/**
 * @file component.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "component.h"
#include "xml/config_xml.h"

static int yac_components_new_first_call = 1;

// general component struct
struct component {

  // general component data

  char * name;
  unsigned int global_comp_id;
  MPI_Comm comm;
  unsigned io_proc_step;
};

static struct {
  char const * name;
  MPI_Group group;
} * all_components;
static size_t total_num_components;

static unsigned compute_io_proc_step(unsigned comm_size) {

  unsigned num_procs_per_node;

  // estimate the number of processes per node
  if (comm_size % 36 == 0) {
    num_procs_per_node = 36;
  } else if (comm_size % 24 == 0) {
    num_procs_per_node = 24;
  } else if (comm_size % 16 == 0) {
    num_procs_per_node = 16;
  } else if (comm_size % 12 == 0) {
    num_procs_per_node = 12;
  } else if (comm_size % 8 == 0) {
    num_procs_per_node = 8;
  } else if (comm_size % 6 == 0) {
    num_procs_per_node = 6;
  } else if (comm_size % 4 == 0) {
    num_procs_per_node = 4;
  } else if (comm_size % 2 == 0) {
    num_procs_per_node = 2;
  } else {
    num_procs_per_node = 1;
  }

  unsigned num_nodes = comm_size / num_procs_per_node;
  unsigned num_io_procs = MIN((unsigned)sqrt((float)num_nodes), 64);
  return ((num_nodes + num_io_procs - 1) / num_io_procs) * num_procs_per_node;
}

//taken from http://beige.ucs.indiana.edu/I590/node85.html
static void mpi_err_handler(int error_code, MPI_Comm comm) {

   if (error_code != MPI_SUCCESS) {

      int rank;

      MPI_Comm_rank(comm, &rank);

      char error_string[1024];
      int length_of_error_string, error_class;

      MPI_Error_class(error_code, &error_class);
      MPI_Error_string(error_class, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Error_string(error_code, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Abort(comm, error_code);
   }
}

void yac_components_new(char const ** names, unsigned num_names,
                        struct component ** comps) {

   if (yac_components_new_first_call) yac_components_new_first_call = 0;
   else yac_internal_abort_message(
     "ERROR: yac_component_new can only be called once", __FILE__, __LINE__ );

#ifdef YAC_XML_ENABLED
  total_num_components = (size_t)yac_xml_get_number_of_components();
#else
  total_num_components = (size_t)num_names;
#endif

  void *attr;
  int flag;
  xt_mpi_call(MPI_Comm_get_attr(MPI_COMM_WORLD, MPI_TAG_UB, &attr, &flag),
              MPI_COMM_WORLD);
  if (!flag)
    yac_internal_abort_message(
      "ERROR(yac_components_new): unable to retrieve MPI_TAG_UB value",
      __FILE__, __LINE__);
  int mpi_tag_ub_val = *(int *)attr;


  if (((1 << total_num_components) - 1) > mpi_tag_ub_val)
    yac_internal_abort_message(
      "ERROR(yac_components_new): too many components, please contact the "
      "author", __FILE__, __LINE__);

  unsigned * global_comp_ids = xmalloc(num_names * sizeof(*global_comp_ids));
  int * idx = xmalloc(num_names * sizeof(*idx));

  // for all names get component ids from xml
  for (unsigned i = 0; i < num_names; ++i) {
#ifdef YAC_XML_ENABLED
    global_comp_ids[i] = yac_xml_get_component_id_by_component_name(names[i]);
#else
    global_comp_ids[i] = yac_hash(names[i]);
#endif
    if (global_comp_ids[i] == (unsigned)-1) {
      fprintf (stderr, "Check def_comp call and xml file for %s!\n", names[i]);
      yac_internal_abort_message(
        "ERROR(yac_components_new): mismatch in component name",
        __FILE__, __LINE__);
    }
    idx[i] = i;
  }

  yac_quicksort_index((int*)global_comp_ids, (size_t)num_names, idx);

  // check for duplicated component ids
  for (unsigned i = 1; i < num_names; ++i) {
    if (global_comp_ids[i-1] == global_comp_ids[i])
      yac_internal_abort_message(
        "ERROR(yac_components_new): duplicated component ids",
        __FILE__, __LINE__);
  }

  all_components = xmalloc(total_num_components * sizeof(*all_components));

  MPI_Group yac_mpi_world_group;
  mpi_err_handler(
    MPI_Comm_group(yac_mpi_comm_world, &yac_mpi_world_group),
    yac_mpi_comm_world);

  // for all components in xml
  for (size_t i = 0, comp_idx = 0; i < total_num_components; ++i) {
#ifdef YAC_XML_ENABLED
    int curr_comp_id = yac_xml_get_component_id(i);
    all_components[i].name = yac_xml_get_component_name(i);
#else
    int curr_comp_id = global_comp_ids[i];
    all_components[i].name = NULL;
#endif

    unsigned curr_comp_idx;
    for (curr_comp_idx = 0; curr_comp_idx < num_names; ++curr_comp_idx)
      if (global_comp_ids[curr_comp_idx] == curr_comp_id) break;

    // if the local process has the current component
    if (curr_comp_idx < num_names) {

      struct component * comp = xmalloc(1 * sizeof(*comp));

      comps[comp_idx++] = comp;

      char const * curr_comp_name = names[idx[curr_comp_idx]];
      comp->name = xmalloc(strlen(curr_comp_name) + 1);
      strcpy(comp->name, curr_comp_name);
      
      comp->global_comp_id = curr_comp_id;
      mpi_err_handler(
        MPI_Comm_split(yac_mpi_comm_world, 0, 0, &(comp->comm)),
        yac_mpi_comm_world);

      // create group for all processes in the component communicator
      MPI_Group comp_group;
      mpi_err_handler(
        MPI_Comm_group(comp->comm, &comp_group), comp->comm);
      int comp_group_size;
      mpi_err_handler(
        MPI_Group_size(comp_group, &comp_group_size), comp->comm);
      int * rank_buffer =
        xmalloc(2 * (size_t)comp_group_size * sizeof(*rank_buffer));
      int * comp_ranks = rank_buffer,
          * world_ranks = rank_buffer + (size_t)comp_group_size;
      for (int j = 0; j < comp_group_size; ++j) comp_ranks[j] = j;
      mpi_err_handler(
        MPI_Group_translate_ranks(
          comp_group, comp_group_size, comp_ranks,
          yac_mpi_world_group, world_ranks), comp->comm);
      mpi_err_handler(
        MPI_Group_incl(yac_mpi_world_group, comp_group_size, world_ranks,
                       &(all_components[i].group)), comp->comm);
      mpi_err_handler(MPI_Group_free(&comp_group), yac_mpi_comm_world);
      free(rank_buffer);

      comp->io_proc_step = compute_io_proc_step(comp_group_size);

    } else {

      MPI_Comm dummy_comm;
      mpi_err_handler(
        MPI_Comm_split(yac_mpi_comm_world, 1, 0, &dummy_comm),
        yac_mpi_comm_world);

      // create group for all processes in the component communicator
      MPI_Group non_comp_group;
      mpi_err_handler(
        MPI_Comm_group(dummy_comm, &non_comp_group), dummy_comm);
      int non_comp_group_size;
      mpi_err_handler(
        MPI_Group_size(non_comp_group, &non_comp_group_size), dummy_comm);
      int * rank_buffer =
        xmalloc(2 * (size_t)non_comp_group_size * sizeof(*rank_buffer));
      int * non_comp_ranks = rank_buffer,
          * world_ranks = rank_buffer + (size_t)non_comp_group_size;
      for (int j = 0; j < non_comp_group_size; ++j) non_comp_ranks[j] = j;
      mpi_err_handler(
        MPI_Group_translate_ranks(
          non_comp_group, non_comp_group_size, non_comp_ranks,
          yac_mpi_world_group, world_ranks), dummy_comm);
      mpi_err_handler(
        MPI_Group_excl(yac_mpi_world_group, non_comp_group_size, world_ranks,
                       &(all_components[i].group)), yac_mpi_comm_world);
      mpi_err_handler(MPI_Group_free(&non_comp_group), yac_mpi_comm_world);
      free(rank_buffer);
      mpi_err_handler(MPI_Comm_free(&dummy_comm), yac_mpi_comm_world);
    }  
  }

  mpi_err_handler(MPI_Group_free(&yac_mpi_world_group), yac_mpi_comm_world);

  free(idx);
  free(global_comp_ids);
}

struct component * yac_component_new(char const * name) {

  struct component * comp;

  yac_components_new(&name, 1, &comp);

  return comp;
}

const char * yac_component_get_name(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

  return comp->name;
}

unsigned int yac_component_get_global_id(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->global_comp_id;
}

MPI_Comm yac_component_get_communicator(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->comm;
}

void yac_component_set_io_proc_step(struct component * comp,
                                    unsigned io_proc_step) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   if (io_proc_step == 0)
      yac_internal_abort_message("ERROR: invalid io_proc_step",
                                 __FILE__, __LINE__ );

   comp->io_proc_step = io_proc_step;
}

unsigned yac_component_get_io_proc_step(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->io_proc_step;
}

MPI_Comm yac_component_get_pair_rootcomm(
  const char * comp_name1, const char * comp_name2) {


  if (!strcmp(comp_name1, comp_name2))
    yac_internal_abort_message(
      "ERROR(yac_component_get_pair_rootcomm): both component names are "
      "identical", __FILE__, __LINE__);

  size_t comp_idx1 = SIZE_MAX, comp_idx2 = SIZE_MAX;

  for (size_t i = 0; i < total_num_components; ++i) {

    if (!strcmp(all_components[i].name, comp_name1)) comp_idx1 = i;
    if (!strcmp(all_components[i].name, comp_name2)) comp_idx2 = i;
  }

  char format_string[] =
    "ERROR(yac_component_get_pair_rootcomm): invalid component name: \"%d\"";
  char err_string[2048];
  if (comp_idx1 == SIZE_MAX) {
    snprintf(err_string, 2048, format_string, comp_name1);
    yac_internal_abort_message(err_string, __FILE__, __LINE__);
  }
  if (comp_idx2 == SIZE_MAX) {
    snprintf(err_string, 2048, format_string, comp_name2);
    yac_internal_abort_message(err_string, __FILE__, __LINE__);
  }

  MPI_Group root_group;

  mpi_err_handler(
    MPI_Group_union(
      all_components[comp_idx1].group, all_components[comp_idx2].group,
      &root_group), yac_mpi_comm_world);

  MPI_Comm root_comm;
  int tag = (1 << comp_idx1) | (1 << comp_idx2);
  mpi_err_handler(
    MPI_Comm_create_group(
      yac_mpi_comm_world, root_group, tag, &root_comm), yac_mpi_comm_world);

  mpi_err_handler(MPI_Group_free(&root_group), yac_mpi_comm_world);

return root_comm;
}

void yac_component_delete(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

  free(comp->name);
  mpi_err_handler(MPI_Comm_free(&(comp->comm)), yac_mpi_comm_world);
  free(comp);
}

void yac_components_reset() {
  yac_components_new_first_call = 1;
  for (size_t i = 0; i < total_num_components; ++i)
    mpi_err_handler(
      MPI_Group_free(&(all_components[i].group)), yac_mpi_comm_world);
  free(all_components);
  all_components = NULL;
  total_num_components = 0;
}

