/**
 * @file couple_config.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "couple_config.h"
#include "utils.h"
#include "xml/config_xml.h"
#include "yac_interface.h"
#include "event.h"
#include "yac_mpi.h"
#include "fields.h"
#include "component.h"
#include "interpolation.h"
#include "xml/interpolation_enum.h"
#include "ensure_array_size.h"
#include "geometry.h"
#include "dist_grid.h"
#include "interp_grid.h"
#include "interp_weights.h"
#include "interp_method.h"
#include "interp_method_avg.h"
#include "interp_method_check.h"
#include "interp_method_conserv.h"
#include "interp_method_nnn.h"
#include "interp_method_fixed.h"
#include "interp_method_file.h"
#include "interp_method_hcsbb.h"
#include "interp_method_spmap.h"

#ifdef YAC_XML_ENABLED
#define MAX_FILE_NAME_LENGHT (512)
#define MAX_ROUTINE_NAME_LENGTH (256)

enum couple_status {
  INITIALISED  = 0,
  FINALIZED    = 1,
  FREED        = 2
};

enum transient_type {
  SRC = 1,
  TGT = 2
};

struct comp_grid_config {
  const char * grid_name;
  const char * comp_name;
};

struct comp_grid_pair_config {
  struct comp_grid_config config[2];
};

struct src_transient_config {
  struct coupling_field * field;
  int id;

  union { // - every interpolation type needs to have the entry type as first
          //   entry in order to be able to compare them using memcmp
          // - no pointers are allowed here in order to be comparable
    struct {
      enum yac_interpolation_list type;
    } general;
    struct {
      enum yac_interpolation_list type;
      enum yac_interp_avg_weight_type reduction_type;
      int partial_coverage;
    } average;
    struct {
      enum yac_interpolation_list type;
      struct yac_nnn_config config;
    } n_nearest_neighbor,
      radial_basis_function;
    struct {
      enum yac_interpolation_list type;
      int order;
      int enforced_conserv;
      int partial_coverage;
      enum yac_interp_method_conserv_normalisation normalisation;
    } conservative;
    struct {
      enum yac_interpolation_list type;
    } spmap;
    struct {
      enum yac_interpolation_list type;
    } hcsbb;
    struct {
      enum yac_interpolation_list type;
      char filename[MAX_FILE_NAME_LENGHT];
      char src_grid_name[MAX_FILE_NAME_LENGHT];
      char tgt_grid_name[MAX_FILE_NAME_LENGHT];
      int io_proc_step;
    } user_file;
    struct {
      enum yac_interpolation_list type;
      double value;
    } fixed;
    struct {
      enum yac_interpolation_list type;
      char constructor_key[MAX_ROUTINE_NAME_LENGTH];
      char do_search_key[MAX_ROUTINE_NAME_LENGTH];
    } check;
  } * interp_stack_config;
  size_t interp_stack_size;
  const char * weight_file_name;
  int event_id;
};

struct tgt_transient_config {
  struct coupling_field * field;
  int id;
  int event_id;
};

struct transient_config {

  int id;

  struct comp_grid_pair_config comp_grid_pair;
  int src_comp_idx;

  struct src_transient_config src_interp_config;
  struct tgt_transient_config tgt_interp_config;

  enum interp_weights_reorder_type reorder_type;

  size_t collection_size;
};

static struct basic_grid_data *
  get_basic_grid_data(const char * grid_name, int global_component_id,
                      unsigned num_fields,
                      struct coupling_field * coupling_fields[num_fields]) {

  for (unsigned i = 0; i < num_fields; ++i) {

    struct coupling_field * field = coupling_fields[i];

    if (global_component_id != yac_get_coupling_field_global_comp_id(field))
      continue;

    struct basic_grid * grid = yac_coupling_field_get_basic_grid(field);

    if (strcmp(grid_name, grid->name) == 0) return &(grid->data);
  }

  return &basic_grid_data_empty;
}

static struct component *
get_local_component(int global_comp_id, unsigned num_components,
                    struct component * components[num_components]) {

  for (unsigned i = 0; i < num_components; ++i)
    if (yac_component_get_global_id(components[i]) == global_comp_id)
      return components[i];

  return NULL;
}

static
int compare_comp_grid_config(const void * a, const void * b) {

  struct comp_grid_config * a_ = (struct comp_grid_config *)a,
                          * b_ = (struct comp_grid_config *)b;

  int ret;
  if ((ret = strcmp(a_->comp_name, b_->comp_name))) return ret;
  else return strcmp(a_->grid_name, b_->grid_name);
}

static struct coupling_field * get_coupling_field(
  int comp_id, const char * transient_name,
  unsigned num_fields, struct coupling_field * coupling_fields[num_fields]) {

  for (unsigned i = 0; i < num_fields; ++i) {
    struct coupling_field * curr_field = coupling_fields[i];
    if ((comp_id ==
         yac_get_coupling_field_global_comp_id(curr_field)) &&
         !strcmp(transient_name, yac_get_coupling_field_name(curr_field)))
      return curr_field;
  }

  return NULL;
}

static int bool_str_2_int(const char * bool_str) {

  int ret;

  if (!strcmp(bool_str, "true") ||
      !strcmp(bool_str, "True") ||
      !strcmp(bool_str, "TRUE")) {

    ret = 1;
  } else {
    ret = 0;
    if (!(!strcmp(bool_str, "false") ||
          !strcmp(bool_str, "False") ||
          !strcmp(bool_str, "FALSE")))
      yac_internal_abort_message(
        "ERROR(bool_str_2_int): invalid value boolean string",
        __FILE__, __LINE__);
  }
  return ret;
}

static struct src_transient_config read_src_interp_config_from_xml(
  int couple_idx, int transient_idx) {

  struct src_transient_config src_config;

  src_config.interp_stack_size =
     (size_t)yac_xml_get_number_of_interpolations(couple_idx, transient_idx);
  src_config.interp_stack_config =
    xcalloc(
      src_config.interp_stack_size, sizeof(*(src_config.interp_stack_config)));

  for (size_t i = 0; i < src_config.interp_stack_size; ++i) {

    src_config.interp_stack_config[i].general.type =
      yac_xml_get_interpolation_method(couple_idx, transient_idx, i);

    if (src_config.interp_stack_config[i].general.type == UNDEFINED)
      yac_internal_abort_message(
        "ERROR(read_src_interp_config_from_xml): couple config: unsupported or "
        "undefined interpolation in xml file!", __FILE__, __LINE__ );

    switch ((int)(src_config.interp_stack_config[i].general.type)) {

      case (AVERAGE): {

        int partial_cov = 0;

        if (yac_xml_interpolation_has_parameter(
              couple_idx, transient_idx, i, "partial_coverage")) {

          const char * partial_cov_str =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i, "partial_coverage");
          partial_cov = bool_str_2_int(partial_cov_str);
          yac_xml_free_xml_string(partial_cov_str);
        }

        enum yac_interp_avg_weight_type reduction_type = AVG;

        if (yac_xml_interpolation_has_parameter(
              couple_idx, transient_idx, i, "weighted")) {

            const char * weighted_str =
              yac_xml_get_interpolation_parameter_value_string(
                couple_idx, transient_idx, i, "weighted");

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

                reduction_type = DIST;

            } else if (!strcmp(weighted_str, "arithmetic_average") ||
                       !strcmp(weighted_str, "ARITHMETIC_AVERAGE")) {

                reduction_type = AVG;
            } else {

              yac_internal_abort_message(
                "ERROR(read_src_interp_config_from_xml): "
                "invalid \"weighted\" value for n-nearest_neighbor",
                __FILE__, __LINE__);
            }

            yac_xml_free_xml_string(weighted_str);
        }

        src_config.interp_stack_config[i].average.partial_coverage =
          partial_cov;
        src_config.interp_stack_config[i].average.reduction_type =
          reduction_type;
        break;
      }
      case (N_NEAREST_NEIGHBOR): {

        if (yac_xml_interpolation_has_parameter(
              couple_idx, transient_idx, i, "n") == 0)
          yac_internal_abort_message(
            "ERROR: start_search:  N_NEAREST_NEIGHBOR requires n!",
            __FILE__, __LINE__ );

        int nn =
          yac_xml_get_interpolation_parameter_value_int(
            couple_idx, transient_idx, i, "n" );

        enum yac_interp_nnn_weight_type reduction_type = NNN_AVG;

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "weighted")) {

            const char * weighted_str =
              yac_xml_get_interpolation_parameter_value_string(
                couple_idx, transient_idx, i,
                "weighted");

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

                reduction_type = NNN_DIST;

            } else if (!strcmp(weighted_str, "gauss_weighted") ||
                       !strcmp(weighted_str, "GAUSS_WEIGHTED")) {

                reduction_type = NNN_GAUSS;

            } else if (!strcmp(weighted_str, "arithmetic_average") ||
                       !strcmp(weighted_str, "ARITHMETIC_AVERAGE")) {

                reduction_type = NNN_AVG;
            } else {

              yac_internal_abort_message(
                "ERROR(read_src_interp_config_from_xml): "
                "invalid \"weighted\" value for n-nearest_neighbor",
                __FILE__, __LINE__);
            }

            yac_xml_free_xml_string(weighted_str);
        }

        double gauss_scale = -1;

        if (reduction_type == NNN_GAUSS) {
          if (yac_xml_interpolation_has_parameter(couple_idx,
                                                  transient_idx,
                                                  i, "gauss_scale")) {


          gauss_scale =
            yac_xml_get_interpolation_parameter_value_double(
              couple_idx, transient_idx, i, "gauss_scale");
          } else {
            gauss_scale = YAC_DEFAULT_GAUSS_SCALE;
          }
        }

        src_config.interp_stack_config[i].
          n_nearest_neighbor.config.data.gauss_scale = gauss_scale;
        src_config.interp_stack_config[i].
          n_nearest_neighbor.config.n = (size_t)nn;
        src_config.interp_stack_config[i].
          n_nearest_neighbor.config.type = reduction_type;
        break;
      }
      case (CONSERVATIVE): {
        int order = 1;
        int enforced_conserv = 0;
        int partial_coverage = 0;
        enum yac_interp_method_conserv_normalisation
          normalisation = CONSERV_DESTAREA;

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx, i,
                                                "order" ) == 1)
           order =
             yac_xml_get_interpolation_parameter_value_int(
               couple_idx, transient_idx, i,
               "order" );

        if ((order != 1) && (order != 2))
          yac_internal_abort_message(
            "ERROR: invalid value for conservative remapping XML parameter "
            "order", __FILE__, __LINE__);

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "enforced_conservation")) {
          const char * enforced_conserv_str =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i,
              "enforced_conservation");
          if (!strcmp(enforced_conserv_str, "true") ||
              !strcmp(enforced_conserv_str, "True") ||
              !strcmp(enforced_conserv_str, "TRUE")) {

            enforced_conserv = 1;
          } else if (!(!strcmp(enforced_conserv_str, "false") ||
                       !strcmp(enforced_conserv_str, "False") ||
                       !strcmp(enforced_conserv_str, "FALSE"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for conservative remapping XML parameter "
              "enforced_conservation", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(enforced_conserv_str);
        }

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "partial_coverage")) {

          const char * partial_cov_str =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i,
              "partial_coverage");

          if (!strcmp(partial_cov_str, "true") ||
              !strcmp(partial_cov_str, "True") ||
              !strcmp(partial_cov_str, "TRUE")) {

            partial_coverage = 1;
          } else if (!(!strcmp(partial_cov_str, "false") ||
                       !strcmp(partial_cov_str, "False") ||
                       !strcmp(partial_cov_str, "FALSE"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for conservative remapping XML parameter "
              "partial_coverage", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(partial_cov_str);
        }

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx, i,
                                                "normalisation")) {

          const char * normalisation_str =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i,
              "normalisation");

          if (!strcmp(normalisation_str, "fracarea") ||
              !strcmp(normalisation_str, "FRACAREA")) {

            normalisation = CONSERV_FRACAREA;

          } else if (!(!strcmp(normalisation_str, "destarea") ||
                       !strcmp(normalisation_str, "DESTAREA"))) {

            normalisation = CONSERV_DESTAREA;

          }

          yac_xml_free_xml_string(normalisation_str);
        }

        src_config.interp_stack_config[i]
          .conservative.order = order;
        src_config.interp_stack_config[i]
          .conservative.enforced_conserv = enforced_conserv;
        src_config.interp_stack_config[i]
          .conservative.partial_coverage = partial_coverage;
        src_config.interp_stack_config[i]
          .conservative.normalisation = normalisation;
        break;
      }
      case (SOURCE_TO_TARGET_MAP):
      {
        break;
      }
      case (FIXED_VALUE):
      {
        if (yac_xml_interpolation_has_parameter(
              couple_idx, transient_idx, i, "user_value" ) == 0)
          yac_internal_abort_message(
            "ERROR(read_src_interp_config_from_xml): start_search: FIXED_VALUE "
            "requires a value defined by the user!", __FILE__, __LINE__);

        src_config.interp_stack_config[i].fixed.value =
          yac_xml_get_interpolation_parameter_value_double(
            couple_idx, transient_idx, i, "user_value" );
        break;
      }
      case (USER_FILE):
      {

        /* get the filename containing the weights */

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                 transient_idx, i,
                                                 "filename" ) == 0)
          yac_internal_abort_message(
            "ERROR(read_src_interp_config_from_xml): "
            "USER_FILE requires a filename containing "
            "weights defined by the user!", __FILE__, __LINE__);

        char const * filename =
          yac_xml_get_interpolation_parameter_value_string(
            couple_idx, transient_idx, i, "filename" );

        if (strlen(filename) + 1 > MAX_FILE_NAME_LENGHT)
          yac_internal_abort_message("ERROR: USER_FILE filename is too long!",
                                     __FILE__, __LINE__);
        strcpy(src_config.interp_stack_config[i].user_file.filename,
               filename);

        yac_xml_free_xml_string(filename);

        src_config.interp_stack_config[i].user_file.io_proc_step = -1;

        char const * src_grid_name =
          yac_xml_get_grid_alias_name_by_grid_id(
            yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
              yac_xml_get_source_component_id(couple_idx, transient_idx),
              yac_xml_get_source_transient_grid_ref_id(
                couple_idx, transient_idx)));
        if (strlen(src_grid_name) + 1 > MAX_FILE_NAME_LENGHT)
          yac_internal_abort_message("ERROR: USER_FILE src_grid_name is too long!",
                                     __FILE__, __LINE__);
        strcpy(src_config.interp_stack_config[i].user_file.src_grid_name,
               src_grid_name);
        yac_xml_free_xml_string(src_grid_name);

        char const * tgt_grid_name =
          yac_xml_get_grid_alias_name_by_grid_id(
            yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
              yac_xml_get_target_component_id(couple_idx, transient_idx),
              yac_xml_get_target_transient_grid_ref_id(
                couple_idx, transient_idx)));
        if (strlen(tgt_grid_name) + 1 > MAX_FILE_NAME_LENGHT)
          yac_internal_abort_message("ERROR: USER_FILE tgt_grid_name is too long!",
                                     __FILE__, __LINE__);
        strcpy(src_config.interp_stack_config[i].user_file.tgt_grid_name,
               tgt_grid_name);
        yac_xml_free_xml_string(tgt_grid_name);

        break;
      }
      case (CHECK):
      {
        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "constructor_key") != 0) {

          char const * constructor_key =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i,
              "constructor_key");

          if (strlen(constructor_key) + 1 > MAX_ROUTINE_NAME_LENGTH)
            yac_internal_abort_message(
              "ERROR: CHECK constructor_key is too long!", __FILE__, __LINE__);
          strcpy(src_config.interp_stack_config[i].
                   check.constructor_key, constructor_key);

          yac_xml_free_xml_string(constructor_key);
        }

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "do_search_key") != 0) {

          char const * do_search_key =
            yac_xml_get_interpolation_parameter_value_string(
              couple_idx, transient_idx, i,
              "do_search_key");

          if (strlen(do_search_key) + 1 > MAX_ROUTINE_NAME_LENGTH)
            yac_internal_abort_message(
              "ERROR: CHECK do_search_key is too long!", __FILE__, __LINE__);
          strcpy(src_config.interp_stack_config[i].
                   check.do_search_key, do_search_key);

          yac_xml_free_xml_string(do_search_key);
        }

        break;
      }
      case (BERNSTEIN_BEZIER):
      {
        break;
      }
      case (RADIAL_BASIS_FUNCTION):
      {

        int nn;

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx, i,
                                                "n")) {
          nn =
            yac_xml_get_interpolation_parameter_value_int(
              couple_idx, transient_idx, i, "n" );
        } else {
          nn = 9;
        }

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "rbf_kernel")) {

            const char * weighted_str =
              yac_xml_get_interpolation_parameter_value_string(
                couple_idx, transient_idx, i,
                "rbf_kernel");

            if (!strcmp(weighted_str, "gauss_kernel") &&
                !strcmp(weighted_str, "GAUSS_KERNEL")) {

              yac_internal_abort_message(
                "ERROR: invalid value for RBF XML parameter "
                "rbf_kernel", __FILE__, __LINE__);
            }

            yac_xml_free_xml_string(weighted_str);
        }

        double rbf_scale;

        if (yac_xml_interpolation_has_parameter(couple_idx,
                                                transient_idx,
                                                i, "rbf_scale")) {

            rbf_scale =
              yac_xml_get_interpolation_parameter_value_double(
                couple_idx, transient_idx, i,
                "rbf_scale");
        } else {
            rbf_scale = YAC_DEFAULT_RBF_SCALE;
        }

        src_config.interp_stack_config[i].
          radial_basis_function.config.data.rbf_scale = rbf_scale;
        src_config.interp_stack_config[i].
          radial_basis_function.config.n = (size_t)nn;
        src_config.interp_stack_config[i].
          n_nearest_neighbor.config.type = NNN_RBF;
        break;
      }
      default:
        yac_internal_abort_message(
          "ERROR(read_src_interp_config_from_xml): start_search: unknown "
          "interpolation!", __FILE__, __LINE__ );
    };
  }

  // default is 1
  src_config.weight_file_name =
    (yac_xml_enforce_write_weight_file(couple_idx, transient_idx))?
    yac_xml_get_weight_file_name(couple_idx, transient_idx):NULL;

  return src_config;
}

static int compare_interp_field(const void * a, const void * b) {

  struct interp_field * a_ = (struct interp_field *)a,
                      * b_ = (struct interp_field *)b;

  int ret;
  if ((ret = (int)(a_->location) - (int)(b_->location))) return ret;
  if ((ret = (a_->coordinates > b_->coordinates) -
             (a_->coordinates < b_->coordinates))) return ret;
  return (a_->mask > b_->mask) < (a_->mask < b_->mask);
}

static int compare_transient_config_fields(const void * a, const void * b) {

  struct coupling_field * a_ = (struct coupling_field *)a,
                        * b_ = (struct coupling_field *)b;

  if ((a_ == NULL) || (b_ == NULL)) return (a_ == NULL) - (b_ == NULL);

  struct interp_field const * interp_fields[2] =
    {yac_coupling_field_get_interp_fields(a_),
     yac_coupling_field_get_interp_fields(b_)};
  size_t num_interp_fields[2] =
    {yac_coupling_field_get_num_interp_fields(a_),
     yac_coupling_field_get_num_interp_fields(b_)};

  // if both transients have different numbers of interpolation fields
  if (num_interp_fields[0] != num_interp_fields[1])
    return (int)num_interp_fields[0] - (int)num_interp_fields[1];

  // if interpolation fields are defined for only one transient
  if ((interp_fields[0] == NULL) ^ (interp_fields[1] == NULL))
    return (interp_fields[0] == NULL) - (interp_fields[1] == NULL);

  for (size_t i = 0; i < num_interp_fields[0]; ++i) {
    int ret =
      compare_interp_field(&(interp_fields[0][i]), &(interp_fields[1][i]));
    if (ret) return ret;
  }

  return 0;
}

static int compare_transient_config_field_ids(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret = a_->src_interp_config.id - b_->src_interp_config.id;
  if (ret) return ret;
  return a_->tgt_interp_config.id - b_->tgt_interp_config.id;
}

static int
compare_transient_config_interp_method(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret;
  if ((ret = (int)(a_->src_interp_config.weight_file_name == NULL) -
             (int)(b_->src_interp_config.weight_file_name == NULL))) return ret;
  if ((a_->src_interp_config.weight_file_name != NULL) &&
      (ret = strcmp(a_->src_interp_config.weight_file_name,
                    b_->src_interp_config.weight_file_name))) return ret;
  if ((ret = (int)(a_->src_interp_config.interp_stack_size) -
             (int)(b_->src_interp_config.interp_stack_size))) return ret;

  size_t stack_size = a_->src_interp_config.interp_stack_size;
  for (size_t i = 0; i < stack_size; ++i) {

    if ((ret =
           (int)(a_->src_interp_config.interp_stack_config[i].general.type) -
           (int)(b_->src_interp_config.interp_stack_config[i].general.type)))
      return ret;

    switch ((int)(a_->src_interp_config.interp_stack_config[i].general.type)) {
      case(AVERAGE):
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           average.reduction_type) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           average.reduction_type)))
          return ret;
        if ((ret = a_->src_interp_config.interp_stack_config[i].
                     average.partial_coverage -
                   b_->src_interp_config.interp_stack_config[i].
                     average.partial_coverage))
          return ret;
        break;
      case (N_NEAREST_NEIGHBOR):
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           n_nearest_neighbor.config.type) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           n_nearest_neighbor.config.type)))
          return ret;
        if ((ret =
               (a_->src_interp_config.interp_stack_config[i].
                  n_nearest_neighbor.config.n >
                b_->src_interp_config.interp_stack_config[i].
                  n_nearest_neighbor.config.n) -
               (a_->src_interp_config.interp_stack_config[i].
                  n_nearest_neighbor.config.n <
                b_->src_interp_config.interp_stack_config[i].
                  n_nearest_neighbor.config.n)))
          return ret;
        if (a_->src_interp_config.interp_stack_config[i].
                 n_nearest_neighbor.config.type == NNN_GAUSS) {
          if ((ret =
                 (a_->src_interp_config.interp_stack_config[i].
                    n_nearest_neighbor.config.data.gauss_scale >
                  b_->src_interp_config.interp_stack_config[i].
                    n_nearest_neighbor.config.data.gauss_scale) -
                 (a_->src_interp_config.interp_stack_config[i].
                    n_nearest_neighbor.config.data.gauss_scale <
                  b_->src_interp_config.interp_stack_config[i].
                    n_nearest_neighbor.config.data.gauss_scale)))
            return ret;
        }
      case (BERNSTEIN_BEZIER):
        return 0;
      case (RADIAL_BASIS_FUNCTION):
        if ((ret =
               (a_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.n >
                b_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.n) -
               (a_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.n <
                b_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.n)))
          return ret;
        if ((ret =
               (a_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.data.rbf_scale >
                b_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.data.rbf_scale) -
               (a_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.data.rbf_scale <
                b_->src_interp_config.interp_stack_config[i].
                  radial_basis_function.config.data.rbf_scale)))
          return ret;
      case(CONSERVATIVE):
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           conservative.order) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           conservative.order)))
          return ret;
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           conservative.enforced_conserv) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           conservative.enforced_conserv)))
          return ret;
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           conservative.partial_coverage) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           conservative.partial_coverage)))
          return ret;
        if ((ret = (int)(a_->src_interp_config.interp_stack_config[i].
                           conservative.normalisation) -
                   (int)(b_->src_interp_config.interp_stack_config[i].
                           conservative.normalisation)))
          return ret;
      case(FIXED_VALUE):
        if ((ret =
               (a_->src_interp_config.interp_stack_config[i].fixed.value >
                b_->src_interp_config.interp_stack_config[i].fixed.value) -
               (a_->src_interp_config.interp_stack_config[i].fixed.value <
                b_->src_interp_config.interp_stack_config[i].fixed.value)))
          return ret;
        break;
      case(CHECK):
        if ((ret = strcmp(a_->src_interp_config.interp_stack_config[i].
                            check.constructor_key,
                          b_->src_interp_config.interp_stack_config[i].
                            check.constructor_key)))
          return ret;
        if ((ret = strcmp(a_->src_interp_config.interp_stack_config[i].
                            check.do_search_key,
                          b_->src_interp_config.interp_stack_config[i].
                            check.do_search_key)))
          return ret;
        break;
      default:
        yac_internal_abort_message(
          "ERROR(compare_transient_config_interp_method): "
          "invalid interpolation type", __FILE__, __LINE__);
    }
  }
  return 0;
}

static
int compare_transient_config(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[0].comp_name,
                    b_->comp_grid_pair.config[0].comp_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[1].comp_name,
                    b_->comp_grid_pair.config[1].comp_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[0].grid_name,
                    b_->comp_grid_pair.config[0].grid_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[1].grid_name,
                    b_->comp_grid_pair.config[1].grid_name))) return ret;
  if ((ret = a_->src_interp_config.id - b_->src_interp_config.id)) return ret;
  if ((ret = a_->tgt_interp_config.id - b_->tgt_interp_config.id)) return ret;
  if ((ret = compare_transient_config_interp_method(a_, b_))) return ret;
  return (int)(a_->reorder_type) - (int)(b_->reorder_type);
}

int read_event_info_from_xml(
  int couple_idx, int transient_idx, enum transient_type transient_type) {

  enum yac_reduction_type reduction_operation;

  switch (yac_xml_get_coupling_period_operation(couple_idx, transient_idx)) {
    case (T_AVERAGE):
      reduction_operation = TIME_AVERAGE;
      break;
    case (T_ACCUMULATE):
      reduction_operation = ACCUMULATE;
      break;
    default:
      reduction_operation = INSTANT;
      break;
  };

  enum yac_time_unit_type time_unit;

  switch ( yac_xml_get_timestep_unit() ) {
    case (T_MILLISECOND):
      time_unit = C_MILLISECOND;
      break;
    case (T_SECOND):
      time_unit = C_SECOND;
      break;
    case (T_MINUTE):
      time_unit = C_MINUTE;
      break;
    case (T_HOUR):
      time_unit = C_HOUR;
      break;
    case (T_DAY):
      time_unit = C_DAY;
      break;
    case (T_MONTH):
      time_unit = C_MONTH;
      break;
    case (T_YEAR):
      time_unit = C_YEAR;
      break;
    case (T_ISO_FORMAT):
      time_unit = C_ISO_FORMAT;
      break;
    default:
      yac_internal_abort_message(
        "ERROR in xml configuration file: unsupported time unit",
        __FILE__, __LINE__ );
  };

  // make sure that the calendar is defined and consistent

  calendarType set_calendar = getCalendarType ();
  calendarType my_calendar  = PROLEPTIC_GREGORIAN; //< \todo get calendar type from XML.

  if ( set_calendar == CALENDAR_NOT_SET )
    initCalendar(my_calendar);
  else if ( set_calendar != my_calendar )
    yac_internal_abort_message(
      "ERROR: different calendars in use! Reconfigure model setup.\n",
      __FILE__, __LINE__);

  int event_id;

  yac_event_init(&event_id);

  if ( time_unit == C_ISO_FORMAT) {

      char *timestep;
      int timelag;

      if ( transient_type == SRC ) {
          timestep = yac_xml_get_source_isotimestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_source_timelag(couple_idx, transient_idx);
      } else {
          timestep = yac_xml_get_target_isotimestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_target_timelag(couple_idx, transient_idx);
      }

      char * period = yac_xml_get_isocoupling_period(couple_idx, transient_idx);

      char * start_datetime = yac_get_start_datetime();
      char * end_datetime = yac_get_end_datetime();

      yac_event_add(
        event_id, timestep, period, time_unit,
        timelag, reduction_operation, start_datetime, end_datetime);

  } else {

      int timestep;
      int timelag;

      if ( transient_type == SRC ) {
          timestep = yac_xml_get_source_timestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_source_timelag(couple_idx, transient_idx);
      } else {
          timestep = yac_xml_get_target_timestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_target_timelag(couple_idx, transient_idx);
      }

      int period = yac_xml_get_coupling_period(couple_idx, transient_idx);
      char * start_datetime = yac_get_start_datetime();
      char * end_datetime = yac_get_end_datetime();

      yac_event_add(
        event_id, &timestep, &period, time_unit,
        timelag, reduction_operation, start_datetime, end_datetime);
  }

  return event_id;
}

static void read_transient_config_from_xml(
  struct transient_config ** transient_configs_, size_t * count,
  unsigned num_fields, struct coupling_field * coupling_fields[num_fields]) {

  int num_couples = yac_xml_get_number_of_couples();
  size_t total_num_transients = 0;
  for (int couple_idx = 0; couple_idx < num_couples; ++couple_idx)
    total_num_transients +=
      (size_t)yac_xml_get_number_of_transient_couples(couple_idx);

  struct transient_config * transient_configs =
    xmalloc(total_num_transients * sizeof(*transient_configs));

  size_t i = 0;

  // extract component, grid configurations from all transients in
  // the xml file
  for (int couple_idx = 0; couple_idx < num_couples; ++couple_idx) {

    int curr_num_transients =
      yac_xml_get_number_of_transient_couples(couple_idx);

    for (int transient_idx = 0; transient_idx < curr_num_transients;
         ++transient_idx, ++i) {

      struct comp_grid_config src_config, tgt_config;

      int src_comp_id =
        yac_xml_get_source_component_id(couple_idx, transient_idx);
      int tgt_comp_id =
        yac_xml_get_target_component_id(couple_idx, transient_idx);

      src_config.comp_name =
        yac_xml_get_component_name_by_component_id(src_comp_id);
      tgt_config.comp_name =
        yac_xml_get_component_name_by_component_id(tgt_comp_id);;

      int src_transient_grid_ref_id =
        yac_xml_get_source_transient_grid_ref_id(couple_idx, transient_idx);
      int tgt_transient_grid_ref_id =
        yac_xml_get_target_transient_grid_ref_id(couple_idx, transient_idx);

      src_config.grid_name =
        yac_xml_get_grid_alias_name_by_grid_id(
          yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
            src_comp_id, src_transient_grid_ref_id));
      tgt_config.grid_name =
        yac_xml_get_grid_alias_name_by_grid_id(
          yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(
            tgt_comp_id, tgt_transient_grid_ref_id));

      int src_comp_idx =
        compare_comp_grid_config(&src_config, &tgt_config) > 0;

      const char * transient_name =
        yac_xml_get_transient_standard_name_by_transient_id(
          yac_xml_get_transient_couple_transient_id(couple_idx, transient_idx));
      struct coupling_field * src_field =
        get_coupling_field(
          src_comp_id, transient_name, num_fields, coupling_fields);
      struct coupling_field * tgt_field =
        get_coupling_field(
          tgt_comp_id, transient_name, num_fields, coupling_fields);
      yac_xml_free_xml_string(transient_name);

      size_t collection_size =
        (size_t)yac_xml_get_transient_collection_size(
          src_comp_id, src_transient_grid_ref_id);
      if ((size_t)yac_xml_get_transient_collection_size(
            tgt_comp_id, tgt_transient_grid_ref_id) != collection_size)
        yac_internal_abort_message(
          "ERROR(read_transient_config_from_xml): collection_size missmatch",
          __FILE__, __LINE__);
      if (src_field != NULL)
        yac_set_coupling_field_collection_size(src_field, collection_size);
      if (tgt_field != NULL)
        yac_set_coupling_field_collection_size(tgt_field, collection_size);

      transient_configs[i].comp_grid_pair.config[src_comp_idx] = src_config;
      transient_configs[i].comp_grid_pair.config[src_comp_idx^1] = tgt_config;
      transient_configs[i].src_comp_idx = src_comp_idx;
      transient_configs[i].src_interp_config =
        read_src_interp_config_from_xml(couple_idx, transient_idx);
      transient_configs[i].src_interp_config.field = src_field;
      transient_configs[i].tgt_interp_config.field = tgt_field;
      transient_configs[i].reorder_type =
        (yac_xml_mapping_on_source(couple_idx, transient_idx))?
        (MAPPING_ON_SRC):(MAPPING_ON_TGT);
      transient_configs[i].src_interp_config.event_id =
        read_event_info_from_xml(couple_idx, transient_idx, SRC);
      transient_configs[i].tgt_interp_config.event_id =
        read_event_info_from_xml(couple_idx, transient_idx, TGT);
      transient_configs[i].collection_size = collection_size;
    }
  }

  { // determine unique transient field config ids, for each unique configuration
    int * transient_src_field_config_compare_matrix =
      xmalloc((size_t)(total_num_transients * (total_num_transients - 1)) *
              sizeof(*transient_src_field_config_compare_matrix));
    int * transient_tgt_field_config_compare_matrix =
      transient_src_field_config_compare_matrix +
      (size_t)(total_num_transients * (total_num_transients - 1)) / 2;

    for (unsigned i = 0, k = 0; i < total_num_transients; ++i) {
      for (unsigned j = i + 1; j < total_num_transients; ++j, ++k) {
        transient_src_field_config_compare_matrix[k] =
          compare_transient_config_fields(
            transient_configs[i].src_interp_config.field,
            transient_configs[j].src_interp_config.field) != 0;
        transient_tgt_field_config_compare_matrix[k] =
          compare_transient_config_fields(
            transient_configs[i].tgt_interp_config.field,
            transient_configs[j].tgt_interp_config.field) != 0;
      }
    }

    yac_mpi_call(
      MPI_Allreduce(MPI_IN_PLACE, transient_src_field_config_compare_matrix,
                    total_num_transients * (total_num_transients - 1),
                    MPI_INT, MPI_LOR, yac_mpi_comm_world), yac_mpi_comm_world);

    int * transient_src_field_config_ids =
      xcalloc(2 * (size_t)total_num_transients,
              sizeof(*transient_src_field_config_ids));
    int * transient_tgt_field_config_ids =
      transient_src_field_config_ids + (size_t)total_num_transients;

    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_src_field_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_src_field_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_src_field_config_compare_matrix[k])
            transient_src_field_config_ids[j] = (int)id;
        id++;
      }
    }
    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_tgt_field_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_tgt_field_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_tgt_field_config_compare_matrix[k])
            transient_tgt_field_config_ids[j] = (int)id;
        id++;
      }
    }
    free(transient_src_field_config_compare_matrix);

    for (unsigned i = 0; i < total_num_transients; ++i) {
      transient_configs[i].src_interp_config.id =
        transient_src_field_config_ids[i];
      transient_configs[i].tgt_interp_config.id =
        transient_tgt_field_config_ids[i];
    }
    free(transient_src_field_config_ids);
  }

  { // determine unique transient config ids, for each unique configuration
    int * transient_config_compare_matrix =
      xmalloc((size_t)(total_num_transients * (total_num_transients - 1)) / 2 *
              sizeof(*transient_config_compare_matrix));
    for (unsigned i = 0, k = 0; i < total_num_transients; ++i)
      for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
        transient_config_compare_matrix[k] =
          compare_transient_config(
            &(transient_configs[i]), &(transient_configs[j])) != 0;

    yac_mpi_call(
      MPI_Allreduce(MPI_IN_PLACE, transient_config_compare_matrix,
                    (total_num_transients * (total_num_transients - 1)) / 2,
                    MPI_INT, MPI_LOR, yac_mpi_comm_world), yac_mpi_comm_world);

    int * transient_config_ids =
      xcalloc((size_t)total_num_transients, sizeof(*transient_config_ids));

    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_config_compare_matrix[k])
            transient_config_ids[j] = (int)id;
        id++;
      }
    }
    free(transient_config_compare_matrix);

    for (unsigned i = 0; i < total_num_transients; ++i)
      transient_configs[i].id = transient_config_ids[i];
    free(transient_config_ids);
  }

  *transient_configs_ = transient_configs;
  *count = total_num_transients;
}

static
int compare_transient_config_ids(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  return a_->id - b_->id;
}

static struct interp_weights * generate_interp_weights(
  struct src_transient_config src_interp_config,
  struct interp_grid * interp_grid) {

  size_t interp_stack_size = src_interp_config.interp_stack_size;
  struct interp_method ** method_stack =
    xmalloc((interp_stack_size + 1) * sizeof(*method_stack));
  method_stack[interp_stack_size] = NULL;

  for (size_t i = 0; i < interp_stack_size; ++i) {

    switch((int)(src_interp_config.interp_stack_config[i].general.type)) {
      case(AVERAGE): {
        enum yac_interp_avg_weight_type weight_type =
          src_interp_config.interp_stack_config[i].average.reduction_type;
        int partial_coverage =
          (int)src_interp_config.interp_stack_config[i].average.partial_coverage;
        method_stack[i] =
          yac_interp_method_avg_new(weight_type,partial_coverage);
        break;
      }
      case(CONSERVATIVE): {
        int order =
          src_interp_config.interp_stack_config[i].conservative.order;
        int enforced_conserv =
          src_interp_config.interp_stack_config[i].conservative.enforced_conserv;
        int partial_coverage =
          src_interp_config.interp_stack_config[i].conservative.partial_coverage;
        enum yac_interp_method_conserv_normalisation normalisation =
          src_interp_config.interp_stack_config[i].conservative.normalisation;

        method_stack[i] =
          yac_interp_method_conserv_new(
            order, enforced_conserv, partial_coverage, normalisation);
        break;
      }
      case(FIXED_VALUE): {
        double fixed_value =
          src_interp_config.interp_stack_config[i].fixed.value;
        method_stack[i] = yac_interp_method_fixed_new(fixed_value);
        break;
      }
      case(USER_FILE): {
        char const * weight_file_name =
          src_interp_config.interp_stack_config[i].user_file.filename;
        int io_proc_step =
          src_interp_config.interp_stack_config[i].user_file.io_proc_step;
        char const * src_grid_name =
          src_interp_config.interp_stack_config[i].user_file.src_grid_name;
        char const * tgt_grid_name =
          src_interp_config.interp_stack_config[i].user_file.tgt_grid_name;
        method_stack[i] =
          yac_interp_method_file_new(
            weight_file_name, io_proc_step, src_grid_name, tgt_grid_name);
        break;
      }
      case(CHECK): {
        func_constructor constructor_callback;
        void * constructor_user_data;
        func_do_search do_search_callback;
        void * do_search_user_data;

        yac_interp_method_check_get_constructor_callback(
          src_interp_config.interp_stack_config[i].check.constructor_key,
          &constructor_callback, &constructor_user_data);
        yac_interp_method_check_get_do_search_callback(
          src_interp_config.interp_stack_config[i].check.do_search_key,
          &do_search_callback, &do_search_user_data);

        method_stack[i] =
          yac_interp_method_check_new(
            constructor_callback, constructor_user_data,
            do_search_callback, do_search_user_data);
        break;
      }
      case (N_NEAREST_NEIGHBOR): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            src_interp_config.interp_stack_config[i].
              n_nearest_neighbor.config);
        break;
      }
      case (BERNSTEIN_BEZIER): {
        method_stack[i] =
          yac_interp_method_hcsbb_new();
        break;
      }
      case (RADIAL_BASIS_FUNCTION): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            src_interp_config.interp_stack_config[i].
              radial_basis_function.config);
        break;
      }
      case (SOURCE_TO_TARGET_MAP): {
        method_stack[i] =
          yac_interp_method_spmap_new();
        break;
      }
      default:
        yac_internal_abort_message(
          "ERROR(generate_interp_weights): unsupported interpolation method",
          __FILE__, __LINE__);
    };
  }

  struct interp_weights * weights =
    yac_interp_method_do_search(method_stack, interp_grid);

  yac_interp_method_delete(method_stack);
  free(method_stack);

  return weights;
}

static struct interp_field get_interp_field_from_coupling_field(
  struct coupling_field * field, MPI_Comm comm) {

  if (field != NULL) {

    if (yac_coupling_field_get_num_interp_fields(field) != 1)
      yac_internal_abort_message(
        "ERROR(get_interp_field_from_coupling_field): unsupported number of "
        "interpolation fields", __FILE__, __LINE__);

    struct interp_field interp_field =
      *(yac_coupling_field_get_interp_fields(field));
    int location = (int)(interp_field.location);

    yac_mpi_call(
      MPI_Allreduce(
        MPI_IN_PLACE, &location, 1, MPI_INT, MPI_MIN, comm), comm);

    if (location != (int)(interp_field.location))
      yac_internal_abort_message(
        "ERROR(get_interp_field_from_coupling_field): location mismatch",
        __FILE__, __LINE__);

    return interp_field;

  } else {

    int location = (int)LOC_UNDEFINED;

    yac_mpi_call(
      MPI_Allreduce(
        MPI_IN_PLACE, &location, 1, MPI_INT, MPI_MIN, comm), comm);

    struct interp_field interp_field;
    interp_field.location = (enum yac_location)location;
    interp_field.coordinates = NULL;
    interp_field.mask = NULL;
    return interp_field;
  }
}

static void generate_interpolations(
  unsigned num_components, struct component * components[num_components],
  unsigned num_fields, struct coupling_field * coupling_fields[num_fields]) {

  // extract information about all transients from the XML file
  struct transient_config * transient_configs;
  size_t transient_count;
  read_transient_config_from_xml(
    &transient_configs, &transient_count, num_fields, coupling_fields);

  // sort transient configurations
  qsort(transient_configs, transient_count, sizeof(*transient_configs),
        compare_transient_config_ids);

  struct dist_grid_pair * dist_grid_pair = NULL;
  struct interp_grid * interp_grid = NULL;
  struct interp_weights * interp_weights = NULL;
  struct interpolation * interp = NULL;
  struct comp_grid_pair_config * prev_comp_grid_pair = NULL;
  MPI_Comm comp_pair_comm = MPI_COMM_NULL;
  int is_active = 0;

  // loop over all transients to build interpolations
  for (size_t i = 0; i < transient_count; ++i) {

    struct transient_config * curr_transient_config = transient_configs + i;
    struct comp_grid_pair_config * curr_comp_grid_pair =
      &(curr_transient_config->comp_grid_pair);

    int build_flag = 0;
    // if the current comp pair is different from the previous one
    // => generate a component pair communicator
    if ((prev_comp_grid_pair == NULL) ||
        (strcmp(prev_comp_grid_pair->config[0].comp_name,
                curr_comp_grid_pair->config[0].comp_name) ||
         strcmp(prev_comp_grid_pair->config[1].comp_name,
                curr_comp_grid_pair->config[1].comp_name))) {

      build_flag = 1;

      if (comp_pair_comm != MPI_COMM_NULL)
        yac_mpi_call(MPI_Comm_free(&comp_pair_comm), yac_mpi_comm_world);

      int comp_is_available[2] = {
        (get_local_component(
           yac_xml_get_component_id_by_component_name(
             curr_comp_grid_pair->config[0].comp_name),
           num_components, components) != NULL),
        (get_local_component(
           yac_xml_get_component_id_by_component_name(
             curr_comp_grid_pair->config[1].comp_name),
           num_components, components) != NULL)};

      is_active = comp_is_available[0] || comp_is_available[1];

      yac_mpi_call(MPI_Comm_split(yac_mpi_comm_world, is_active, 0,
                                  &comp_pair_comm), yac_mpi_comm_world);

      // determine whether both components are available in the setup
      yac_mpi_call(MPI_Allreduce(MPI_IN_PLACE, comp_is_available, 2,
                                 MPI_INT, MPI_LOR, comp_pair_comm),
                   yac_mpi_comm_world);
      is_active = comp_is_available[0] && comp_is_available[1];
    }

    // if the current configuration differs from the previous one and the local
    // process is involved in this configuration
    if (is_active &&
        (build_flag ||
         strcmp(prev_comp_grid_pair->config[0].grid_name,
                curr_comp_grid_pair->config[0].grid_name) ||
         strcmp(prev_comp_grid_pair->config[1].grid_name,
                curr_comp_grid_pair->config[1].grid_name))) {

      build_flag = 1;

      if (dist_grid_pair != NULL) yac_dist_grid_pair_delete(dist_grid_pair);

      struct basic_grid_data * basic_grid_data[2] =
        {get_basic_grid_data(
           curr_comp_grid_pair->config[0].grid_name,
           yac_xml_get_component_id_by_component_name(
             curr_comp_grid_pair->config[0].comp_name),
           num_fields, coupling_fields),
         get_basic_grid_data(
           curr_comp_grid_pair->config[1].grid_name,
           yac_xml_get_component_id_by_component_name(
             curr_comp_grid_pair->config[1].comp_name),
           num_fields, coupling_fields)};

      dist_grid_pair =
        yac_dist_grid_pair_new(
          basic_grid_data[0],
          yac_xml_get_grid_id_by_grid_name(
            curr_comp_grid_pair->config[0].grid_name),
          basic_grid_data[1],
          yac_xml_get_grid_id_by_grid_name(
            curr_comp_grid_pair->config[1].grid_name),
          comp_pair_comm);
    }

    // if the current source or target field data differes from the previous
    // one
    if (is_active &&
        (build_flag ||
         compare_transient_config_field_ids(
           curr_transient_config - 1, curr_transient_config))) {

      build_flag = 1;

      size_t num_src_fields = 1;
      struct interp_field const src_fields[1] =
        {get_interp_field_from_coupling_field(
           curr_transient_config->src_interp_config.field, comp_pair_comm)};
      struct interp_field const tgt_field =
        get_interp_field_from_coupling_field(
          curr_transient_config->tgt_interp_config.field, comp_pair_comm);
      int src_comp_idx = transient_configs[i].src_comp_idx;
      int src_grid_id =
        yac_xml_get_grid_id_by_grid_name(
          curr_comp_grid_pair->config[src_comp_idx].grid_name);
      int tgt_grid_id =
        yac_xml_get_grid_id_by_grid_name(
          curr_comp_grid_pair->config[src_comp_idx^1].grid_name);

      if (interp_grid != NULL) yac_interp_grid_delete(interp_grid);

      // generate interp grid
      interp_grid = yac_interp_grid_new(
        dist_grid_pair, src_grid_id, tgt_grid_id, num_src_fields,
        src_fields, tgt_field);
    }

    // if the current interpolation method stack differes from the previous
    // configuration
    if (is_active &&
        (build_flag ||
         compare_transient_config_interp_method(
           curr_transient_config - 1, curr_transient_config))) {

      build_flag = 1;

      if (interp_weights != NULL) yac_interp_weights_delete(interp_weights);

      // generate interp weights
      interp_weights = generate_interp_weights(
        curr_transient_config->src_interp_config, interp_grid);
    }

    if (is_active &&
        (curr_transient_config->src_interp_config.weight_file_name != NULL)) {

      int src_comp_idx = transient_configs[i].src_comp_idx;

      yac_interp_weights_write_to_file(
        interp_weights,
        curr_transient_config->src_interp_config.weight_file_name,
        curr_comp_grid_pair->config[src_comp_idx].grid_name,
        curr_comp_grid_pair->config[src_comp_idx^1].grid_name, -1);
    }

    // if the current weight reorder method differs from the previous
    // configuration
    if (is_active &&
        (build_flag ||
         (curr_transient_config[-1].reorder_type !=
          curr_transient_config->reorder_type) ||
         (curr_transient_config[-1].collection_size !=
          curr_transient_config->collection_size))) {

      if (interp != NULL) yac_interpolation_delete(interp);

      // generate interpolation
      interp =
        yac_interp_weights_get_interpolation(
          interp_weights, curr_transient_config->reorder_type,
          curr_transient_config->collection_size);
    }

    if (is_active) {

      int is_source = curr_transient_config->src_interp_config.field != NULL;
      int is_target = curr_transient_config->tgt_interp_config.field != NULL;

      if (is_source)
        yac_set_coupling_field_put_op(
          curr_transient_config->src_interp_config.field,
          curr_transient_config->src_interp_config.event_id,
          yac_interpolation_copy(interp));
      else yac_event_free(curr_transient_config->src_interp_config.event_id);

      if (is_target)
        yac_set_coupling_field_get_op(
          curr_transient_config->tgt_interp_config.field,
          curr_transient_config->tgt_interp_config.event_id,
          yac_interpolation_copy(interp));
      else yac_event_free(curr_transient_config->tgt_interp_config.event_id);

    } else {
      yac_event_free(curr_transient_config->src_interp_config.event_id);
      yac_event_free(curr_transient_config->tgt_interp_config.event_id);
    }

    prev_comp_grid_pair = curr_comp_grid_pair;
  }

  yac_interpolation_delete(interp);
  yac_interp_weights_delete(interp_weights);
  yac_interp_grid_delete(interp_grid);
  yac_dist_grid_pair_delete(dist_grid_pair);
  if (comp_pair_comm != MPI_COMM_NULL)
    yac_mpi_call(MPI_Comm_free(&comp_pair_comm), yac_mpi_comm_world);
  for (size_t i = 0; i < transient_count; ++i) {
    free((void*)(transient_configs[i].comp_grid_pair.config[0].grid_name));
    free((void*)(transient_configs[i].comp_grid_pair.config[1].grid_name));
    free(transient_configs[i].src_interp_config.interp_stack_config);
    free((void*)(transient_configs[i].src_interp_config.weight_file_name));
  }
  free(transient_configs);
}

void yac_setup_couple_config(unsigned num_components,
                             struct component * components[num_components],
                             unsigned num_fields,
                             struct coupling_field * coupling_fields[num_fields]) {

  generate_interpolations(
    num_components, components, num_fields, coupling_fields);
}

MPI_Comm yac_get_pair_rootcomm_ ( char const * comp_name1,
                                  char const * comp_name2 ) {

  return yac_component_get_pair_rootcomm(comp_name1, comp_name2);
}

#else // YAC_XML_ENABLED

void yac_setup_couple_config(unsigned num_components,
                             struct component * components[num_components],
                             unsigned num_fields,
                             struct coupling_field * coupling_fields[num_fields]) {
  fprintf (stderr, "WARNING: yac was compiled without xml support!\n");
  fprintf (stdout, "WARNING: yac was compiled without xml support!\n");
}

#endif // YAC_XML_ENABLED

