/**
 * @file grid.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Thomas Jahns <jahns@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>

#include "grid.h"
#include "grid_cell.h"

// the order must match the entries in enum  yac_location
static char const * yac_location_str[] = {"CELL",
                                          "CORNER",
                                          "EDGE",
                                          "UNDEFINED"};

enum yac_location yac_str2loc(char const * location) {

  if (!strncmp("CELL", location, YAC_MAX_LOC_STR_LEN))
    return CELL;
  else if (!strncmp("CORNER", location, YAC_MAX_LOC_STR_LEN))
    return CORNER;
  else if (!strncmp("EDGE", location, YAC_MAX_LOC_STR_LEN))
    return EDGE;
  else if (!strncmp("UNDEFINED", location, YAC_MAX_LOC_STR_LEN))
    return LOC_UNDEFINED;
  else
      yac_internal_abort_message("ERROR(yac_str2loc): invalid location\n",
                                 __FILE__, __LINE__);

  // routine will never reach this point
  return LOC_UNDEFINED;
}

char const * yac_loc2str(enum yac_location location) {

  switch (location) {
    case (CELL):
    case (CORNER):
    case (EDGE):
    case (LOC_UNDEFINED):
      break;
    default:
      yac_internal_abort_message("ERROR(yac_loc2str): location must be one of "
                                 "CORNER/EDGE/CELL/LOC_UNDEFINED.", __FILE__,
                                 __LINE__ );
  };

  return yac_location_str[location];
}

enum yac_location yac_get_location(int const location) {

  switch (location) {
    case (CELL):
    case (CORNER):
    case (EDGE):
    case (LOC_UNDEFINED):
      return (enum yac_location) location;
    default:
      yac_internal_abort_message("ERROR: get_location: location must be one "
                                 "of CORNER/EDGE/CELL/LOC_UNDEFINED.", __FILE__,
                                 __LINE__ );
  };

  // routine will never reach this point
  return CORNER;
}

void yac_basic_grid_data_get_cell(
  struct basic_grid_data * grid, size_t cell_idx, struct grid_cell * cell) {

  if (cell_idx >= grid->num_total_cells)
    yac_internal_abort_message(
      "ERROR(yac_basic_grid_data_get_cell): invalid cell_idx",
      __FILE__, __LINE__);

  size_t num_vertices =
    (cell->num_corners = grid->num_vertices_per_cell[cell_idx]);
  if (num_vertices > cell->array_size) {

    cell->coordinates_xyz =
      xrealloc(cell->coordinates_xyz,
               num_vertices * sizeof(*(cell->coordinates_xyz)));
    cell->edge_type =
      xrealloc(cell->edge_type, num_vertices * sizeof(*(cell->edge_type)));
    cell->array_size = num_vertices;
  }

  size_t * vertices =
    grid->cell_to_vertex + grid->cell_to_vertex_offsets[cell_idx];
  for (int j = 0; j < num_vertices; ++j) {
    size_t vertex_idx = vertices[j];
    for (int k = 0; k < 3; ++k)
      cell->coordinates_xyz[j][k] = grid->vertex_coordinates[vertex_idx][k];
  }
}

void yac_basic_grid_data_free(struct basic_grid_data grid) {

  free(grid.vertex_coordinates);
  free(grid.cell_ids);
  free(grid.vertex_ids);
  free(grid.edge_ids);
  free(grid.core_cell_mask);
  free(grid.core_vertex_mask);
  free(grid.core_edge_mask);
  free(grid.num_vertices_per_cell);
  free(grid.num_cells_per_vertex);
  free(grid.cell_to_vertex);
  free(grid.cell_to_vertex_offsets);
  free(grid.cell_to_edge);
  if (grid.cell_to_vertex_offsets != grid.cell_to_edge_offsets)
    free(grid.cell_to_edge_offsets);
  free(grid.vertex_to_cell);
  free(grid.vertex_to_cell_offsets);
  free(grid.edge_to_vertex);
  free(grid.edge_type);
}

void yac_basic_grid_free(struct basic_grid grid) {

  yac_basic_grid_data_free(grid.data);
  free(grid.name);
}
