/**
 * @file interp_method_file.c
 *
 * @copyright Copyright  (C)  2020 Moritz Hanke <hanke@dkrz.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>

#include "interp_method_file.h"
#include "dist_grid.h"
#include "yac_mpi.h"
#include "io_utils.h"

static size_t do_search_file(struct interp_method * method,
                              struct interp_grid * interp_grid,
                              size_t * tgt_points, size_t count,
                              struct interp_weights * weights);
static void delete_file(struct interp_method * method);

static struct interp_method_vtable
  interp_method_file_vtable = {
    .do_search = do_search_file,
    .delete = delete_file};

struct interp_method_file {

  struct interp_method_vtable * vtable;
  char const * weight_file_name;
  char const * src_grid_name;
  char const * tgt_grid_name;
  int io_proc_step;
};

struct link_data {
  struct {
    yac_int global_id;
  } src, tgt;
  double weight;
  size_t src_field_idx;
};

struct fixed_data {
  double value;
  struct {
    yac_int global_id;
  } * tgt;
  size_t num_tgt;
};

struct temp_fixed_data {
  double value;
  struct {
    yac_int global_id;
  } tgt;
};

struct temp_result {
  enum {
    LINK = 0,
    FIXED = 1,
    NO_INTERP = 2,
  } type;
  union {
    struct {
      struct link_data * links;
      size_t count;
    } link;
    struct {
      double value;
    } fixed;
  } data;
  struct {
    yac_int global_id;
    size_t local_id;
  } tgt;
  int owner;
  size_t reorder_idx;
};

static void read_weight_file(
  char const * weight_file_name, int io_proc_step, MPI_Comm comm,
  char const * src_grid_name, char const * tgt_grid_name,
  enum yac_location * src_locations, enum yac_location tgt_location,
  size_t num_src_fields,
  struct link_data ** links, size_t * num_links,
  struct fixed_data ** fixed, size_t * num_fixed) {

  *links = NULL;
  *num_links = 0;
  *fixed = NULL;
  *num_fixed = 0;

#ifndef YAC_NETCDF_ENABLED
  UNUSED(weight_file_name);
  UNUSED(io_proc_step);
  UNUSED(comm);
  UNUSED(src_grid_name);
  UNUSED(tgt_grid_name);
  UNUSED(src_locations);
  UNUSED(tgt_location);
  UNUSED(num_src_fields);
  UNUSED(links);
  UNUSED(num_links);

  yac_internal_abort_message(
    "ERROR(read_weight_file): YAC_NETCDF_ENABLED not defined", __FILE__, __LINE__);
#else

  int num_io_procs, io_idx;

  // if the local process does not have to read anything from the file
  if (!yac_is_io_proc(&io_proc_step, comm, &num_io_procs, &io_idx)) return;

  // open weight file on io processes
  int ncid;
  int status = nc_open(weight_file_name, NC_NOWRITE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  //\todo use parallel netcdf if available

  int dimid, var_id;

  // global attributes
  size_t version_len;
  char * version;
  status = nc_inq_attlen(ncid, NC_GLOBAL, "version", &version_len);
  if (status != NC_NOERR) handle_error(status);
  version = xmalloc(version_len + 1);
  version[version_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "version", version);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(YAC_WEIGHT_FILE_VERSION_STRING) != version_len) ||
      strncmp(version, YAC_WEIGHT_FILE_VERSION_STRING, version_len)) {
    if (io_idx == 0)
      fprintf(stderr, "version: %s (weight file) %s (interface)\n",
              version, YAC_WEIGHT_FILE_VERSION_STRING);
    yac_internal_abort_message(
      "ERROR(read_weight_file): version string from weight file does not match "
      "with YAC version\n", __FILE__, __LINE__);
  }
  free(version);

  size_t grid_name_len;
  char * grid_name;
  status = nc_inq_attlen(ncid, NC_GLOBAL, "src_grid_name", &grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  grid_name = xmalloc(grid_name_len + 1);
  grid_name[grid_name_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "src_grid_name", grid_name);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(src_grid_name) != grid_name_len) ||
      strncmp(src_grid_name, grid_name, grid_name_len)) {
    if (io_idx == 0)
      fprintf(stderr, "grid name: %s (weight file) %s (interface)\n",
              grid_name, src_grid_name);
    yac_internal_abort_message(
      "ERROR(read_weight_file): source grid name from weight file does not "
      "match with the one provided through the interface\n",
      __FILE__, __LINE__);
  }

  free(grid_name);

  status = nc_inq_attlen(ncid, NC_GLOBAL, "dst_grid_name", &grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  grid_name = xmalloc(grid_name_len + 1);
  grid_name[grid_name_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "dst_grid_name", grid_name);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(tgt_grid_name) != grid_name_len) ||
      strncmp(tgt_grid_name, grid_name, grid_name_len)) {
    if (io_idx == 0)
      fprintf(stderr, "grid name: %s (weight file) %s (interface)\n",
              grid_name, tgt_grid_name);
    yac_internal_abort_message(
      "ERROR(read_weight_file): target grid name from weight file does not match "
      "with the one provided through the interface\n", __FILE__, __LINE__);
  }

  free(grid_name);

  //---------------
  // read link data
  //---------------

  size_t str_link_len;
  char * str_link;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_links", &str_link_len);
  if (status != NC_NOERR) handle_error(status);
  str_link = xmalloc(str_link_len + 1);
  str_link[str_link_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_links", str_link);
  if (status != NC_NOERR) handle_error(status);

  int contains_links = (strlen("TRUE") == str_link_len) &&
                       !strncmp("TRUE", str_link, str_link_len);
  if (!contains_links &&
      ((strlen("FALSE") != str_link_len) ||
       strncmp("FALSE", str_link, str_link_len)))
    yac_internal_abort_message("ERROR(read_weight_file): invalid global "
                               "attribute contains_links", __FILE__,
                               __LINE__);
  free(str_link);

  size_t num_wgts = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_wgts", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_wgts);
    if (status != NC_NOERR) handle_error(status);
    if (num_wgts != 1)
        yac_internal_abort_message("ERROR(read_weight_file): YAC only supports "
                                   "num_wgts == 1\n", __FILE__, __LINE__);
  }

  // get number of links from file

  size_t num_links_in_file = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_links", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_links_in_file);
    if (status != NC_NOERR) handle_error(status);
    if (num_links_in_file == 0)
        yac_internal_abort_message("ERROR(read_weight_file): no links defined",
                                   __FILE__, __LINE__);
  }

  // get number of source fields
  size_t tmp_num_src_fields = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_src_fields", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &tmp_num_src_fields);
    if (status != NC_NOERR) handle_error(status);
    if (tmp_num_src_fields == 0)
        yac_internal_abort_message(
          "ERROR(read_weight_file): no source fields in file\n",
          __FILE__, __LINE__);
    if (tmp_num_src_fields != num_src_fields) {
      if (io_idx == 0)
        fprintf(
          stderr, "number of source fields from file: %zu interface: %zu\n",
          tmp_num_src_fields, num_src_fields);
      yac_internal_abort_message(
        "ERROR(read_weight_file): number of source fields in file does not "
        "match with the number provided through the interface\n",
        __FILE__, __LINE__);
    }
  }

  // get max location string length from file
  size_t max_loc_str_len;
  status = nc_inq_dimid(ncid, "max_loc_str_len", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &max_loc_str_len);
  if (status != NC_NOERR) handle_error(status);
  if (max_loc_str_len != YAC_MAX_LOC_STR_LEN)
    yac_internal_abort_message(
      "ERROR(read_weight_file): wrong max location string length in weight "
      "file", __FILE__, __LINE__);

  // get source locations
  enum yac_location * tmp_src_locations =
    xmalloc(num_src_fields * sizeof(*tmp_src_locations));
  status = nc_inq_varid(ncid, "src_locations", &var_id);
  if (status != NC_NOERR) handle_error(status);

  for (size_t i = 0; i < num_src_fields; ++i) {

    char loc_str[YAC_MAX_LOC_STR_LEN];

    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, YAC_MAX_LOC_STR_LEN};

    status = nc_get_vara_text(ncid, var_id, str_start, str_count, loc_str);
    if (status != NC_NOERR) handle_error(status);

    tmp_src_locations[i] = yac_str2loc(loc_str);
  }

  for (unsigned i = 0; i < num_src_fields; ++i) {

    if (tmp_src_locations[i] != src_locations[i]) {

      int collumn_width[2] = {MAX(strlen("weight file"), YAC_MAX_LOC_STR_LEN),
                              MAX(strlen("interface"), YAC_MAX_LOC_STR_LEN)};
      if (io_idx == 0) {
        fprintf(stderr, "source locations:\n"
                        "%*s : %*s\n",
                        collumn_width[0], "weight file",
                        collumn_width[1], "interface");
        for (unsigned j = 0; j < num_src_fields; ++j) {

          char const * str_location_wgt = yac_loc2str(tmp_src_locations[j]);
          char const * str_location_tgt = yac_loc2str(src_locations[j]);
          fprintf(stderr, "%*s : %*s\n", collumn_width[0], str_location_wgt,
                  collumn_width[1], str_location_tgt);
        }
      }

      yac_internal_abort_message(
        "ERROR(read_weight_file): source locations in file does not match with "
        "the locations provided through the interface\n", __FILE__, __LINE__);
    }
  }

  free(tmp_src_locations);

  // get target location
  status = nc_inq_varid(ncid, "dst_location", &var_id);
  {
    char loc_str[YAC_MAX_LOC_STR_LEN];
    status = nc_get_var_text(ncid, var_id, loc_str);
    if (status != NC_NOERR) handle_error(status);
    if (tgt_location != yac_str2loc(loc_str))
      yac_internal_abort_message(
        "ERROR(read_weight_file): target locations in file does not match with "
        "the locations provided through the interface\n", __FILE__, __LINE__);
  }

    if (contains_links) {

    // get number of links per source field
    unsigned * num_links_per_src_field =
      xmalloc(num_src_fields * sizeof(*num_links_per_src_field));
    status = nc_inq_varid(ncid, "num_links_per_src_field", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_uint(ncid, var_id, num_links_per_src_field);
    if (status != NC_NOERR) handle_error(status);

    size_t local_size =
      ((num_links_in_file + (size_t)num_io_procs) - 1) / (size_t)num_io_procs;

    size_t offset = MIN(local_size * (size_t)io_idx, num_links_in_file);
    size_t count =
      MAX(MIN(local_size, num_links_in_file - local_size * (size_t)io_idx), 0);

    *num_links = count;

    *links = xmalloc(count * sizeof(**links));

    if (count > 0) {
      size_t src_field_idx = 0, src_field_start_offset = 0;
      while ((src_field_idx < num_src_fields) &&
             (src_field_start_offset +
              (size_t)num_links_per_src_field[src_field_idx] <
              offset)) {

        src_field_start_offset += (size_t)num_links_per_src_field[src_field_idx];
        ++src_field_idx;
      };
      if (src_field_idx >= num_src_fields)
        yac_internal_abort_message(
          "ERROR(read_weight_file): problem in num_links_per_src_field",
          __FILE__, __LINE__);
      size_t src_field_offset = offset - src_field_start_offset;
      for (size_t k = 0; (src_field_idx < num_src_fields) && (k < count);
           src_field_offset = 0, ++src_field_idx) {

        for (; (src_field_offset <
                (size_t)num_links_per_src_field[src_field_idx]) &&
               (k < count); ++src_field_offset, ++k) {
          (*links)[k].src_field_idx = src_field_idx;
        }
      }
    }
    free(num_links_per_src_field);

    // get links
    int * address = xmalloc(count * sizeof(*address));
    status = nc_inq_varid(ncid, "src_address", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_vara_int(ncid, var_id, &offset, &count, address);
    if (status != NC_NOERR) handle_error(status);
    for (unsigned i = 0; i < count; ++i)
      (*links)[i].src.global_id = (yac_int)(address[i] - 1);

    status = nc_inq_varid(ncid, "dst_address", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_vara_int(ncid, var_id, &offset, &count, address);
    if (status != NC_NOERR) handle_error(status);
    for (unsigned i = 0; i < count; ++i)
      (*links)[i].tgt.global_id = (yac_int)(address[i] - 1);
    free(address);

    double * weights = xmalloc(count * sizeof(*weights));

    status = nc_inq_varid(ncid, "remap_matrix", &var_id);
    if (status != NC_NOERR) handle_error(status);
    size_t offsets[2] = {offset, 0};
    size_t counts[2] = {count, 1};
    status = nc_get_vara_double(ncid, var_id, offsets, counts, weights);
    if (status != NC_NOERR) handle_error(status);
    for (size_t i = 0; i < count; ++i)
      (*links)[i].weight = weights[i];

    free(weights);
  }

  //----------------
  // read fixed data
  //----------------

  // global attributes
  size_t str_fixed_len;
  char * str_fixed;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_fixed_dst", &str_fixed_len);
  if (status != NC_NOERR) handle_error(status);
  str_fixed = xmalloc(str_fixed_len + 1);
  str_fixed[str_fixed_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_fixed_dst", str_fixed);
  if (status != NC_NOERR) handle_error(status);

  int contains_fixed = (strlen("TRUE") == str_fixed_len) &&
                       !strncmp("TRUE", str_fixed, str_fixed_len);

  if (!contains_fixed &&
      ((strlen("FALSE") != str_fixed_len) ||
       strncmp("FALSE", str_fixed, str_fixed_len)))
    yac_internal_abort_message("ERROR(read_weight_file): invalid global "
                               "attribute contains_fixed_dst", __FILE__,
                               __LINE__);
  free(str_fixed);

  if (contains_fixed) {

    // get number of fixed values
    size_t num_fixed_values;
    status = nc_inq_dimid(ncid, "num_fixed_values", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_fixed_values);
    if (status != NC_NOERR) handle_error(status);

    // get number of fixed target points
    size_t num_fixed_tgt;
    status = nc_inq_dimid(ncid, "num_fixed_dst", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_fixed_tgt);
    if (status != NC_NOERR) handle_error(status);

    size_t local_size =
      ((num_fixed_tgt + (size_t)num_io_procs) - 1) / (size_t)num_io_procs;

    size_t offset = MIN(local_size * (size_t)io_idx, num_fixed_tgt);
    size_t count =
      MAX(MIN(local_size, num_fixed_tgt - local_size * (size_t)io_idx), 0);

    *fixed = xmalloc(num_fixed_values * sizeof(**fixed));
    *num_fixed = num_fixed_values;

    double * fixed_values = xmalloc(num_fixed_values * sizeof(*fixed_values));
    int * num_tgt_indices_per_fixed_value =
      xmalloc(num_fixed_values * sizeof(*num_tgt_indices_per_fixed_value));
    int * global_tgt_indices = xmalloc(count * sizeof(*global_tgt_indices));

    // get number of fixed target points per fixed value
    status = nc_inq_varid(ncid, "num_dst_per_fixed_value", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, num_tgt_indices_per_fixed_value);
    if (status != NC_NOERR) handle_error(status);

    // read data
    status = nc_inq_varid(ncid, "fixed_values", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_double(ncid, var_id, fixed_values);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "dst_address_fixed", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, global_tgt_indices);
    if (status != NC_NOERR) handle_error(status);

    if (count > 0) {
      void * tgt_global_id_buffer = xmalloc(count * sizeof(*((*fixed)->tgt)));
      size_t fixed_value_idx = 0, fixed_value_start_offset = 0;
      while ((fixed_value_idx < num_fixed_values) &&
             (fixed_value_start_offset +
              (size_t)num_tgt_indices_per_fixed_value[fixed_value_idx] <
              offset)) {

        fixed_value_start_offset +=
          (size_t)num_tgt_indices_per_fixed_value[fixed_value_idx];
        (*fixed)[fixed_value_idx].value = fixed_values[fixed_value_idx];
        (*fixed)[fixed_value_idx].num_tgt = 0;
        (*fixed)[fixed_value_idx].tgt = NULL;
        ++fixed_value_idx;
      };
      if (fixed_value_idx >= num_fixed_values)
        yac_internal_abort_message(
          "ERROR(read_weight_file): problem in num_tgt_indices_per_fixed_value",
          __FILE__, __LINE__);
      size_t fixed_value_offset = offset - fixed_value_start_offset;
      for (size_t k = 0; (fixed_value_idx < num_fixed_values) && (k < count);
           fixed_value_offset = 0, ++fixed_value_idx) {

        size_t curr_tgt_count =
          MIN((size_t)num_tgt_indices_per_fixed_value[fixed_value_idx] -
              fixed_value_offset, count - k);

        (*fixed)[fixed_value_idx].value = fixed_values[fixed_value_idx];
        (*fixed)[fixed_value_idx].num_tgt = curr_tgt_count;
        (*fixed)[fixed_value_idx].tgt =
          (void*)(((unsigned char *)tgt_global_id_buffer) +
                  k * sizeof(*((*fixed)->tgt)));

        for (size_t i = 0; i < curr_tgt_count; ++fixed_value_offset, ++k, ++i)
          (*fixed)[fixed_value_idx].tgt[i].global_id =
            (yac_int)(global_tgt_indices[k] - 1);
      }
    }
    free(fixed_values);
    free(global_tgt_indices);
    free(num_tgt_indices_per_fixed_value);
  }

#endif
}

static inline int
compute_bucket(yac_int value, int comm_size) {
  return (int)(value / 128) % comm_size;
}

static int get_tgt_pack_size(MPI_Comm comm) {

  int tgt_pack_size;
  yac_mpi_call(MPI_Pack_size(1, Xt_int_dt, comm, &tgt_pack_size), comm);
  return tgt_pack_size;
}

static int get_link_pack_size(MPI_Comm comm) {

  int global_id_pack_size;
  int weight_pack_size;
  int src_field_idx_pack_size;

  yac_mpi_call(
    MPI_Pack_size(1, Xt_int_dt, comm, &global_id_pack_size), comm);
  yac_mpi_call(
    MPI_Pack_size(1, MPI_DOUBLE, comm, &weight_pack_size), comm);
  yac_mpi_call(
    MPI_Pack_size(1, MPI_UINT64_T, comm, &src_field_idx_pack_size), comm);

  return 2 * global_id_pack_size + weight_pack_size +
         src_field_idx_pack_size;
}

static int get_fixed_pack_size(MPI_Comm comm) {

  int value_pack_size;
  int global_id_pack_size;
  int count_pack_size;

  yac_mpi_call(
    MPI_Pack_size(1, MPI_DOUBLE, comm, &value_pack_size), comm);
  yac_mpi_call(
    MPI_Pack_size(1, Xt_int_dt, comm, &global_id_pack_size), comm);
  yac_mpi_call(
    MPI_Pack_size(1, MPI_UINT64_T, comm, &count_pack_size), comm);

  return value_pack_size + global_id_pack_size + count_pack_size;
}

static void pack_tgt(
  yac_int global_tgt_id, void * buffer, int buffer_size, MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
      MPI_Pack(&global_tgt_id, 1, Xt_int_dt, buffer,
               buffer_size, &position, comm), comm);
}

static void pack_link(
  struct link_data * link, void * buffer, int buffer_size, MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
      MPI_Pack(&(link->src.global_id), 1, Xt_int_dt, buffer,
               buffer_size, &position, comm), comm);
  yac_mpi_call(
      MPI_Pack(&(link->tgt.global_id), 1, Xt_int_dt, buffer,
               buffer_size, &position, comm), comm);
  yac_mpi_call(
      MPI_Pack(&(link->weight), 1, MPI_DOUBLE, buffer,
               buffer_size, &position, comm), comm);
  uint64_t temp_src_field_idx = (uint64_t)(link->src_field_idx);
  yac_mpi_call(
      MPI_Pack(&temp_src_field_idx, 1, MPI_UINT64_T, buffer,
               buffer_size, &position, comm), comm);

}

static void pack_fixed(
  double fixed_value, yac_int global_tgt_id,
  void * buffer, int buffer_size, MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
      MPI_Pack(&fixed_value, 1, MPI_DOUBLE, buffer,
               buffer_size, &position, comm), comm);
  yac_mpi_call(
      MPI_Pack(&global_tgt_id, 1, Xt_int_dt, buffer,
               buffer_size, &position, comm), comm);
}

static void unpack_tgt(
  void * buffer, int buffer_size, yac_int * global_tgt_id, MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, global_tgt_id, 1,
               Xt_int_dt, comm), comm);
}

static void unpack_link(
  void * buffer, int buffer_size, struct link_data * link, MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &(link->src.global_id), 1,
               Xt_int_dt, comm), comm);
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &(link->tgt.global_id), 1,
               Xt_int_dt, comm), comm);
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &(link->weight), 1,
               MPI_DOUBLE, comm), comm);
  uint64_t temp_src_field_idx;
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &temp_src_field_idx, 1,
               MPI_UINT64_T, comm), comm);
  link->src_field_idx = (size_t)temp_src_field_idx;
}

static void unpack_fixed(
  void * buffer, int buffer_size, struct temp_fixed_data * fixed,
  MPI_Comm comm) {

  int position = 0;
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &(fixed->value), 1,
               MPI_DOUBLE, comm), comm);
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &(fixed->tgt.global_id), 1,
               Xt_int_dt, comm), comm);
}

static int compare_temp_result_tgt (void const * a, void const * b) {

  struct temp_result const * result_a = (struct temp_result const *)a;
  struct temp_result const * result_b = (struct temp_result const *)b;

  return (result_a->tgt.global_id > result_b->tgt.global_id) -
         (result_a->tgt.global_id < result_b->tgt.global_id);
}

static int compare_link_data_tgt (void const * a, void const * b) {

  struct link_data const * link_a = (struct link_data const *)a;
  struct link_data const * link_b = (struct link_data const *)b;

  return (link_a->tgt.global_id > link_b->tgt.global_id) -
         (link_a->tgt.global_id < link_b->tgt.global_id);
}

static int compare_temp_result_type (void const * a, void const * b) {

  struct temp_result const * result_a = (struct temp_result const *)a;
  struct temp_result const * result_b = (struct temp_result const *)b;

  return (result_a->type > result_b->type) - (result_a->type < result_b->type);
}

static int compare_temp_result_fixed (void const * a, void const * b) {

  struct temp_result const * result_a = (struct temp_result const *)a;
  struct temp_result const * result_b = (struct temp_result const *)b;

  return (result_a->data.fixed.value > result_b->data.fixed.value) -
         (result_a->data.fixed.value < result_b->data.fixed.value);
}

static int compare_temp_result_reorder_idx (void const * a, void const * b) {

  struct temp_result const * result_a = (struct temp_result const *)a;
  struct temp_result const * result_b = (struct temp_result const *)b;

  return (result_a->reorder_idx > result_b->reorder_idx) -
         (result_a->reorder_idx < result_b->reorder_idx);
}

static int compare_temp_fixed_data_tgt (void const * a, void const * b) {

  struct temp_fixed_data const * fixed_a = (struct temp_fixed_data const *)a;
  struct temp_fixed_data const * fixed_b = (struct temp_fixed_data const *)b;

  return (fixed_a->tgt.global_id > fixed_b->tgt.global_id) -
         (fixed_a->tgt.global_id < fixed_b->tgt.global_id);
}

int get_temp_result_pack_size(struct temp_result * result, MPI_Comm comm) {

  int type_pack_size;
  int data_pack_size;
  int owner_pack_size;

  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &type_pack_size), comm);
  switch(result->type) {
    case(LINK): {
      int count_pack_size;
      int links_pack_size;
      yac_mpi_call(
        MPI_Pack_size(1, MPI_UINT64_T, comm, &count_pack_size), comm);
      links_pack_size =
        (int)(result->data.link.count) * get_link_pack_size(comm);
      data_pack_size = links_pack_size + count_pack_size;
      break;
    }
    case(FIXED): {
      int fixed_value_pack_size;
      yac_mpi_call(
        MPI_Pack_size(1, MPI_DOUBLE, comm, &fixed_value_pack_size), comm);
      data_pack_size = fixed_value_pack_size;
      break;
    }
    case(NO_INTERP): {
      data_pack_size = 0;
      break;
    }
    default: {
      data_pack_size = -1;
      yac_internal_abort_message(
        "ERROR(get_temp_result_pack_size): invalid result type",
        __FILE__, __LINE__);
    }
  };
  int tgt_global_id_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, Xt_int_dt, comm, &tgt_global_id_pack_size), comm);
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &owner_pack_size), comm);

  return type_pack_size + data_pack_size + tgt_global_id_pack_size +
         owner_pack_size;
}

static void pack_temp_result(
  struct temp_result * result, void * buffer, int buffer_size, MPI_Comm comm) {

  int position = 0;
  int temp_type = (int)(result->type);
  yac_mpi_call(
      MPI_Pack(&temp_type, 1, MPI_INT, buffer,
               buffer_size, &position, comm), comm);
  switch(result->type) {
    case(LINK): {
      uint64_t temp_count = (uint64_t)(result->data.link.count);
      yac_mpi_call(
          MPI_Pack(&temp_count, 1, MPI_UINT64_T, buffer,
                   buffer_size, &position, comm), comm);
      for (uint64_t i = 0; i < temp_count; ++i) {
        yac_mpi_call(
            MPI_Pack(&(result->data.link.links[i].src.global_id), 1,
                     Xt_int_dt, buffer, buffer_size, &position, comm), comm);
        yac_mpi_call(
            MPI_Pack(&(result->data.link.links[i].tgt.global_id), 1,
                     Xt_int_dt, buffer, buffer_size, &position, comm), comm);
        yac_mpi_call(
            MPI_Pack(&(result->data.link.links[i].weight), 1,
                     MPI_DOUBLE, buffer, buffer_size, &position, comm), comm);
        uint64_t temp_src_field_idx =
          (uint64_t)result->data.link.links[i].src_field_idx;
        yac_mpi_call(
            MPI_Pack(&temp_src_field_idx, 1, MPI_UINT64_T,
                     buffer, buffer_size, &position, comm), comm);
      }
      break;
    }
    case(FIXED): {
      yac_mpi_call(
          MPI_Pack(&(result->data.fixed.value), 1,
                   MPI_DOUBLE, buffer, buffer_size, &position, comm), comm);
      break;
    }
    case(NO_INTERP): {
      break;
    }
    default: {
      yac_internal_abort_message(
        "ERROR(get_temp_result_pack_size): invalid result type",
        __FILE__, __LINE__);
    }
  };
  yac_mpi_call(
      MPI_Pack(&(result->tgt.global_id), 1,
               Xt_int_dt, buffer, buffer_size, &position, comm), comm);
  yac_mpi_call(
      MPI_Pack(&(result->owner), 1, MPI_INT,
               buffer, buffer_size, &position, comm), comm);
}

static void unpack_temp_result(
  void * buffer, int buffer_size, struct temp_result * result,
  MPI_Comm comm) {

  int position = 0;

  int temp_type;
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position, &temp_type, 1,
               MPI_INT, comm), comm);
  switch (temp_type) {
    case(LINK): {
      result->type = LINK;
      uint64_t temp_count;
      yac_mpi_call(
        MPI_Unpack(buffer, buffer_size, &position, &temp_count, 1,
                   MPI_UINT64_T, comm), comm);
      result->data.link.count = (size_t)temp_count;
      result->data.link.links =
        xmalloc((size_t)temp_count * sizeof(*(result->data.link.links)));
      for (uint64_t i = 0; i < temp_count; ++i) {

        yac_mpi_call(
          MPI_Unpack(buffer, buffer_size, &position,
                     &(result->data.link.links[i].src.global_id), 1,
                     Xt_int_dt, comm), comm);
        yac_mpi_call(
          MPI_Unpack(buffer, buffer_size, &position,
                     &(result->data.link.links[i].tgt.global_id), 1,
                     Xt_int_dt, comm), comm);
        yac_mpi_call(
          MPI_Unpack(buffer, buffer_size, &position,
                     &(result->data.link.links[i].weight), 1,
                     MPI_DOUBLE, comm), comm);
        uint64_t temp_src_field_idx;
        yac_mpi_call(
          MPI_Unpack(buffer, buffer_size, &position, &temp_src_field_idx, 1,
                     MPI_UINT64_T, comm), comm);
        result->data.link.links[i].src_field_idx = (size_t)temp_src_field_idx;
      }
      break;
    }
    case (FIXED): {
      result->type = FIXED;
      yac_mpi_call(
        MPI_Unpack(buffer, buffer_size, &position,
                   &(result->data.fixed.value), 1, MPI_DOUBLE, comm), comm);
      break;
    }
    case (NO_INTERP): {
      result->type = NO_INTERP;
      break;
    }
    default: {
      yac_internal_abort_message(
        "ERROR(unpack_temp_result): invalid result type", __FILE__, __LINE__);
    }
  };
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position,
               &(result->tgt.global_id), 1, Xt_int_dt, comm), comm);
  yac_mpi_call(
    MPI_Unpack(buffer, buffer_size, &position,
               &(result->owner), 1, MPI_INT, comm), comm);
}

static void redist_weight_file_data(
  struct interp_grid * interp_grid, size_t * tgt_points, size_t tgt_count,
  size_t * num_interpolated_points,
  struct link_data ** links, size_t * num_links,
  struct fixed_data ** fixed, size_t * num_fixed) {

  MPI_Comm comm = yac_interp_grid_get_MPI_Comm(interp_grid);

  int comm_size;
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  // get global id of target point that have to be interpolated
  yac_int * tgt_global_ids = xmalloc(tgt_count * sizeof(*tgt_global_ids));
  yac_interp_grid_get_tgt_global_ids(
    interp_grid, tgt_points, tgt_count, tgt_global_ids);

  yac_quicksort_index_yac_int_size_t(tgt_global_ids, tgt_count, tgt_points);

  int tgt_pack_size = get_tgt_pack_size(comm),
      link_pack_size = get_link_pack_size(comm),
      fixed_pack_size = get_fixed_pack_size(comm);

  int * int_buffer =
    xmalloc(4 * (size_t)comm_size * sizeof(*int_buffer));
  int * total_sendcounts = int_buffer + 0 * comm_size;
  int * total_recvcounts = int_buffer + 1 * comm_size;
  int * total_sdispls =    int_buffer + 2 * comm_size;
  int * total_rdispls =    int_buffer + 3 * comm_size;
  int * sendcounts, * recvcounts, * sdispls, *rdispls;
  yac_get_comm_buffers(
    3, &sendcounts, &recvcounts, &sdispls, &rdispls, comm);

  for (size_t i = 0; i < tgt_count; ++i)
    sendcounts[3 * compute_bucket(tgt_global_ids[i], comm_size) + 0]++;
  for (size_t i = 0; i < *num_links; ++i)
    sendcounts[3 * compute_bucket((*links)[i].tgt.global_id, comm_size) + 1]++;
  for (size_t i = 0; i < *num_fixed; ++i)
    for (size_t j = 0; j < (*fixed)[i].num_tgt; ++j)
      sendcounts[
        3 * compute_bucket((*fixed)[i].tgt[j].global_id, comm_size) + 2]++;

  // exchange sendcounts
  yac_mpi_call(MPI_Alltoall(sendcounts, 3, MPI_INT,
                            recvcounts, 3, MPI_INT, comm), comm);

  for (int i = 0, saccu = 0, raccu = 0; i < comm_size; ++i) {
    total_sdispls[i] = saccu;
    sdispls[3*i+0] = saccu;
    sdispls[3*i+1] =
      (saccu + (total_sendcounts[i] = sendcounts[3*i+0] * tgt_pack_size));
    sdispls[3*i+2] =
      (saccu + (total_sendcounts[i] += sendcounts[3*i+1] * link_pack_size));
    total_sendcounts[i] += sendcounts[3*i+2] * fixed_pack_size;
    total_rdispls[i] = raccu;
    rdispls[3*i+0] = raccu;
    rdispls[3*i+1] =
      (raccu + (total_recvcounts[i] = recvcounts[3*i+0] * tgt_pack_size));
    rdispls[3*i+2] =
      (raccu + (total_recvcounts[i] += recvcounts[3*i+1] * link_pack_size));
    total_recvcounts[i] += recvcounts[3*i+2] * fixed_pack_size;
    saccu += total_sendcounts[i];
    raccu += total_recvcounts[i];
  }

  size_t send_size = (size_t)(total_sendcounts[comm_size - 1]) +
                     (size_t)(total_sdispls[comm_size - 1]);
  size_t recv_size = (size_t)(total_recvcounts[comm_size - 1]) +
                     (size_t)(total_rdispls[comm_size - 1]);

  void * pack_buffer = xmalloc(send_size + recv_size);
  void * send_pack_buffer = pack_buffer;
  void * recv_pack_buffer = (void*)((unsigned char *)pack_buffer + send_size);

  // pack data
  for (size_t i = 0; i < tgt_count; ++i) {
    yac_int curr_global_id = tgt_global_ids[i];
    int rank = compute_bucket(curr_global_id, comm_size);
    pack_tgt(
      curr_global_id,
      (void*)((unsigned char*)send_pack_buffer + sdispls[3*rank+0]),
      tgt_pack_size, comm);
    sdispls[3*rank+0] += tgt_pack_size;
  }
  for (size_t i = 0; i < *num_links; ++i) {
    struct link_data * curr_link = (*links) + i;
    int rank = compute_bucket(curr_link->tgt.global_id, comm_size);
    pack_link(
      curr_link,
      (void*)((unsigned char*)send_pack_buffer + sdispls[3*rank+1]),
      link_pack_size, comm);
    sdispls[3*rank+1] += link_pack_size;
  }
  for (size_t i = 0; i < *num_fixed; ++i) {
    double curr_fixed_value = (*fixed)[i].value;
    size_t curr_num_tgt = (*fixed)[i].num_tgt;
    for (size_t j = 0; j < curr_num_tgt; ++j) {
      yac_int curr_global_id = (*fixed)[i].tgt[j].global_id;
      int rank = compute_bucket(curr_global_id, comm_size);
      pack_fixed(
        curr_fixed_value, curr_global_id,
        (void*)((unsigned char*)send_pack_buffer + sdispls[3*rank+2]),
        fixed_pack_size, comm);
      sdispls[3*rank+2] += fixed_pack_size;
    }
  }

  // exchange data
  yac_alltoallv_p2p(
    send_pack_buffer, total_sendcounts, total_sdispls,
    recv_pack_buffer, total_recvcounts, total_rdispls,
    1, MPI_PACKED, comm);

  size_t recv_tgt_count = 0, recv_link_count = 0, recv_fixed_count = 0;
  for (int i = 0; i < comm_size; ++i) {
    recv_tgt_count += (size_t)recvcounts[3*i+0];
    recv_link_count += (size_t)recvcounts[3*i+1];
    recv_fixed_count += (size_t)recvcounts[3*i+2];
  }
  struct link_data * recv_links =
    xrealloc(*links, recv_link_count * sizeof(*recv_links));
  struct temp_fixed_data * recv_fixed =
    xmalloc(recv_fixed_count * sizeof(*recv_fixed));
  struct temp_result * result_buffer =
    xmalloc((recv_tgt_count + tgt_count) * sizeof(*result_buffer));
  struct temp_result * send_results = result_buffer;
  struct temp_result * recv_results = result_buffer + recv_tgt_count;

  // unpack data
  recv_tgt_count = 0, recv_link_count = 0, recv_fixed_count = 0;
  size_t unpack_offset = 0;
  for (int i = 0; i < comm_size; ++i) {
    int curr_num_tgt = recvcounts[3 * i + 0];
    int curr_num_links = recvcounts[3 * i + 1];
    int curr_num_fixed = recvcounts[3 * i + 2];
    for (int j = 0; j < curr_num_tgt; ++j, ++recv_tgt_count) {
      unpack_tgt(
        (void*)((unsigned char *)recv_pack_buffer + unpack_offset),
        tgt_pack_size, &(send_results[recv_tgt_count].tgt.global_id),
        comm);
      unpack_offset += (size_t)tgt_pack_size;
      send_results[recv_tgt_count].owner = i;
      send_results[recv_tgt_count].reorder_idx = recv_tgt_count;
    }
    for (int j = 0; j < curr_num_links; ++j, ++recv_link_count) {
      unpack_link(
        (void*)((unsigned char *)recv_pack_buffer + unpack_offset),
        link_pack_size, recv_links + recv_link_count, comm);
      unpack_offset += (size_t)link_pack_size;
    }
    for (int j = 0; j < curr_num_fixed; ++j, ++recv_fixed_count) {
      unpack_fixed(
        (void*)((unsigned char *)recv_pack_buffer + unpack_offset),
        fixed_pack_size, recv_fixed + recv_fixed_count, comm);
      unpack_offset += (size_t)fixed_pack_size;
    }
  }

  // sort received target point inforatiom by global target id
  qsort(
    send_results, recv_tgt_count, sizeof(*send_results),
    compare_temp_result_tgt);

  // sort links by global target ids
  qsort(
    recv_links, recv_link_count, sizeof(*recv_links), compare_link_data_tgt);

  // sort fixed data by global target ids
  qsort(
    recv_fixed, recv_fixed_count, sizeof(*recv_fixed),
    compare_temp_fixed_data_tgt);

  // match target points with linked and fixed data from the weight file
  for (size_t tgt_idx = 0, link_idx = 0, fixed_idx = 0;
       tgt_idx < recv_tgt_count; ++tgt_idx) {

    yac_int curr_tgt = send_results[tgt_idx].tgt.global_id;

    while ((link_idx < recv_link_count) &&
           (recv_links[link_idx].tgt.global_id < curr_tgt)) ++link_idx;
    while ((fixed_idx < recv_fixed_count) &&
           (recv_fixed[fixed_idx].tgt.global_id < curr_tgt)) ++fixed_idx;

    if ((link_idx < recv_link_count) &&
        (recv_links[link_idx].tgt.global_id == curr_tgt) &&
        (fixed_idx < recv_fixed_count) &&
        (recv_fixed[fixed_idx].tgt.global_id == curr_tgt))
      yac_internal_abort_message(
        "ERROR(redist_weight_file_data): inconsistent data in weight file;"
        "link and fixed data available for a target point",
        __FILE__, __LINE__);
    if ((link_idx < recv_link_count) &&
        (recv_links[link_idx].tgt.global_id == curr_tgt)) {

      size_t temp_link_idx = link_idx;
      while ((temp_link_idx < recv_link_count) &&
             (recv_links[temp_link_idx].tgt.global_id == curr_tgt))
        ++temp_link_idx;
      send_results[tgt_idx].type = LINK;
      send_results[tgt_idx].data.link.links = recv_links + link_idx;
      send_results[tgt_idx].data.link.count = temp_link_idx - link_idx;

    } else if ((fixed_idx < recv_fixed_count) &&
               (recv_fixed[fixed_idx].tgt.global_id == curr_tgt)) {

      send_results[tgt_idx].type = FIXED;
      send_results[tgt_idx].data.fixed.value = recv_fixed[fixed_idx].value;

    } else {
      send_results[tgt_idx].type = NO_INTERP;
    }
  }
  free(recv_fixed);

  // sort received target point inforatiom into origional receive order
  qsort(
    send_results, recv_tgt_count, sizeof(*send_results),
    compare_temp_result_reorder_idx);

  int * pack_size_buffer =
    xmalloc((recv_tgt_count + tgt_count) * sizeof(*pack_size_buffer));
  int * send_pack_sizes = pack_size_buffer;
  int * recv_pack_sizes = pack_size_buffer + recv_tgt_count;
  for (size_t i = 0; i < recv_tgt_count; ++i)
    send_pack_sizes[i] = get_temp_result_pack_size(send_results + i, comm);

  for (int i = 0, saccu = 0, raccu = 0; i < comm_size; ++i) {
    sdispls[i] = saccu;
    rdispls[i] = raccu;
    int sendcount = recvcounts[3*i];
    int recvcount = sendcounts[3*i];
    saccu += (sendcounts[i] = sendcount);
    raccu += (recvcounts[i] = recvcount);
  }

  yac_alltoallv_int_p2p(
    send_pack_sizes, sendcounts, sdispls,
    recv_pack_sizes, recvcounts, rdispls, comm);

  for (int i = 0, k = 0, l = 0, saccu = 0, raccu = 0; i < comm_size; ++i) {
    int sendcount = sendcounts[i];
    int recvcount = recvcounts[i];
    sendcounts[i] = 0;
    recvcounts[i] = 0;
    for (int j = 0; j < sendcount; ++j, ++k)
      sendcounts[i] += send_pack_sizes[k];
    for (int j = 0; j < recvcount; ++j, ++l)
      recvcounts[i] += recv_pack_sizes[l];
    sdispls[i] = saccu;
    rdispls[i] = raccu;
    saccu += sendcounts[i];
    raccu += recvcounts[i];
  }
  size_t total_send_pack_size =
    (size_t)(sdispls[comm_size-1]) + (size_t)(sendcounts[comm_size-1]);
  size_t total_recv_pack_size =
    (size_t)(rdispls[comm_size-1]) + (size_t)(recvcounts[comm_size-1]);

  pack_buffer =
    xrealloc(pack_buffer, total_send_pack_size + total_recv_pack_size);
  send_pack_buffer = pack_buffer;
  recv_pack_buffer =
    (void*)((unsigned char*)pack_buffer + total_send_pack_size);

  for (size_t i = 0, offset = 0; i < recv_tgt_count; ++i) {
    pack_temp_result(
      send_results + i, (void*)((unsigned char*)send_pack_buffer + offset),
      send_pack_sizes[i], comm);
    offset += (size_t)(send_pack_sizes[i]);
  }

  // redistribute link and fixed data to processes requiring this data for
  // their local target points
  yac_alltoallv_p2p(
    send_pack_buffer, sendcounts, sdispls,
    recv_pack_buffer, recvcounts, rdispls,
    1, MPI_PACKED, comm);

  // unpack results
  for (size_t i = 0, offset = 0; i < tgt_count; ++i) {
    unpack_temp_result(
      (void*)((unsigned char*)recv_pack_buffer + offset),
      recv_pack_sizes[i], recv_results + i, comm);
    offset += (size_t)(recv_pack_sizes[i]);
  }
  free(pack_size_buffer);

  // sort received results by global target id
  qsort(
    recv_results, tgt_count, sizeof(*recv_results),
    compare_temp_result_tgt);

  for (size_t i = 0; i < tgt_count; ++i)
    recv_results[i].tgt.local_id = tgt_points[i];

  // sort received results by type
  qsort(
    recv_results, tgt_count, sizeof(*recv_results),
    compare_temp_result_type);

  size_t new_num_links = 0;
  size_t new_total_num_links = 0;
  size_t new_num_fixed = 0;
  {
    size_t i = 0;
    while ((i < tgt_count) && (recv_results[i].type == LINK))
      new_total_num_links += recv_results[i++].data.link.count;
    new_num_links = i;
    while ((i < tgt_count) && (recv_results[i].type == FIXED)) ++i;
    new_num_fixed = i - new_num_links;
  }

  // extract links from results
  *links = realloc(recv_links, new_total_num_links * sizeof(**links));
  for (size_t i = 0, offset = 0; i < new_num_links;
       offset += recv_results[i++].data.link.count)
    memcpy(*links + offset, recv_results[i].data.link.links,
           recv_results[i].data.link.count * sizeof(**links));
  *num_links = new_total_num_links;

  // extract fixed data from results
  if (new_num_fixed > 0) {
    recv_results += new_num_links;
    qsort(
      recv_results, new_num_fixed, sizeof(*recv_results),
      compare_temp_result_fixed);
    double curr_fixed_value =
      (recv_results[0].data.fixed.value == 1337.0)?(-1337.0):(1337.0);
    size_t num_fixed_values = 0;
    void * tgt_global_id_buffer =
      xrealloc((*num_fixed > 0)?(*fixed)->tgt:NULL,
               new_num_fixed * sizeof(*((*fixed)->tgt)));
    struct fixed_data * curr_fixed = NULL;
    for (size_t i = 0, offset = 0; i < new_num_fixed; ++i) {
      if (recv_results[i].data.fixed.value != curr_fixed_value) {
        curr_fixed_value = recv_results[i].data.fixed.value;
        *fixed = xrealloc(*fixed, (num_fixed_values + 1) * sizeof(**fixed));
        curr_fixed = *fixed + num_fixed_values;
        ++num_fixed_values;
        curr_fixed->value = curr_fixed_value;
        curr_fixed->tgt = (void*)(((unsigned char *)tgt_global_id_buffer) +
                                  offset * sizeof(*((*fixed)->tgt)));
        curr_fixed->num_tgt = 0;
      }
      curr_fixed->tgt[curr_fixed->num_tgt++].global_id =
        recv_results[i].tgt.global_id;
    }
    *num_fixed = num_fixed_values;
    recv_results -= new_num_links;

  } else {
    if (*num_fixed > 0) free((*fixed)->tgt);
    free(*fixed);
    *fixed = NULL;
    *num_fixed = 0;
  }

  for (size_t i = 0; i < tgt_count; ++i)
    tgt_points[i] = recv_results[i].tgt.local_id;
  *num_interpolated_points = new_num_links + new_num_fixed;

  for (size_t i = 0; i < new_num_links; ++i)
    free(recv_results[i].data.link.links);
  free(result_buffer);
  free(pack_buffer);
  yac_free_comm_buffers(sendcounts, recvcounts, sdispls, rdispls);
  free(int_buffer);
  free(tgt_global_ids);
}

static size_t do_search_file(struct interp_method * method,
                             struct interp_grid * interp_grid,
                             size_t * tgt_points, size_t count,
                             struct interp_weights * weights) {

  struct interp_method_file * method_file =
    (struct interp_method_file*)(method);

  MPI_Comm comm = yac_interp_grid_get_MPI_Comm(interp_grid);

  size_t num_src_fields = yac_interp_grid_get_num_src_fields(interp_grid);
  enum yac_location * src_locations =
    xmalloc(num_src_fields * sizeof(*src_locations));
  for (size_t i = 0; i < num_src_fields; ++i)
    src_locations[i] = yac_interp_grid_get_src_field_location(interp_grid, i);
  enum yac_location tgt_location =
    yac_interp_grid_get_tgt_field_location(interp_grid);

  struct link_data * links;
  size_t num_links;
  struct fixed_data * fixed;
  size_t num_fixed;

  // read data from file
  read_weight_file(
    method_file->weight_file_name, method_file->io_proc_step, comm,
    method_file->src_grid_name, method_file->tgt_grid_name,
    src_locations, tgt_location, num_src_fields,
    &links, &num_links, &fixed, &num_fixed);

  free(src_locations);

  size_t num_interpolated_points;
  redist_weight_file_data(
    interp_grid, tgt_points, count, &num_interpolated_points,
    &links, &num_links, &fixed, &num_fixed);

  size_t num_fixed_tgt = 0;
  for (size_t i = 0; i < num_fixed; ++i) num_fixed_tgt += fixed[i].num_tgt;

  if (num_links > 0) {

    size_t num_link_tgt = num_interpolated_points - num_fixed_tgt;
    yac_int * src_global_ids = xmalloc(num_links * sizeof(*src_global_ids));

    // get global source ids
    for (size_t i = 0; i < num_links; ++i)
      src_global_ids[i] = links[i].src.global_id;

    size_t * src_local_ids = xmalloc(num_links * sizeof(*src_local_ids));

    // get local ids for all source points
    yac_intep_grid_src_global_to_local(
      interp_grid, 0, src_global_ids, num_links, src_local_ids);

    const_int_pointer src_mask =
      yac_interp_grid_get_src_field_mask(interp_grid, 0);

    size_t * num_weights_per_tgt =
      xcalloc(num_link_tgt, sizeof(*num_weights_per_tgt));
    double * w = xmalloc(num_links * sizeof(*w));
    size_t * tgt_reorder_idx = xmalloc(count * sizeof(*tgt_reorder_idx));
    size_t masked_count = 0;

    for (size_t i = 0; i < count; ++i) tgt_reorder_idx[i] = i;

    size_t new_num_link_tgt = 0;
    size_t new_num_links = 0;
    for (size_t i = 0, link_idx = 0; i < num_link_tgt; ++i) {

      yac_int curr_tgt_global_id = links[link_idx].tgt.global_id;
      int mask_flag = 1;

      // check the source mask
      size_t temp_link_idx = link_idx;
      while ((temp_link_idx < num_links) &&
             (links[temp_link_idx].tgt.global_id == curr_tgt_global_id)) {

        if (links[temp_link_idx].src_field_idx != 0)
          yac_internal_abort_message(
            "ERROR(do_search_file): src_field_idx != 0 not yet supported",
            __FILE__, __LINE__);

        if (src_mask != NULL)
          mask_flag &= src_mask[src_local_ids[temp_link_idx]];
        ++temp_link_idx;
      }

      // if any of the source points for the current target point is masked
      if (!mask_flag) {
        tgt_reorder_idx[i] = count + masked_count;
        link_idx = temp_link_idx;
        ++masked_count;
        continue;
      }

      num_weights_per_tgt[new_num_link_tgt] = temp_link_idx - link_idx;
      for (; link_idx < temp_link_idx; ++link_idx, ++new_num_links) {
        src_global_ids[new_num_links] = src_global_ids[link_idx];
        src_local_ids[new_num_links] = src_local_ids[link_idx];
        w[new_num_links] = links[link_idx].weight;
      }

      ++new_num_link_tgt;
    }

    if (masked_count > 0) {
      yac_quicksort_index_size_t_size_t(tgt_reorder_idx, count, tgt_points);
      num_interpolated_points -= masked_count;
      num_links = new_num_links;
      num_link_tgt = new_num_link_tgt;
    }
    free(tgt_reorder_idx);

    struct remote_point * srcs =
      yac_interp_grid_get_src_remote_points(
        interp_grid, 0, src_local_ids, num_links);

    struct remote_points tgts = {
      .data =
        yac_interp_grid_get_tgt_remote_points(
          interp_grid, tgt_points, num_link_tgt),
      .count = num_link_tgt};

    // store weights
    yac_interp_weights_add_wsum(
      weights, &tgts, num_weights_per_tgt, srcs, w);

    free(tgts.data);
    free(srcs);
    free(w);
    free(num_weights_per_tgt);
    free(src_local_ids);
    free(src_global_ids);
  } else {
    // get local ids for all source points
    yac_intep_grid_src_global_to_local(
      interp_grid, 0, NULL, 0, NULL);
  }
  free(links);

  if (num_fixed) {

    size_t tgt_points_offset = num_interpolated_points - num_fixed_tgt;

    for (size_t i = 0; i < num_fixed; ++i) {

      struct remote_points tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, tgt_points + tgt_points_offset, fixed[i].num_tgt),
        .count = fixed[i].num_tgt};

      yac_interp_weights_add_fixed(weights, &tgts, fixed[i].value);
      free(tgts.data);
      tgt_points_offset += fixed[i].num_tgt;
    }
    free(fixed->tgt);
  }
  free(fixed);

  return num_interpolated_points;
}

struct interp_method * yac_interp_method_file_new(
  char const * weight_file_name, int io_proc_step,
  char const * src_grid_name, char const * tgt_grid_name) {

  struct interp_method_file * method = xmalloc(1 * sizeof(*method));

  method->vtable = &interp_method_file_vtable;
  method->io_proc_step = io_proc_step;
  method->weight_file_name = xmalloc(strlen(weight_file_name) + 1);
  strcpy((char*)(method->weight_file_name), weight_file_name);
  method->src_grid_name = xmalloc(strlen(src_grid_name) + 1);
  strcpy((char*)(method->src_grid_name), src_grid_name);
  method->tgt_grid_name = xmalloc(strlen(tgt_grid_name) + 1);
  strcpy((char*)(method->tgt_grid_name), tgt_grid_name);


  return (struct interp_method*)method;
}

static void delete_file(struct interp_method * method) {

  struct interp_method_file * method_file =
    (struct interp_method_file *)method;
  free((void*)(method_file->weight_file_name));
  free((void*)(method_file->src_grid_name));
  free((void*)(method_file->tgt_grid_name));
  free(method_file);
}
