/**
 * @file interpolation.c
 *
 * @copyright Copyright  (C)  2019 Moritz Hanke <hanke@dkrz.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
// Get the definition of the 'restrict' keyword.
#include "config.h"
#endif

#include <string.h>

#include "utils.h"
#include "interpolation.h"
#include "yac_mpi.h"

struct t_rsh_wsb_rrf {
  Xt_redist redist_halo;
  size_t * halo_sizes;
  size_t tgt_count;
  double ** result_buffer;
  size_t * num_src_per_tgt;
  double * weights;
  size_t * src_field_idx;
  size_t * src_idx;
  size_t num_src_fields;
  double ** src_fields_buffer;
  Xt_redist redist_results;
  Xt_request request_results;
  int is_source;
  int is_target;
};

struct t_rsh_wsf {
  Xt_redist redist_src;
  Xt_request request_send_src;
  double ** src_send_buffer;
  double ** src_fields_buffer;
  size_t * src_send_buffer_sizes;
  size_t * src_recv_buffer_sizes;
  size_t * tgt_pos;
  size_t tgt_count;
  size_t * num_src_per_tgt;
  double * weights;
  size_t * src_field_idx;
  size_t * src_idx;
  size_t num_src_fields;
  int is_source;
  int is_target;
};

struct t_direct {
  Xt_redist redist;
  double ** src_field_buffer;
  double ** field_dummy;
  int is_source;
  int is_target;
};

struct t_fixed {
  double value;
  size_t * pos;
  size_t count;
};

struct interpolation {

  struct t_rsh_wsb_rrf * rsh_wsb_rrf; // with weights
  struct t_rsh_wsb_rrf * rsh_sb_rrf;  // without weights (all weights == 1.0)

  struct t_rsh_wsf * rsh_wsf; // with weights
  struct t_rsh_wsf * rsh_sf;  // without weights (all weights == 1.0)

  struct t_direct * direct;

  struct t_fixed * fixed;
  size_t fixed_count;

  int is_source, is_target;
  size_t collection_size;
};

struct interpolation * yac_interpolation_new(size_t collection_size) {

  struct interpolation * interp = xmalloc(1 * sizeof(*interp));

  interp->rsh_wsb_rrf = NULL;
  interp->rsh_sb_rrf = NULL;
  interp->rsh_wsf = NULL;
  interp->rsh_sf = NULL;
  interp->direct = NULL;
  interp->fixed = NULL;
  interp->fixed_count = 0;

  interp->is_source = 0;
  interp->is_target = 0;
  interp->collection_size = collection_size;

  return interp;
}

static struct t_fixed generate_fixed(
  double value, size_t count, size_t * pos) {

  struct t_fixed fixed;

  fixed.value = value;
  fixed.count = count;
  fixed.pos = xmalloc(count * sizeof(*fixed.pos));
  memcpy(fixed.pos, pos, count * sizeof(*pos));

  return fixed;
}

void yac_interpolation_add_fixed(
  struct interpolation * interp, double value, size_t count, size_t * pos) {

  interp->is_target |= (count > 0);
  interp->fixed =
    xrealloc(
      interp->fixed, (interp->fixed_count + 1) * sizeof(*(interp->fixed)));
  interp->fixed[interp->fixed_count] = generate_fixed(value, count, pos);
  interp->fixed_count++;
}

static void interpolation_direct_init_buffers(
  struct t_direct * direct, size_t collection_size) {

  if (direct->is_source)
    direct->src_field_buffer =
      xmalloc(collection_size * sizeof(*(direct->src_field_buffer)));
  else
    direct->src_field_buffer = NULL;
  direct->field_dummy =
    xmalloc(collection_size * sizeof(*(direct->field_dummy)));
  for (size_t i = 0; i < collection_size; ++i)
    direct->field_dummy[i] = NULL;
}

void yac_interpolation_add_direct(
  struct interpolation * interp, Xt_redist redist_) {

  struct t_direct * direct = xmalloc(1 * sizeof(*direct));

  Xt_redist redist;
  size_t collection_size = (size_t)(interp->collection_size);

  if (collection_size > 1) {

    Xt_redist * redist_buffer =
      xmalloc(collection_size  * sizeof(*redist_buffer));

    for (size_t i = 0; i < collection_size; ++i)
      redist_buffer[i] = redist_;
    redist =
      xt_redist_collection_new(
        redist_buffer, (int)(collection_size), -1,
        xt_redist_get_MPI_Comm(redist_));
    xt_redist_delete(redist_);
    free(redist_buffer);

  } else {
    redist = redist_;
  }

  direct->is_source =
    (redist != NULL) && (xt_redist_get_num_send_msg(redist) > 0);
  direct->is_target =
    (redist != NULL) && (xt_redist_get_num_recv_msg(redist) > 0);
  direct->redist = redist;
  interpolation_direct_init_buffers(direct, collection_size);

  interp->is_source |= direct->is_source;
  interp->is_target |= direct->is_target;
  interp->direct = direct;
}

size_t * get_buffer_sizes(
  Xt_redist * redist, size_t num_src_fields,
  size_t (*xt_redist_get_buffer_size)(Xt_redist)) {

  size_t * buffer_sizes;
  if (redist == NULL) {
    buffer_sizes = xcalloc(num_src_fields, sizeof(*buffer_sizes));
  } else {
    buffer_sizes = xmalloc(num_src_fields * sizeof(*buffer_sizes));
    for (size_t i = 0; i < num_src_fields; ++i)
      buffer_sizes[i] = xt_redist_get_buffer_size(redist[i]);
  }
  return buffer_sizes;
}

double ** setup_src_buffer(
  size_t * src_buffer_sizes, size_t num_src_fields, size_t collection_size) {

  size_t src_buffer_size = 0;
  for (int i = 0; i < num_src_fields; ++i)
    src_buffer_size += src_buffer_sizes[i];
  double ** src_buffer =
    xmalloc(collection_size * num_src_fields * sizeof(*src_buffer));
  src_buffer[0] = xmalloc(collection_size * src_buffer_size);
  for (size_t i = 0, offset = 0; i < collection_size; ++i) {
    for (size_t j = 0; j < num_src_fields; offset += src_buffer_sizes[j++]) {
      src_buffer[i * num_src_fields + j] =
        (double*)((char*)(src_buffer[0]) + offset);
    }
  }
  return src_buffer;
}

size_t xt_redist_get_buffer_size(
  Xt_redist redist, MPI_Datatype(*xt_redist_get_MPI_Datatype)(Xt_redist, int)) {

  if (redist == 0) return 0;

  MPI_Comm comm = xt_redist_get_MPI_Comm(redist);
  int comm_size;
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  size_t max_size = 0;

  for (int i = 0; i < comm_size; ++i) {

    MPI_Datatype dt = xt_redist_get_MPI_Datatype(redist, i);
    if (dt == MPI_DATATYPE_NULL) continue;
    MPI_Aint lb, extent;
    yac_mpi_call(MPI_Type_get_extent(dt, &lb, &extent), comm);
    size_t curr_size = (size_t)extent + (size_t)lb;
    if (curr_size > max_size) max_size = curr_size;
    yac_mpi_call(MPI_Type_free(&dt), comm);
  }
  return max_size;
}

size_t xt_redist_get_send_buffer_size(Xt_redist redist) {

  return xt_redist_get_buffer_size(redist, xt_redist_get_send_MPI_Datatype);
}

size_t xt_redist_get_recv_buffer_size(Xt_redist redist) {

  return xt_redist_get_buffer_size(redist, xt_redist_get_recv_MPI_Datatype);
}

static struct t_rsh_wsb_rrf * generate_sum_rsh_wsb_rrf(
  Xt_redist redist_halo, size_t * halo_sizes, size_t tgt_count,
  size_t * num_src_per_tgt, double * weights, size_t * src_field_idx,
  size_t * src_idx, size_t num_src_fields, Xt_redist redist_results,
  size_t collection_size, int is_source, int is_target) {

  size_t total_num_src = 0;
  for (size_t i = 0; i < tgt_count; ++i) total_num_src += num_src_per_tgt[i];

  struct t_rsh_wsb_rrf * rsh_wsb_rrf = xmalloc(1 * sizeof(*(rsh_wsb_rrf)));
  rsh_wsb_rrf->request_results = XT_REQUEST_NULL;
  rsh_wsb_rrf->redist_halo = (redist_halo == NULL)?NULL:xt_redist_copy(redist_halo);
  rsh_wsb_rrf->halo_sizes =
    xmalloc(num_src_fields * sizeof(*halo_sizes));
  memcpy(rsh_wsb_rrf->halo_sizes, halo_sizes,
         num_src_fields * sizeof(*halo_sizes));
  rsh_wsb_rrf->tgt_count = tgt_count;
  size_t result_buffer_size = tgt_count * sizeof(*(rsh_wsb_rrf->result_buffer));
  rsh_wsb_rrf->result_buffer =
    setup_src_buffer(&result_buffer_size, 1, collection_size);
  rsh_wsb_rrf->num_src_per_tgt =
    xmalloc(tgt_count * sizeof(*num_src_per_tgt));
  memcpy(rsh_wsb_rrf->num_src_per_tgt, num_src_per_tgt,
         tgt_count * sizeof(*num_src_per_tgt));
  if (weights != NULL) {
    rsh_wsb_rrf->weights = xmalloc(total_num_src * sizeof(*weights));
    memcpy(rsh_wsb_rrf->weights, weights, total_num_src * sizeof(*weights));
  } else {
    rsh_wsb_rrf->weights = NULL;
  }
  rsh_wsb_rrf->src_field_idx =
    xmalloc(total_num_src * sizeof(*src_field_idx));
  memcpy(rsh_wsb_rrf->src_field_idx, src_field_idx,
         total_num_src * sizeof(*src_field_idx));
  rsh_wsb_rrf->src_idx =
    xmalloc(total_num_src * sizeof(*src_idx));
  memcpy(rsh_wsb_rrf->src_idx, src_idx,
         total_num_src * sizeof(*src_idx));
  rsh_wsb_rrf->num_src_fields = num_src_fields;
  rsh_wsb_rrf->src_fields_buffer =
    xrealloc(
      setup_src_buffer(
        halo_sizes, num_src_fields, collection_size),
        2 * collection_size * num_src_fields *
        sizeof(*(rsh_wsb_rrf->src_fields_buffer)));
  rsh_wsb_rrf->redist_results =
    (redist_results == NULL)?NULL:xt_redist_copy(redist_results);
  rsh_wsb_rrf->is_source = is_source;
  rsh_wsb_rrf->is_target = is_target;

  return rsh_wsb_rrf;
}

void yac_interpolation_add_weight_sum_rsh_w_sb_rrf(
  struct interpolation * interp, Xt_redist * redist_halo_,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist redist_results_,
  int is_source, int is_target, int with_weights) {

  Xt_redist redist_halo;
  Xt_redist redist_results;
  size_t collection_size = (size_t)(interp->collection_size);

  if (collection_size > 1) {

    int redist_halo_is_null = *redist_halo_ == NULL;
    for (size_t i = 1; i < num_src_fields; ++i)
      if (redist_halo_is_null != (redist_halo_[i] == NULL))
        yac_internal_abort_message(
          "ERROR(yac_interpolation_add_weight_sum_rsh_wsb_rrf): "
          "invalid argument", __FILE__, __LINE__);

    Xt_redist * redist_buffer =
      xmalloc(collection_size * num_src_fields * sizeof(*redist_buffer));

    if (redist_halo_is_null) {
      redist_halo = NULL;
    } else {
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          redist_buffer[i * num_src_fields + j] = redist_halo_[j];
      redist_halo =
        xt_redist_collection_new(
          redist_buffer, (int)(num_src_fields * collection_size), -1,
          xt_redist_get_MPI_Comm(*redist_halo_));
    }

    if (redist_results_ == NULL) {
      redist_results = NULL;
    } else {
      for (size_t i = 0; i < collection_size; ++i)
        redist_buffer[i] = redist_results_;
      redist_results =
        xt_redist_collection_new(
          redist_buffer, (int)collection_size, -1,
          xt_redist_get_MPI_Comm(redist_results_));

      free(redist_buffer);
    }

  } else {
    redist_halo = (redist_halo_ != NULL)?(*redist_halo_):(NULL);
    redist_results = (redist_results_ != NULL)?(redist_results_):(NULL);
  }

  size_t * halo_sizes =
    get_buffer_sizes(
      redist_halo_, num_src_fields, xt_redist_get_recv_buffer_size);

  if (with_weights) {
    interp->rsh_wsb_rrf =
      generate_sum_rsh_wsb_rrf(
        redist_halo, halo_sizes, tgt_count, num_src_per_tgt, weights,
        src_field_idx, src_idx, num_src_fields, redist_results, collection_size,
        is_source, is_target);
  } else {
    interp->rsh_sb_rrf =
      generate_sum_rsh_wsb_rrf(
        redist_halo, halo_sizes, tgt_count, num_src_per_tgt, NULL,
        src_field_idx, src_idx, num_src_fields, redist_results, collection_size,
        is_source, is_target);
  }
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  free(halo_sizes);

  if (collection_size > 1) {

    if (redist_halo != NULL) xt_redist_delete(redist_halo);
    if (redist_results != NULL) xt_redist_delete(redist_results);
  }
}

static struct t_rsh_wsf * generate_sum_rsh_wsf(
  Xt_redist redist_src, size_t * src_send_buffer_sizes,
  size_t * src_recv_buffer_sizes, size_t * tgt_pos, size_t tgt_count,
  size_t * num_src_per_tgt, double * weights, size_t * src_field_idx,
  size_t * src_idx, size_t num_src_fields, size_t collection_size,
  int is_source, int is_target) {

  size_t total_num_src = 0;
  for (size_t i = 0; i < tgt_count; ++i) total_num_src += num_src_per_tgt[i];

  struct t_rsh_wsf * rsh_wsf = xmalloc(1 * sizeof(*(rsh_wsf)));
  rsh_wsf->redist_src = (redist_src == NULL)?NULL:xt_redist_copy(redist_src);
  rsh_wsf->request_send_src = XT_REQUEST_NULL;
  rsh_wsf->src_send_buffer =
    setup_src_buffer(src_send_buffer_sizes, num_src_fields, collection_size);
  rsh_wsf->src_send_buffer_sizes =
    xmalloc(num_src_fields * sizeof(*src_send_buffer_sizes));
  memcpy(rsh_wsf->src_send_buffer_sizes, src_send_buffer_sizes,
         num_src_fields * sizeof(*src_send_buffer_sizes));
  rsh_wsf->src_fields_buffer =
    xrealloc(
      setup_src_buffer(
        src_recv_buffer_sizes, num_src_fields, collection_size),
        2 * collection_size * num_src_fields *
        sizeof(*(rsh_wsf->src_fields_buffer)));
  rsh_wsf->src_recv_buffer_sizes =
    xmalloc(num_src_fields * sizeof(*src_recv_buffer_sizes));
  memcpy(rsh_wsf->src_recv_buffer_sizes, src_recv_buffer_sizes,
         num_src_fields * sizeof(*src_recv_buffer_sizes));
  rsh_wsf->tgt_pos = xmalloc(tgt_count * sizeof(*tgt_pos));
  memcpy(rsh_wsf->tgt_pos, tgt_pos, tgt_count * sizeof(*tgt_pos));
  rsh_wsf->tgt_count = tgt_count;
  rsh_wsf->num_src_per_tgt =
    xmalloc(tgt_count * sizeof(*num_src_per_tgt));
  memcpy(rsh_wsf->num_src_per_tgt, num_src_per_tgt,
         tgt_count * sizeof(*num_src_per_tgt));
  if (weights != NULL) {
    rsh_wsf->weights = xmalloc(total_num_src * sizeof(*weights));
    memcpy(rsh_wsf->weights, weights, total_num_src * sizeof(*weights));
  } else {
    rsh_wsf->weights = NULL;
  }
  rsh_wsf->src_field_idx =
    xmalloc(total_num_src * sizeof(*src_field_idx));
  memcpy(rsh_wsf->src_field_idx, src_field_idx,
         total_num_src * sizeof(*src_field_idx));
  rsh_wsf->src_idx =
    xmalloc(total_num_src * sizeof(*src_idx));
  memcpy(rsh_wsf->src_idx, src_idx,
         total_num_src * sizeof(*src_idx));
  rsh_wsf->num_src_fields = num_src_fields;
  rsh_wsf->is_source = is_source;
  rsh_wsf->is_target = is_target;

  return rsh_wsf;
}

void yac_interpolation_add_weight_sum_rsh_w_sf(
  struct interpolation * interp, Xt_redist * redist_src_, size_t * tgt_pos,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, int with_weights) {

  Xt_redist redist_src;
  size_t collection_size = (size_t)(interp->collection_size);

  if (collection_size > 1) {

    int redist_src_is_null = *redist_src_ == NULL;
    for (size_t i = 1; i < num_src_fields; ++i)
      if (redist_src_is_null != (redist_src_[i] == NULL))
        yac_internal_abort_message(
          "ERROR(yac_interpolation_add_weight_sum_rsh_wsb_rrf): "
          "invalid argument", __FILE__, __LINE__);


    if (redist_src_is_null) {
      redist_src = NULL;
    } else {
      Xt_redist * redist_buffer =
        xmalloc(collection_size * num_src_fields * sizeof(*redist_buffer));

      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          redist_buffer[i * num_src_fields + j] = redist_src_[j];
      redist_src =
        xt_redist_collection_new(
          redist_buffer, (int)(num_src_fields * collection_size), -1,
          xt_redist_get_MPI_Comm(*redist_src_));

      free(redist_buffer);
    }

  } else {
    redist_src = (redist_src_ != NULL)?(*redist_src_):(NULL);
  }

  int is_source =
    (redist_src != NULL) && (xt_redist_get_num_send_msg(redist_src) > 0);
  int is_target = tgt_count > 0;

  size_t * src_send_buffer_sizes =
    get_buffer_sizes(
      redist_src_, num_src_fields, xt_redist_get_send_buffer_size);
  size_t * src_recv_buffer_sizes =
    get_buffer_sizes(
      redist_src_, num_src_fields, xt_redist_get_recv_buffer_size);

  if (with_weights) {
    interp->rsh_wsf =
      generate_sum_rsh_wsf(
        redist_src, src_send_buffer_sizes, src_recv_buffer_sizes,
        tgt_pos, tgt_count, num_src_per_tgt, weights, src_field_idx, src_idx,
        num_src_fields, collection_size, is_source, is_target);
  } else {
    interp->rsh_sf =
      generate_sum_rsh_wsf(
        redist_src, src_send_buffer_sizes, src_recv_buffer_sizes,
        tgt_pos, tgt_count, num_src_per_tgt, NULL, src_field_idx, src_idx,
        num_src_fields, collection_size, is_source, is_target);
  }
  interp->is_source |= is_source;
  interp->is_target |= is_target;

  free(src_recv_buffer_sizes);
  free(src_send_buffer_sizes);

  if ((collection_size > 1) && (redist_src != NULL))
    xt_redist_delete(redist_src);
}

void yac_interpolation_add_sum_rsh_sb_rrf(
  struct interpolation * interp, Xt_redist * redist_halo,
  size_t tgt_count, size_t * num_src_per_tgt,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist redist_results,
  int is_source, int is_target) {

  yac_interpolation_add_weight_sum_rsh_w_sb_rrf(
    interp, redist_halo, tgt_count, num_src_per_tgt, NULL, src_field_idx,
    src_idx, num_src_fields, redist_results, is_source, is_target, 0);
}

void yac_interpolation_add_sum_rsh_sf(
  struct interpolation * interp, Xt_redist * redist_src,
  size_t * tgt_pos, size_t tgt_count, size_t * num_src_per_tgt,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields) {

  yac_interpolation_add_weight_sum_rsh_w_sf(
    interp, redist_src, tgt_pos, tgt_count, num_src_per_tgt, NULL,
    src_field_idx, src_idx, num_src_fields, 0);
}

void yac_interpolation_add_weight_sum_rsh_wsb_rrf(
  struct interpolation * interp, Xt_redist * redist_halo,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields, Xt_redist redist_results,
  int is_source, int is_target) {

  yac_interpolation_add_weight_sum_rsh_w_sb_rrf(
    interp, redist_halo, tgt_count, num_src_per_tgt, weights, src_field_idx,
    src_idx, num_src_fields, redist_results, is_source, is_target, 1);
};

void yac_interpolation_add_weight_sum_rsh_wsf(
  struct interpolation * interp, Xt_redist * redist_src, size_t * tgt_pos,
  size_t tgt_count, size_t * num_src_per_tgt, double * weights,
  size_t * src_field_idx, size_t * src_idx,
  size_t num_src_fields) {

  yac_interpolation_add_weight_sum_rsh_w_sf(
    interp, redist_src, tgt_pos, tgt_count, num_src_per_tgt, weights,
    src_field_idx, src_idx, num_src_fields, 1);
}

struct interpolation * yac_interpolation_copy(struct interpolation * interp) {

  struct interpolation * interp_copy =
    yac_interpolation_new(interp->collection_size);

  if (interp->rsh_wsb_rrf != NULL) {
    interp_copy->rsh_wsb_rrf =
      generate_sum_rsh_wsb_rrf(
        interp->rsh_wsb_rrf->redist_halo, interp->rsh_wsb_rrf->halo_sizes,
        interp->rsh_wsb_rrf->tgt_count, interp->rsh_wsb_rrf->num_src_per_tgt,
        interp->rsh_wsb_rrf->weights, interp->rsh_wsb_rrf->src_field_idx,
        interp->rsh_wsb_rrf->src_idx, interp->rsh_wsb_rrf->num_src_fields,
        interp->rsh_wsb_rrf->redist_results, interp->collection_size,
        interp->rsh_wsb_rrf->is_source, interp->rsh_wsb_rrf->is_target);
    interp_copy->is_source |= interp->rsh_wsb_rrf->is_source;
    interp_copy->is_target |= interp->rsh_wsb_rrf->is_target;
  }

  if (interp->rsh_wsf != NULL) {
    interp_copy->rsh_wsf =
      generate_sum_rsh_wsf(
        interp->rsh_wsf->redist_src, interp->rsh_wsf->src_send_buffer_sizes,
        interp->rsh_wsf->src_recv_buffer_sizes, interp->rsh_wsf->tgt_pos,
        interp->rsh_wsf->tgt_count, interp->rsh_wsf->num_src_per_tgt,
        interp->rsh_wsf->weights, interp->rsh_wsf->src_field_idx,
        interp->rsh_wsf->src_idx, interp->rsh_wsf->num_src_fields,
        interp->collection_size,
        interp->rsh_wsf->is_source, interp->rsh_wsf->is_target);
    interp_copy->is_source |= interp->rsh_wsf->is_source;
    interp_copy->is_target |= interp->rsh_wsf->is_target;
  }

  if (interp->rsh_sb_rrf != NULL) {
    interp_copy->rsh_sb_rrf =
      generate_sum_rsh_wsb_rrf(
        interp->rsh_sb_rrf->redist_halo, interp->rsh_sb_rrf->halo_sizes,
        interp->rsh_sb_rrf->tgt_count, interp->rsh_sb_rrf->num_src_per_tgt,
        interp->rsh_sb_rrf->weights, interp->rsh_sb_rrf->src_field_idx,
        interp->rsh_sb_rrf->src_idx, interp->rsh_sb_rrf->num_src_fields,
        interp->rsh_sb_rrf->redist_results, interp->collection_size,
        interp->rsh_sb_rrf->is_source, interp->rsh_sb_rrf->is_target);
    interp_copy->is_source |= interp->rsh_sb_rrf->is_source;
    interp_copy->is_target |= interp->rsh_sb_rrf->is_target;
  }

  if (interp->rsh_sf != NULL) {
    interp_copy->rsh_sf =
      generate_sum_rsh_wsf(
        interp->rsh_sf->redist_src, interp->rsh_sf->src_send_buffer_sizes,
        interp->rsh_sf->src_recv_buffer_sizes, interp->rsh_sf->tgt_pos,
        interp->rsh_sf->tgt_count, interp->rsh_sf->num_src_per_tgt,
        interp->rsh_sf->weights, interp->rsh_sf->src_field_idx,
        interp->rsh_sf->src_idx, interp->rsh_sf->num_src_fields,
        interp->collection_size,
        interp->rsh_sf->is_source, interp->rsh_sf->is_target);
    interp_copy->is_source |= interp->rsh_sf->is_source;
    interp_copy->is_target |= interp->rsh_sf->is_target;
  }

  if (interp->fixed != NULL) {
    interp_copy->fixed =
      xmalloc(interp->fixed_count * sizeof(*(interp_copy->fixed)));
    interp_copy->fixed_count = interp->fixed_count;
    for (size_t i = 0; i < interp->fixed_count; ++i) {
      interp_copy->fixed[i] =
        generate_fixed(
          interp->fixed[i].value, interp->fixed[i].count, interp->fixed[i].pos);
      interp_copy->is_target |= (interp->fixed[i].count > 0);
    }
  }

  if (interp->direct != NULL) {
    interp_copy->direct = xmalloc(1 * sizeof(*(interp_copy->direct)));
    interp_copy->direct->is_source = interp->direct->is_source;
    interp_copy->direct->is_target = interp->direct->is_target;
    interp_copy->direct->redist =
      (interp->direct->redist == NULL)?
        NULL:xt_redist_copy(interp->direct->redist);
    interpolation_direct_init_buffers(
      interp_copy->direct, interp->collection_size);
    interp_copy->is_source |= interp->direct->is_source;
    interp_copy->is_target |= interp->direct->is_target;
  }

  return interp_copy;
}

// src_fields dimensions [collection_idx]
//                       [field index]
//                       [local_idx]
// tgt_field dimensions [collection_idx]
//                      [local_idx]
void yac_interpolation_execute(
  struct interpolation * interp, double *** src_fields, double ** tgt_field) {

  size_t collection_size = (size_t)(interp->collection_size);

  if (interp->rsh_wsb_rrf != NULL) {

    double ** results = interp->rsh_wsb_rrf->result_buffer;

    if (interp->rsh_wsb_rrf->is_source) {

      size_t num_src_fields = interp->rsh_wsb_rrf->num_src_fields;
      size_t send_offset = num_src_fields * collection_size;
      double ** temp_src_fields = interp->rsh_wsb_rrf->src_fields_buffer;
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] =
            src_fields[i][j];

      // do halo exchange
      if (interp->rsh_wsb_rrf->redist_halo != NULL)
        xt_redist_s_exchange(
          interp->rsh_wsb_rrf->redist_halo,
          (int)(collection_size * num_src_fields),
          (void const**)(temp_src_fields + send_offset), (void**)temp_src_fields);

      size_t tgt_count = interp->rsh_wsb_rrf->tgt_count;
      size_t * num_src_per_tgt = interp->rsh_wsb_rrf->num_src_per_tgt;
      double * weights = interp->rsh_wsb_rrf->weights;
      size_t * src_field_idx = interp->rsh_wsb_rrf->src_field_idx;
      size_t * src_idx = interp->rsh_wsb_rrf->src_idx;
      size_t local_data_offset = send_offset - num_src_fields;

      // apply weights
      for (size_t l = 0; l < collection_size; ++l) {
        for (size_t i = 0, k = 0; i < tgt_count; ++i) {
          size_t curr_num_src_per_tgt = num_src_per_tgt[i];
          double result = 0.0;
          for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
            result +=
              temp_src_fields
                [l * num_src_fields + src_field_idx[k] +
                 ((src_field_idx[k] >= num_src_fields)?(local_data_offset):(0))]
                [src_idx[k]] * weights[k];
          results[l][i] = result;
        }
      }
    }

    double * dummy = NULL;
    if (!interp->rsh_wsb_rrf->is_target) tgt_field = &dummy;

    // redistribute results
    if (interp->rsh_wsb_rrf->redist_results != NULL)
      xt_redist_s_exchange(
        interp->rsh_wsb_rrf->redist_results, interp->collection_size,
        (const void **)results, (void **)tgt_field);
  }

  if (interp->rsh_wsf != NULL) {

    size_t num_src_fields = interp->rsh_wsf->num_src_fields;
      size_t send_offset = num_src_fields * collection_size;
    double ** temp_src_fields = interp->rsh_wsf->src_fields_buffer;
    if (interp->rsh_wsf->is_source) {
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] =
            src_fields[i][j];
    } else {
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] = NULL;
    }

    // do halo exchange
    if (interp->rsh_wsf->redist_src != NULL)
      xt_redist_s_exchange(
        interp->rsh_wsf->redist_src, (int)(num_src_fields * collection_size),
        (void const**)(temp_src_fields + send_offset), (void**)temp_src_fields);

    size_t * tgt_pos = interp->rsh_wsf->tgt_pos;
    size_t tgt_count = interp->rsh_wsf->tgt_count;
    size_t * num_src_per_tgt = interp->rsh_wsf->num_src_per_tgt;
    double * weights = interp->rsh_wsf->weights;
    size_t * src_field_idx = interp->rsh_wsf->src_field_idx;
    size_t * src_idx = interp->rsh_wsf->src_idx;

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
          result +=
            temp_src_fields[l * num_src_fields + src_field_idx[k]]
                           [src_idx[k]] * weights[k];
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  }

  if (interp->rsh_sb_rrf != NULL) {

    double ** results = interp->rsh_sb_rrf->result_buffer;

    if (interp->rsh_sb_rrf->is_source) {

      size_t num_src_fields = interp->rsh_sb_rrf->num_src_fields;
      size_t send_offset = num_src_fields * collection_size;
      double ** temp_src_fields = interp->rsh_sb_rrf->src_fields_buffer;
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] =
            src_fields[i][j];

      // do halo exchange
      if (interp->rsh_sb_rrf->redist_halo != NULL)
        xt_redist_s_exchange(
          interp->rsh_sb_rrf->redist_halo,
          (int)(collection_size * num_src_fields),
          (void const**)(temp_src_fields + send_offset), (void**)temp_src_fields);

      size_t tgt_count = interp->rsh_sb_rrf->tgt_count;
      size_t * num_src_per_tgt = interp->rsh_sb_rrf->num_src_per_tgt;
      size_t * src_field_idx = interp->rsh_sb_rrf->src_field_idx;
      size_t * src_idx = interp->rsh_sb_rrf->src_idx;
      size_t local_data_offset = send_offset - num_src_fields;

      // sum source points
      for (size_t l = 0; l < collection_size; ++l) {
        for (size_t i = 0, k = 0; i < tgt_count; ++i) {
          size_t curr_num_src_per_tgt = num_src_per_tgt[i];
          double result = 0.0;
          for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
            result +=
              temp_src_fields
                [l * num_src_fields + src_field_idx[k] +
                 ((src_field_idx[k] >= num_src_fields)?(local_data_offset):(0))]
                [src_idx[k]];
          results[l][i] = result;
        }
      }
    }

    double * dummy = NULL;
    if (!interp->rsh_sb_rrf->is_target) tgt_field = &dummy;

    // redistribute results
    if (interp->rsh_sb_rrf->redist_results != NULL)
      xt_redist_s_exchange(
        interp->rsh_sb_rrf->redist_results, interp->collection_size,
        (const void **)results, (void **)tgt_field);
  }

  if (interp->rsh_sf != NULL) {

    size_t num_src_fields = interp->rsh_sf->num_src_fields;
      size_t send_offset = num_src_fields * collection_size;
    double ** temp_src_fields = interp->rsh_sf->src_fields_buffer;
    if (interp->rsh_sf->is_source) {
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] =
            src_fields[i][j];
    } else {
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          temp_src_fields[i * num_src_fields + j + send_offset] = NULL;
    }

    // do halo exchange
    if (interp->rsh_sf->redist_src != NULL)
      xt_redist_s_exchange(
        interp->rsh_sf->redist_src, (int)(num_src_fields * collection_size),
        (void const**)(temp_src_fields + send_offset), (void**)temp_src_fields);

    size_t * tgt_pos = interp->rsh_sf->tgt_pos;
    size_t tgt_count = interp->rsh_sf->tgt_count;
    size_t * num_src_per_tgt = interp->rsh_sf->num_src_per_tgt;
    size_t * src_field_idx = interp->rsh_sf->src_field_idx;
    size_t * src_idx = interp->rsh_sf->src_idx;

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
          result +=
            temp_src_fields[l * num_src_fields + src_field_idx[k]]
                           [src_idx[k]];
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  }

  size_t const fixed_count = interp->fixed_count;
  for (size_t l = 0; l < collection_size; ++l) {
    for (size_t i = 0; i < fixed_count; ++i) {
      size_t count = interp->fixed[i].count;
      size_t * pos = interp->fixed[i].pos;
      double value = interp->fixed[i].value;
      for (size_t j = 0; j < count; ++j)
        tgt_field[l][pos[j]] = value;
    }
  }

  if (interp->direct != NULL) {

    double ** temp_src_fields;
    if (interp->direct->is_source) {
      temp_src_fields = interp->direct->src_field_buffer;
      for (size_t i = 0; i < collection_size; ++i)
        temp_src_fields[i] = src_fields[i][0];
    } else {
      temp_src_fields = interp->direct->field_dummy;
    }

    if (!interp->direct->is_target) tgt_field = interp->direct->field_dummy;

    // send source points to the target processes
    if (interp->direct->is_source || interp->direct->is_target)
      xt_redist_s_exchange(
        interp->direct->redist, interp->collection_size,
        (const void **)temp_src_fields, (void **)tgt_field);
  }
}

// src_fields dimensions [collection_idx]
//                       [field index]
//                       [local_idx]
void yac_interpolation_execute_put(
  struct interpolation * interp, double *** src_fields) {

  // put cannot be called by processes that are source and target
  if (interp->is_source && interp->is_target)
    yac_internal_abort_message(
      "ERROR(yac_interpolation_execute_put): put called on process, which is "
      "source and target", __FILE__, __LINE__);
  else
    if (!interp->is_source) return;

  size_t collection_size = (size_t)(interp->collection_size);

  if (interp->rsh_wsb_rrf != NULL) {

    // wait until previous put is completed
    xt_request_wait(&(interp->rsh_wsb_rrf->request_results));

    size_t num_src_fields = interp->rsh_wsb_rrf->num_src_fields;
    size_t send_offset = num_src_fields * collection_size;
    double ** temp_src_fields = interp->rsh_wsb_rrf->src_fields_buffer;
    for (size_t i = 0; i < collection_size; ++i)
      for (size_t j = 0; j < num_src_fields; ++j)
        temp_src_fields[i * num_src_fields + j + send_offset] =
          src_fields[i][j];

    // do halo exchange
    if (interp->rsh_wsb_rrf->redist_halo != NULL)
      xt_redist_s_exchange(
        interp->rsh_wsb_rrf->redist_halo,
        (int)(collection_size * num_src_fields),
        (void const**)(temp_src_fields + send_offset), (void**)temp_src_fields);

    size_t tgt_count = interp->rsh_wsb_rrf->tgt_count;
    size_t * num_src_per_tgt = interp->rsh_wsb_rrf->num_src_per_tgt;
    double * weights = interp->rsh_wsb_rrf->weights;
    size_t * src_field_idx = interp->rsh_wsb_rrf->src_field_idx;
    size_t * src_idx = interp->rsh_wsb_rrf->src_idx;
    double ** results = interp->rsh_wsb_rrf->result_buffer;
    size_t local_data_offset = send_offset - num_src_fields;

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
          result +=
            temp_src_fields
              [l * num_src_fields + src_field_idx[k] +
               ((src_field_idx[k] >= num_src_fields)?(local_data_offset):(0))]
              [src_idx[k]] * weights[k];
        results[l][i] = result;
      }
    }

    // redistribute results
    double ** dummy = xmalloc(collection_size * sizeof(*dummy));
    for (size_t i = 0; i < collection_size; ++i) dummy[i] = NULL;
    xt_redist_a_exchange(
      interp->rsh_wsb_rrf->redist_results, interp->collection_size,
      (const void **)results, (void **)dummy,
      &(interp->rsh_wsb_rrf->request_results));
    free(dummy);
  }

  if (interp->rsh_wsf != NULL) {

    // wait until previous put is completed
    xt_request_wait(&(interp->rsh_wsf->request_send_src));

    size_t num_src_fields = interp->rsh_wsf->num_src_fields;
    size_t * src_send_buffer_sizes = interp->rsh_wsf->src_send_buffer_sizes;
    double ** src_send_buffer = interp->rsh_wsf->src_send_buffer;
    double ** src_fields_buffer = interp->rsh_wsf->src_fields_buffer;
    if (interp->rsh_wsf->is_source)
      for (size_t i = 0; i < collection_size; ++i)
        for (size_t j = 0; j < num_src_fields; ++j)
          memcpy(src_send_buffer[i * num_src_fields + j], src_fields[i][j],
                 src_send_buffer_sizes[j]);

    // send source field data to target processes
    if (interp->rsh_wsf->redist_src != NULL)
      xt_redist_a_exchange(
        interp->rsh_wsf->redist_src, (int)(num_src_fields * collection_size),
        (const void **)src_send_buffer, (void**)src_fields_buffer,
        &(interp->rsh_wsf->request_send_src));
  }

  if ((interp->direct != NULL) && (interp->direct->is_source)){

    double ** temp_src_fields = interp->direct->src_field_buffer;
    for (size_t i = 0; i < collection_size; ++i)
      temp_src_fields[i] = src_fields[i][0];

    // send source points to the target processes
    xt_redist_s_exchange(
      interp->direct->redist, interp->collection_size,
      (const void **)temp_src_fields, (void **)interp->direct->field_dummy);
  }
}

// tgt_field dimensions [collection_idx]
//                      [local_idx]
void yac_interpolation_execute_get(
  struct interpolation * interp, double ** tgt_field) {

  // get cannot be called by processes that are source and target
  if (interp->is_source && interp->is_target)
    yac_internal_abort_message(
      "ERROR(yac_interpolation_execute_get): get called on process, which is "
      "source and target", __FILE__, __LINE__);
  else
    if (!interp->is_target) return;

  size_t collection_size = interp->collection_size;

  if (interp->rsh_wsb_rrf != NULL) {

    // redistribute results
    double ** dummy = xmalloc(collection_size * sizeof(*dummy));
    for (size_t i = 0; i < collection_size; ++i) dummy[i] = NULL;
    xt_redist_a_exchange(
      interp->rsh_wsb_rrf->redist_results, collection_size,
      (const void **)dummy, (void **)tgt_field,
      &(interp->rsh_wsb_rrf->request_results));
    xt_request_wait(&(interp->rsh_wsb_rrf->request_results));
    free(dummy);
  }

  if (interp->rsh_wsf != NULL) {

    size_t num_src_fields = interp->rsh_wsf->num_src_fields;
    double ** src_fields_buffer = interp->rsh_wsf->src_fields_buffer;

    // receive source field data
    Xt_request request_src_field;
    double ** dummy =
      xmalloc(num_src_fields * collection_size * sizeof(*dummy));
    for (size_t i = 0; i < num_src_fields * collection_size; ++i)
      dummy[i] = NULL;
    xt_redist_a_exchange(
      interp->rsh_wsf->redist_src, (int)(num_src_fields * collection_size),
      (const void **)dummy, (void**)src_fields_buffer,
      &request_src_field);
    xt_request_wait(&request_src_field);
    free(dummy);

    size_t * tgt_pos = interp->rsh_wsf->tgt_pos;
    size_t tgt_count = interp->rsh_wsf->tgt_count;
    size_t * num_src_per_tgt = interp->rsh_wsf->num_src_per_tgt;
    double * weights = interp->rsh_wsf->weights;
    size_t * src_field_idx = interp->rsh_wsf->src_field_idx;
    size_t * src_idx = interp->rsh_wsf->src_idx;

    // apply weights
    for (size_t l = 0; l < collection_size; ++l) {
      for (size_t i = 0, k = 0; i < tgt_count; ++i) {
        size_t curr_num_src_per_tgt = num_src_per_tgt[i];
        double result = 0.0;
        for (size_t j = 0; j < curr_num_src_per_tgt; ++j, ++k)
          result +=
            src_fields_buffer[src_field_idx[k] * collection_size + l]
                             [src_idx[k]] * weights[k];
        tgt_field[l][tgt_pos[i]] = result;
      }
    }
  }

  size_t const fixed_count = interp->fixed_count;
  for (size_t l = 0; l < collection_size; ++l) {
    for (size_t i = 0; i < fixed_count; ++i) {
      size_t count = interp->fixed[i].count;
      size_t * pos = interp->fixed[i].pos;
      double value = interp->fixed[i].value;
      for (size_t j = 0; j < count; ++j)
        tgt_field[l][pos[j]] = value;
    }
  }

  if ((interp->direct != NULL) && (interp->direct->is_target)) {

    // receive data from source processes
    xt_redist_s_exchange(
      interp->direct->redist, interp->collection_size,
      (const void **)interp->direct->field_dummy, (void **)tgt_field);
  }
}

void yac_interpolation_delete(struct interpolation * interp) {

  if (interp == NULL) return;

  if (interp->rsh_wsb_rrf != NULL) {
    xt_request_wait(&(interp->rsh_wsb_rrf->request_results));
    if (interp->rsh_wsb_rrf->redist_results != NULL)
      xt_redist_delete(interp->rsh_wsb_rrf->redist_results);
    free(interp->rsh_wsb_rrf->result_buffer[0]);
    free(interp->rsh_wsb_rrf->result_buffer);
    free(interp->rsh_wsb_rrf->src_idx);
    free(interp->rsh_wsb_rrf->src_field_idx);
    free(interp->rsh_wsb_rrf->weights);
    free(interp->rsh_wsb_rrf->num_src_per_tgt);
    free(interp->rsh_wsb_rrf->halo_sizes);
    if (interp->rsh_wsb_rrf->redist_halo != NULL)
      xt_redist_delete(interp->rsh_wsb_rrf->redist_halo);
    free(interp->rsh_wsb_rrf->src_fields_buffer[0]);
    free(interp->rsh_wsb_rrf->src_fields_buffer);
    free(interp->rsh_wsb_rrf);
  }

  if (interp->rsh_sb_rrf != NULL) {
    xt_request_wait(&(interp->rsh_sb_rrf->request_results));
    if (interp->rsh_sb_rrf->redist_results != NULL)
      xt_redist_delete(interp->rsh_sb_rrf->redist_results);
    free(interp->rsh_sb_rrf->result_buffer[0]);
    free(interp->rsh_sb_rrf->result_buffer);
    free(interp->rsh_sb_rrf->src_idx);
    free(interp->rsh_sb_rrf->src_field_idx);
    free(interp->rsh_sb_rrf->num_src_per_tgt);
    free(interp->rsh_sb_rrf->halo_sizes);
    if (interp->rsh_sb_rrf->redist_halo != NULL)
      xt_redist_delete(interp->rsh_sb_rrf->redist_halo);
    free(interp->rsh_sb_rrf->src_fields_buffer[0]);
    free(interp->rsh_sb_rrf->src_fields_buffer);
    free(interp->rsh_sb_rrf);
  }

  if (interp->rsh_wsf != NULL) {

    if (interp->rsh_wsf->redist_src != NULL)
      xt_redist_delete(interp->rsh_wsf->redist_src);
    xt_request_wait(&(interp->rsh_wsf->request_send_src));
    free(interp->rsh_wsf->tgt_pos);
    free(interp->rsh_wsf->src_send_buffer[0]);
    free(interp->rsh_wsf->src_send_buffer);
    free(interp->rsh_wsf->src_fields_buffer[0]);
    free(interp->rsh_wsf->src_fields_buffer);
    free(interp->rsh_wsf->src_send_buffer_sizes);
    free(interp->rsh_wsf->src_recv_buffer_sizes);
    free(interp->rsh_wsf->src_idx);
    free(interp->rsh_wsf->src_field_idx);
    free(interp->rsh_wsf->weights);
    free(interp->rsh_wsf->num_src_per_tgt);
    free(interp->rsh_wsf);
  }

  if (interp->rsh_sf != NULL) {

    if (interp->rsh_sf->redist_src != NULL)
      xt_redist_delete(interp->rsh_sf->redist_src);
    xt_request_wait(&(interp->rsh_sf->request_send_src));
    free(interp->rsh_sf->tgt_pos);
    free(interp->rsh_sf->src_send_buffer[0]);
    free(interp->rsh_sf->src_send_buffer);
    free(interp->rsh_sf->src_fields_buffer[0]);
    free(interp->rsh_sf->src_fields_buffer);
    free(interp->rsh_sf->src_send_buffer_sizes);
    free(interp->rsh_sf->src_recv_buffer_sizes);
    free(interp->rsh_sf->src_idx);
    free(interp->rsh_sf->src_field_idx);
    free(interp->rsh_sf->num_src_per_tgt);
    free(interp->rsh_sf);
  }

  if (interp->direct != NULL) {

    xt_redist_delete(interp->direct->redist);
    free(interp->direct->src_field_buffer);
    free(interp->direct->field_dummy);
    free(interp->direct);
  }

  for (size_t i = 0; i < interp->fixed_count; ++i)
    free(interp->fixed[i].pos);
  free(interp->fixed);
  free(interp);
}
