!>
!! @file mo_yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface description
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!                                 Thomas Jahns <jahns@dkrz.de>
!! Fortran user interface for YAC.
!!
!!  Interface descriptions for
!!
!!  - \ref yac_finit
!!  - \ref yac_ffinalize
!!  - \ref yac_fsearch
!!
!!  - \ref yac_fdef_datetime
!!  - \ref yac_fdef_comp
!!  - \ref yac_fdef_comps
!!  - yac_fdef_grid
!!  - yac_fdef_points
!!  - \ref yac_fset_global_index
!!  - yac_fset_core_mask
!!  - yac_fset_mask
!!  - \ref yac_fdef_field
!!
!!  - yac_fput
!!  - yac_fget
!!
!!  - yac_fget_local_comm
!!  - \ref yac_fget_pair_rootcomm
!!  - \ref yac_fget_nbr_fields
!!  - \ref yac_fget_field_ids
!!  - yac_fget_model_timestep
!!  - yac_fget_coupling_timestep
!!
!!  - \ref yac_abort_message
!!  - \ref yac_redirstdout
!!
!! @version 1.0
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!         Thomas Jahns <jahns@dkrz.de>
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
!             Thomas Jahns <jahns@dkrz.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!

!> \example test_dummy_coupling.f90
!! This example simulates a whole model setup with three components (ocean,
!! atmosphere, io). It uses one process for each component.

module mo_yac_finterface

  !----------------------------------------------------------------------
  !>
  !!   Constants
  !!
  !----------------------------------------------------------------------

  public

  integer, parameter :: YAC_MAX_CHARLEN = 132


  !----------------------------------------------------------------------
  !>
  !!   Point types
  !!
  !----------------------------------------------------------------------

  integer, parameter :: YAC_LOCATION_CELL = 0
  integer, parameter :: YAC_LOCATION_CORNER = 1
  integer, parameter :: YAC_LOCATION_EDGE = 2

  !----------------------------------------------------------------------
  !>
  !!   Action types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: NONE               = 0
   enumerator :: COUPLING           = 1
   enumerator :: RESTART            = 2
   enumerator :: GET_FOR_RESTART    = 3
   enumerator :: PUT_FOR_RESTART    = 4
   enumerator :: GET_FOR_CHECKPOINT = 5
   enumerator :: PUT_FOR_CHECKPOINT = 6
   enumerator :: OUT_OF_BOUND       = 7
  end enum

  !----------------------------------------------------------------------
  !>
  !!   Reduction types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: INSTANT      = 0
   enumerator :: ACCUMULATE   = 1
   enumerator :: TIME_AVERAGE = 2
  end enum

  !----------------------------------------------------------------------
  !>
  !!   pointer data types
  !!
  !----------------------------------------------------------------------

  type yac_r4_ptr
    real, pointer :: p(:)
  end type yac_r4_ptr

  type yac_r8_ptr
    double precision, pointer :: p(:)
  end type yac_r8_ptr

  !----------------------------------------------------------------------
  !>
  !!   Fortran2C interface for yac collective routines
  !!
  !----------------------------------------------------------------------

  !> \example test_init_final.f90
  !! This contains an example on how to use yac_init and yac_ffinalize.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler initialisation
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_finit ( xml_filename, schema_filename )

       character(len=*), intent(in) :: xml_filename    !< [IN] xml description
       character(len=*), intent(in) :: schema_filename !< [IN] xml schema

     end subroutine yac_finit

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface
     subroutine yac_cinit ( xml_filename, schema_filename ) &
    &      bind ( c, name='yac_cinit' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: xml_filename
       character ( kind=c_char), dimension(*) :: schema_filename

     end subroutine yac_cinit

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler termination
  !!
  !----------------------------------------------------------------------

  interface
     subroutine yac_ffinalize ()
     end subroutine yac_ffinalize
  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface
     subroutine yac_cfinalize () bind ( c, name='yac_cfinalize' )
     end subroutine yac_cfinalize
  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting the yac version
  !!
  !----------------------------------------------------------------------

  interface
     function yac_fget_version () result (version_string)
     import :: YAC_MAX_CHARLEN
     character(len=YAC_MAX_CHARLEN) :: version_string
     end function yac_fget_version
  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface
     function yac_cget_version () bind ( c, name='yac_cget_version' )

     use, intrinsic :: iso_c_binding, only : c_ptr
     type(c_ptr) :: yac_cget_version

     end function yac_cget_version
  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the component definition
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_comp ( comp_name, comp_id )

       character(len=*), intent(in) :: comp_name !< [IN]   comp name
       integer, intent(out)         :: comp_id   !< [OUT]  component identifier

     end subroutine yac_fdef_comp

  end interface

  interface

     subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

        use, intrinsic :: iso_c_binding, only : c_char
        import :: YAC_MAX_CHARLEN

        integer, intent(in)   :: num_comps
        character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                                 comp_names(num_comps)
        integer, intent(out)  :: comp_ids(num_comps)

     end subroutine yac_fdef_comps

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_comp ( comp_name, comp_id ) &
    &      bind ( c, name='yac_cdef_comp' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: comp_name
       integer   ( kind=c_int )               :: comp_id

     end subroutine yac_cdef_comp

  end interface

  interface

     subroutine yac_cdef_comps ( comp_names, num_comps, comp_ids ) &
    &      bind ( c, name='yac_cdef_comps' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       type ( c_ptr )                :: comp_names(*)
       integer ( kind=c_int ), value :: num_comps
       integer ( kind=c_int )        :: comp_ids(*)

     end subroutine yac_cdef_comps

  end interface

  !> \example test_def_datetime.f90
  !! This shows how to use yac_fdef_datetime.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of start- and end datetime
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_datetime ( start_datetime, end_datetime )

       character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
       character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

     end subroutine yac_fdef_datetime

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_datetime ( start_datetime, end_datetime ) &
    &      bind ( c, name='yac_cdef_datetime' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: start_datetime
       character ( kind=c_char), dimension(*) :: end_datetime

     end subroutine yac_cdef_datetime

  end interface

  !> \example test_def_grid.f90
  !! This shows how to use yac_fdef_grid

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of grids
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_grid

     subroutine yac_fdef_grid_nonuniform_r4 ( grid_name,             &
                                              nbr_vertices,          &
                                              nbr_cells,             &
                                              nbr_connections,       &
                                              nbr_vertices_per_cell, &
                                              x_vertices,            &
                                              y_vertices,            &
                                              cell_to_vertex,        &
                                              grid_id)

       character(len=*), intent(in) :: grid_name                        !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell
       real, intent(in)             :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                          !< [OUT] grid ID

     end subroutine yac_fdef_grid_nonuniform_r4

     subroutine yac_fdef_grid_nonuniform_r8 ( grid_name,             &
                                              nbr_vertices,          &
                                              nbr_cells,             &
                                              nbr_connections,       &
                                              nbr_vertices_per_cell, &
                                              x_vertices,            &
                                              y_vertices,            &
                                              cell_to_vertex,        &
                                              grid_id)

       character(len=*), intent(in) :: grid_name                        !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell
       double precision, intent(in) :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                          !< [OUT] grid ID

     end subroutine yac_fdef_grid_nonuniform_r8

     subroutine yac_fdef_grid_unstruct_r4 ( grid_name,             &
                                            nbr_vertices,          &
                                            nbr_cells,             &
                                            nbr_vertices_per_cell, &
                                            x_vertices,            &
                                            y_vertices,            &
                                            cell_to_vertex,        &
                                            grid_id)

       character(len=*), intent(in) :: grid_name                          !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
       integer, intent(in)          :: nbr_vertices_per_cell              !< [IN]  number of vertices for each cell
       real, intent(in)             :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex( &
                                         nbr_vertices_per_cell,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                          !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                            !< [OUT] grid ID

     end subroutine yac_fdef_grid_unstruct_r4

     subroutine yac_fdef_grid_unstruct_r8 ( grid_name,             &
                                            nbr_vertices,          &
                                            nbr_cells,             &
                                            nbr_vertices_per_cell, &
                                            x_vertices,            &
                                            y_vertices,            &
                                            cell_to_vertex,        &
                                            grid_id)

       character(len=*), intent(in) :: grid_name                          !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
       integer, intent(in)          :: nbr_vertices_per_cell              !< [IN]  number of vertices for each cell
       double precision, intent(in) :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex( &
                                         nbr_vertices_per_cell,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                          !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                            !< [OUT] grid ID

     end subroutine yac_fdef_grid_unstruct_r8

     subroutine yac_fdef_grid_reg2d_r4 ( grid_name,             &
                                         nbr_vertices,          &
                                         cyclic,                &
                                         x_vertices,            &
                                         y_vertices,            &
                                         grid_id)

       character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
       integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension
       real, intent(in)             :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices
       integer, intent(out)         :: grid_id                     !< [OUT] grid ID

     end subroutine yac_fdef_grid_reg2d_r4

     subroutine yac_fdef_grid_reg2d_r8 ( grid_name,             &
                                         nbr_vertices,          &
                                         cyclic,                &
                                         x_vertices,            &
                                         y_vertices,            &
                                         grid_id)

       character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
       integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension
       double precision, intent(in) :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices
       integer, intent(out)         :: grid_id                     !< [OUT] grid ID

     end subroutine yac_fdef_grid_reg2d_r8

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_grid_reg2d ( grid_name,    &
                                      nbr_vertices, &
                                      cyclic,       &
                                      x_vertices,   &
                                      y_vertices,   &
                                      grid_id )     &
    &      bind ( c, name='yac_cdef_grid_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    )               :: nbr_vertices(2)
       integer ( kind=c_int    )               :: cyclic(2)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices(1))
       real    ( kind=c_double )               :: y_vertices(nbr_vertices(2))
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_reg2d

  end interface

  interface

     subroutine yac_cdef_grid_unstruct ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_vertices_per_cell, &
                                         x_vertices,            &
                                         y_vertices,            &
                                         cell_to_vertex,        &
                                         grid_id )              &
     &     bind ( c, name='yac_cdef_grid_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char, c_double

       character ( kind=c_char ), dimension(*) :: grid_name
       integer ( kind=c_int    ), value        :: nbr_vertices
       integer ( kind=c_int    ), value        :: nbr_cells
       integer ( kind=c_int    )               :: nbr_vertices_per_cell(nbr_cells)
       real    ( kind=c_double )               :: x_vertices(nbr_vertices)
       real    ( kind=c_double )               :: y_vertices(nbr_vertices)
       integer ( kind=c_int    )               :: cell_to_vertex(nbr_cells,nbr_vertices)
       integer ( kind=c_int    )               :: grid_id

     end subroutine yac_cdef_grid_unstruct

  end interface

  !> \example test_def_points.f90
  !! This shows how to use yac_fdef_points.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of points
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_points

     subroutine yac_fdef_points_reg2d_r4 ( grid_id,       &
                                           nbr_points,    &
                                           location,      &
                                           x_points_real, &
                                           y_points_real, &
                                           point_id )

       integer, intent(in)  :: grid_id                 !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
       real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r4

     subroutine yac_fdef_points_reg2d_r8 ( grid_id,       &
                                           nbr_points,    &
                                           location,      &
                                           x_points,      &
                                           y_points,      &
                                           point_id )

       integer, intent(in)  :: grid_id                         !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points
       integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex
       double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                        !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r8

     subroutine yac_fdef_points_unstruct_r4 ( grid_id,       &
                                              nbr_points,    &
                                              location,      &
                                              x_points_real, &
                                              y_points_real, &
                                              point_id )

       integer, intent(in)  :: grid_id                   !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points                !< [IN]  number of points
       integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex
       real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                  !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r4

     subroutine yac_fdef_points_unstruct_r8 ( grid_id,       &
                                              nbr_points,    &
                                              location,      &
                                              x_points,      &
                                              y_points,      &
                                              point_id )

       integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points                   !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
       double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r8

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_points_reg2d ( grid_id,      &
                                        nbr_points,   &
                                        location,     &
                                        x_points,     &
                                        y_points,     &
                                        point_id )    &
    &      bind ( c, name='yac_cdef_points_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_int )        :: nbr_points(2)
       integer ( kind=c_int ), value :: location

       real    ( kind=c_double )     :: x_points(nbr_points(1))
       real    ( kind=c_double )     :: y_points(nbr_points(2))

       integer ( kind=c_int )        :: point_id

     end subroutine yac_cdef_points_reg2d

  end interface

  interface

     subroutine yac_cdef_points_unstruct ( grid_id,      &
                                           nbr_points,   &
                                           location,     &
                                           x_points,     &
                                           y_points,     &
                                           point_id )    &
    &      bind ( c, name='yac_cdef_points_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer (kind=c_int), value :: grid_id
       integer (kind=c_int), value :: nbr_points
       integer (kind=c_int), value :: location

       real    (kind=c_double)     :: x_points(nbr_points)
       real    (kind=c_double)     :: y_points(nbr_points)

       integer (kind=c_int)        :: point_id

     end subroutine yac_cdef_points_unstruct

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of grid global ids
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fset_global_index ( global_index, &
                                        location,     &
                                        grid_id )

       integer, intent(in)  :: global_index(*)
       integer, intent(in)  :: location
       integer, intent(in)  :: grid_id

     end subroutine yac_fset_global_index

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cset_global_index ( global_index, &
                                        location,     &
                                        grid_id )     &
    &      bind ( c, name='yac_cset_global_index' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: global_index(*)
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id

     end subroutine yac_cset_global_index

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of a grid core masks
  !!
  !----------------------------------------------------------------------

  interface yac_fset_core_mask

     subroutine yac_fset_core_imask ( is_core,  &
                                      location, &
                                      grid_id )

       integer, intent(in)  :: is_core(*)
       integer, intent(in)  :: location
       integer, intent(in)  :: grid_id

     end subroutine yac_fset_core_imask

     subroutine yac_fset_core_lmask ( is_core,  &
                                      location, &
                                      grid_id )

       logical, intent(in)  :: is_core(*) !< [IN] mask array
       integer, intent(in)  :: location   !< [IN] location
       integer, intent(in)  :: grid_id    !< [IN] grid identifier

     end subroutine yac_fset_core_lmask

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cset_core_mask ( mask,         &
                                     location,    &
                                     grid_id )     &
    &      bind ( c, name='yac_cset_core_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: mask(*)
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id

     end subroutine yac_cset_core_mask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of pointset masks
  !!
  !----------------------------------------------------------------------

  interface yac_fset_mask

     subroutine yac_fset_imask ( is_valid,   &
                                 points_id )

       integer, intent(in)  :: is_valid(*)
       integer, intent(in)  :: points_id

     end subroutine yac_fset_imask

     subroutine yac_fset_lmask ( is_valid,   &
                                 points_id )

       logical, intent(in)  :: is_valid(*)  !< [IN]  mask array
       integer, intent(in)  :: points_id    !< [IN]  associated points id

     end subroutine yac_fset_lmask

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cset_mask ( is_valid,     &
                                points_id )   &
    &      bind ( c, name='yac_cset_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: is_valid(*)
       integer ( kind=c_int ), value :: points_id

     end subroutine yac_cset_mask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of coupling fields
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_field ( field_name,    &
                                 component_id,  &
                                 grid_id,       &
                                 point_ids,     &
                                 num_pointsets, &
                                 field_id )

       character(len=*), intent (in) :: field_name    !< [IN] short name of the field
       integer, intent (in)          :: component_id  !< [IN] component identifier
       integer, intent (in)          :: grid_id       !< [IN] grid identifier
       integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
       integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
       integer, intent (out)         :: field_id      !< [OUT] returned field handle

     end subroutine yac_fdef_field

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_field ( field_name,    &
                                 component_id,  &
                                 grid_id,       &
                                 point_ids,     &
                                 num_pointsets, &
                                 field_id )     &
    &      bind ( c, name='yac_cdef_field' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: field_name !< [IN] short name of the field
       integer ( kind=c_int ), value :: component_id         !< [IN] component identifier
       integer ( kind=c_int ), value :: grid_id              !< [IN] grid identifier
       integer ( kind=c_int )        :: point_ids(*)         !< [IN] point identifier
       integer ( kind=c_int ), value :: num_pointsets        !< [IN] number of pointsets per grid
       integer ( kind=c_int )        :: field_id             !< [OUT] returned field handle

     end subroutine yac_cdef_field

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for sending coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fput

     subroutine yac_fput_r4 ( field_id,         &
                              nbr_hor_points,   &
                              nbr_pointsets,    &
                              collection_size,  &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       real, intent (in)             :: send_field(nbr_hor_points, &
         &                                         nbr_pointsets, &
         &                                         collection_size) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r4

     subroutine yac_fput_r4_ptr ( field_id,         &
          &                       nbr_pointsets,    &
          &                       collection_size,  &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

       import :: yac_r4_ptr

       integer, intent (in)     :: field_id        !< [IN] field identifier
       integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                                   !< [IN] returned field
       integer, intent (out)    :: info            !< [OUT] returned info
       integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r4_ptr

     subroutine yac_fput_single_pointset_r4 ( field_id,         &
                                              nbr_hor_points,   &
                                              collection_size,  &
                                              send_field,       &
                                              info,             &
                                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       real, intent (in)             :: send_field(nbr_hor_points, collection_size) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_pointset_r4

     subroutine yac_fput_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (in) :: send_field(nbr_hor_points, nbr_pointsets, collection_size)
                                                        !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r8

     subroutine yac_fput_r8_ptr ( field_id,         &
          &                       nbr_pointsets,    &
          &                       collection_size,  &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

         import :: yac_r8_ptr

         integer, intent (in)     :: field_id        !< [IN] field identifier
         integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
         integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
         type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                                     !< [IN] returned field
         integer, intent (out)    :: info            !< [OUT] returned info
         integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r8_ptr

     subroutine yac_fput_single_pointset_r8 ( field_id,         &
                                              nbr_hor_points,   &
                                              collection_size,  &
                                              send_field,       &
                                              info,             &
                                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (in) :: send_field(nbr_hor_points, collection_size) !< [IN] returned
                                                                                                               !! field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_pointset_r8

  end interface yac_fput

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cput_ ( field_id,         &
                            collection_size,  &
                            send_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cput_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] number of vertical level or bundles
       real    ( kind=c_double )     :: send_field(*)       !< [IN] returned field handle
       integer ( kind=c_int )        :: info                !< [OUT] returned info argument
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cput_

  end interface

  interface

     subroutine yac_cput_ptr_ ( field_id,         &
                                collection_size,  &
                                send_field,       &
                                info,             &
                                ierror ) bind ( c, name='yac_cput_ptr_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] number of vertical level or bundles
       type(c_ptr)                   :: send_field(*)       !< [IN] returned field handle
       integer ( kind=c_int )        :: info                !< [OUT] returned info argument
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cput_ptr_

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for receiving coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fget

     subroutine yac_fget_r4 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       real, intent (inout)  :: recv_field(nbr_hor_points, collection_size) !< [INOUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4

     subroutine yac_fget_r4_ptr ( field_id,         &
                                  collection_size,  &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r4_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r4_ptr)      :: recv_field(collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4_ptr

     subroutine yac_fget_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (inout) :: &
          recv_field(nbr_hor_points, collection_size) !< [INOUT] returned
                                                                                                                      !! field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8



     subroutine yac_fget_r8_ptr ( field_id,         &
                                  collection_size,  &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r8_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r8_ptr)      :: recv_field(collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8_ptr

  end interface yac_fget

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_ ( field_id,         &
                            collection_size,  &
                            recv_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cget_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       real    ( kind=c_double )     :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget_

  end interface

  interface

     subroutine yac_cget ( field_id,         &
                           collection_size,  &
                           recv_field,       &
                           info,             &
                           ierror ) bind ( c, name='yac_cget' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       type(c_ptr)                   :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for setting and getting the io_proc_step of a component
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in)  :: comp_id            !< [IN] component identifier
       integer, intent (out) :: io_proc_step       !< [OUT] stride between ranks of io processes

     end subroutine yac_fget_comp_io_proc_step

     subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in) :: comp_id            !< [IN] component identifier
       integer, intent (in) :: io_proc_step       !< [IN] stride between ranks of io processes

     end subroutine yac_fset_comp_io_proc_step

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_comp_io_proc_step ( comp_id, io_proc_step ) &
    &      bind ( c, name='yac_cget_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ) :: io_proc_step

     end subroutine yac_cget_comp_io_proc_step

     subroutine yac_cset_comp_io_proc_step ( comp_id, io_proc_step ) &
    &      bind ( c, name='yac_cset_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ), value :: io_proc_step

     end subroutine yac_cset_comp_io_proc_step

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a local MPI communicator
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_localcomm ( local_communicator, comp_id )

       integer, intent (out) :: local_communicator             !< [OUT] MPI communicator
       integer, intent (in)  :: comp_id                        !< [IN]  component ID

     end subroutine yac_fget_localcomm

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_get_localcomm ( local_communicator, comp_id ) &
    &      bind ( c, name='yac_get_localcomm' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ) :: local_communicator
       integer ( kind=c_int ), value :: comp_id

     end subroutine yac_get_localcomm

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a MPI communicator for
  !!   communication between a pair of component root proceses
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_pair_rootcomm ( local_communicator, comp_name1, comp_name2 )

       character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
       character(len=*), intent(in) :: comp_name2  !< [IN]  second component name
       integer, intent (out) :: local_communicator !< [OUT] MPI communicator

     end subroutine yac_fget_pair_rootcomm

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_get_pair_rootcomm ( local_communicator, comp_name1, comp_name2 ) &
    &      bind ( c, name='yac_get_pair_rootcomm' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ) :: local_communicator
       character ( kind=c_char), dimension(*) :: comp_name1
       character ( kind=c_char), dimension(*) :: comp_name2

     end subroutine yac_get_pair_rootcomm

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking the end of the definition phase
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fsearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

       integer, intent (in)  :: nbr_comps             !< [IN]  number of components
       integer, intent (in)  :: comp_ids(nbr_comps)   !< [IN]  list of component IDs
       integer, intent (in)  :: nbr_fields            !< [IN]  number of fields
       integer, intent (in)  :: field_ids(nbr_fields) !< [IN]  list of field IDs
       integer, intent (out) :: ierror                !< [OUT] returned error handler

     end subroutine yac_fsearch

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_csearch ( nbr_comps, comp_ids,  nbr_fields, field_ids, ierror ) &
    &      bind ( c, name='yac_csearch' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: nbr_comps            !< [IN]  number of components
       integer ( kind=c_int )        :: comp_ids(nbr_comps)  !< [IN]  list of component IDs
       integer ( kind=c_int ), value :: nbr_fields           !< [IN]  number of fields
       integer ( kind=c_int )        :: field_ids(nbr_comps) !< [IN]  list of field IDs
       integer ( kind=c_int )        :: ierror               !< [OUT] returned error handler

     end subroutine yac_csearch

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking query functions
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_nbr_fields ( nbr_fields )

       integer, intent (out) :: nbr_fields

     end subroutine yac_fget_nbr_fields

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_nbr_fields ( nbr_fields ) bind ( c, name='yac_cget_nbr_fields' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: nbr_fields           !< [OUT]  number of components

     end subroutine yac_cget_nbr_fields

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_ids ( nbr_fields, field_ids )

       integer, intent (in)  :: nbr_fields
       integer, intent (out) :: field_ids(nbr_fields)

     end subroutine yac_fget_field_ids

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_ids ( nbr_fields, field_ids ) &
    &      bind ( c, name='yac_cget_field_ids' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: nbr_fields            !< [IN]  number of components
       integer ( kind=c_int )        :: field_ids(nbr_fields) !< [OUT] list of field IDS

     end subroutine yac_cget_field_ids

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_id_from_name ( field_name, field_id )

       character(len=*), intent (in) :: field_name
       integer, intent (out)                       :: field_id

     end subroutine yac_fget_field_id_from_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_id_from_name ( field_name, field_id ) &
    &      bind ( c, name='yac_cget_field_id_from_name' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN]  field name
       integer ( kind=c_int )                 :: field_id    !< [OUT] field ID

     end subroutine yac_cget_field_id_from_name

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_name_from_id ( field_id, field_name )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: field_name

     end subroutine yac_fget_field_name_from_id

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_name_from_id ( field_id, field_name ) &
    &      bind ( c, name='yac_cget_field_name_from_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN]  field ID
       character ( kind=c_char), dimension(*) :: field_name  !< [OUT] field name

     end subroutine yac_cget_field_name_from_id

  end interface

  ! ---------------------------------------------------------------------

  interface yac_fget_role_of_field

    subroutine yac_fget_role_of_field_by_id ( field_id, role )

      integer, intent (in)  :: field_id
      integer, intent (out) :: role

    end subroutine yac_fget_role_of_field_by_id

    ! ....................................................................

    subroutine yac_fget_role_of_field_by_name ( field_name, role )

      character (len=*), intent (in)    :: field_name
      integer, intent (out)                           :: role

     end subroutine yac_fget_role_of_field_by_name

  end interface yac_fget_role_of_field

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_role_of_field_by_id ( field_id, role ) &
    &      bind ( c, name='yac_cget_role_of_field_by_id' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value  :: field_id    !< [IN]  field ID
       integer ( kind=c_int)          :: role        !< [OUT] nothing (0) | put (1) | get (2)

    end subroutine yac_cget_role_of_field_by_id

    ! ...................................................................

    subroutine yac_cget_role_of_field_by_name ( field_name, role ) &
   &      bind ( c, name='yac_cget_role_of_field_by_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      character ( kind=c_char), dimension(*) :: field_name !< [IN]  field name
      integer ( kind=c_int)                  :: role       !< [OUT] nothing (0) | put (1) | get (2)

    end subroutine yac_cget_role_of_field_by_name

  end interface

! ---------------------------------------------------------------------

  interface yac_fget_model_timestep

     subroutine yac_fget_model_timestep_id ( field_id, string )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: string

     end subroutine yac_fget_model_timestep_id

     ! ..................................................................

     subroutine yac_fget_model_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name
       character(len=*), intent (out) :: string

     end subroutine yac_fget_model_timestep_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_model_timestep_id ( field_id, string ) &
    &      bind ( c, name='yac_cget_model_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN] field identifier
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_model_timestep_id

     ! ..................................................................

     subroutine yac_cget_model_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_model_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN] short name of the field
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_model_timestep_name

   end interface

  ! ---------------------------------------------------------------------

  interface yac_fget_coupling_timestep

     subroutine yac_fget_coupling_timestep_id ( field_id, string )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: string

     end subroutine yac_fget_coupling_timestep_id

     ! ..................................................................

     subroutine yac_fget_coupling_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name
       character(len=*), intent (out) :: string

     end subroutine yac_fget_coupling_timestep_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_coupling_timestep_id ( field_id, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN] field identifier
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_coupling_timestep_id

     ! ..................................................................

     subroutine yac_cget_coupling_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN] short name of the field
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_coupling_timestep_name

   end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_abort_message ( text, file, line ) bind ( c, name='yac_abort_message' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: text       !< [IN] error message
       character ( kind=c_char ), dimension(*) :: file       !< [IN] error message
       integer ( kind=c_int ), value           :: line       !< [IN] line number

     end subroutine yac_abort_message

  end interface

  interface

     subroutine yac_redirstdout ( filestem, lenstr, parallel, my_pe, npes, ierror ) &
    &      bind ( c, name='yac_redirstdout' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: filestem
       integer ( kind=c_int ), value           :: lenstr
       integer ( kind=c_int ), value           :: parallel
       integer ( kind=c_int ), value           :: my_pe
       integer ( kind=c_int ), value           :: npes
       integer ( kind=c_int )                  :: ierror

     end subroutine yac_redirstdout

  end interface

end module mo_yac_finterface
