!>
!! @file yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @par Warranty
!!
!! This code has been tested up to a certain level. Defects and weaknesses,
!! which may be included in the code, do not establish any warranties by the
!! authors. The authors do not make any warranty, express or implied, or assume any
!! liability or responsibility for the use, acquisition or application of this
!! software.
!!
!! @par Fortran User API
!!
!! Initialisation
!!
!! - \ref yac_finit
!! - \ref yac_ffinalize
!! - \ref yac_fget_localcomm
!!
!! Definition
!!
!! - \ref yac_fdef_datetime
!! - \ref yac_fdef_comp
!! - \ref yac_fdef_comps
!! - \ref mo_yac_finterface::yac_fdef_grid "yac_fdef_grid (*)"
!! - \ref mo_yac_finterface::yac_fdef_points "yac_fdef_points (*)"
!! - \ref yac_fset_global_index
!! - \ref yac_fset_core_lmask
!! - \ref yac_fset_core_imask
!! - \ref yac_fset_lmask
!! - \ref yac_fset_imask
!! - \ref yac_fdef_field
!!
!! End of definition and search
!!
!! - \ref yac_fsearch
!!
!! Data exchange
!!
!! - \ref mo_yac_finterface::yac_fput "yac_fput (*)"
!! - \ref mo_yac_finterface::yac_fget "yac_fget (*)"
!!
!! Query routines
!!
!! - \ref yac_fget_field_ids
!! - \ref yac_fget_nbr_fields
!! - \ref yac_fget_field_id_from_name
!! - \ref yac_fget_field_name_from_id
!!
!! - \ref mo_yac_finterface::yac_fget_role_of_field "yac_fget_role_of_field (*)"
!! - \ref mo_yac_finterface::yac_fget_model_timestep "yac_fget_model_timestep (*)"
!! - \ref mo_yac_finterface::yac_fget_coupling_timestep "yac_fget_coupling_timestep (*)"
!!
!! Auxillary routines
!!
!! - \ref yac_abort_message
!! - \ref yac_redirstdout
!!
!! For a detailed description of the Fortran interface please see mo_yac_finterface.f90.
!! Functions marked with (*) are overloaded with respect to the data type.
!! See mo_yac_finterface.f90
!! for details.
!!
!! For a detailed description of the c interface please see interface.h
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!

!
! -------------------------- init -------------------------------------

subroutine yac_finit ( xml_filename, schema_filename )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit ( TRIM(xml_filename)    // c_null_char, &
                   TRIM(schema_filename) // c_null_char )

end subroutine yac_finit

! -------------------------- final -------------------------------------

subroutine yac_ffinalize ( )

  use mo_yac_finterface, dummy => yac_ffinalize

  call yac_cfinalize ( )

end subroutine yac_ffinalize

! -------------------------- version ----------------------------------

function yac_fget_version () result (version_string)

  use, intrinsic :: iso_c_binding, only : c_ptr, c_char, c_null_char, c_f_pointer
  use mo_yac_finterface, dummy => yac_fget_version

  implicit none

  type (c_ptr)                                           :: c_string_ptr
  character (len=1, kind=c_char), dimension (:), pointer :: version_chars => null ()
  character (len=YAC_MAX_CHARLEN)                        :: version_string
  integer :: i

  c_string_ptr = yac_cget_version()

  CALL c_f_pointer ( c_string_ptr, version_chars, [ YAC_MAX_CHARLEN ])

  version_string = " "

  do i=1, YAC_MAX_CHARLEN
    if (version_chars(i) == c_null_char) exit
    version_string(i:i) = version_chars(i)
  end do

end function yac_fget_version

! -------------------------- dates ------------------------------------

subroutine yac_fdef_datetime ( start_datetime, end_datetime )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_datetime

  character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
  character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

  integer :: index

  index = 0


  if (present(start_datetime)) then
    call yac_check_strlength ( start_datetime )
    index = index + 1
  end if

  if (present(end_datetime)) then
    call yac_check_strlength ( end_datetime )
    index = index + 2
  end if

  select case ( index )

    case ( 3 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 2 )
      call yac_cdef_datetime ( c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 1 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               c_null_char )
    end select

end subroutine yac_fdef_datetime

! ------------------------ def_comp ------------------------------------

subroutine yac_fdef_comp ( comp_name, comp_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_comp

  character(len=*), intent(in) :: comp_name  !< [IN]  component name
  integer, intent(out)         :: comp_id    !< [OUT] returned handle to the component

  call yac_check_strlength ( comp_name )

  call yac_cdef_comp ( TRIM(comp_name) // c_null_char, comp_id )

end subroutine yac_fdef_comp

! ------------------------ def_comps ------------------------------------

subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

  use, intrinsic :: iso_c_binding, only : c_null_char, c_ptr, c_loc, c_char
  use mo_yac_finterface, dummy => yac_fdef_comps

  integer, intent(in)   :: num_comps !< [IN]  number of components
  character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                           comp_names(num_comps) !< [IN]  component names
  integer, intent(out)  :: comp_ids(num_comps) !< [OUT] returned handle to the components

  integer :: i, j
  character(kind=c_char), target :: comp_names_cpy(YAC_MAX_CHARLEN+1, num_comps)
  type(c_ptr) :: comp_name_ptrs(num_comps)

  comp_names_cpy = c_null_char

  do i = 1, num_comps
    do j = 1, len_trim(comp_names(i))
      comp_names_cpy(j,i) = comp_names(i)(j:j)
    end do
    comp_name_ptrs(i) = c_loc(comp_names_cpy(1,i));
  end do

  call yac_cdef_comps ( comp_name_ptrs, num_comps, comp_ids )

end subroutine yac_fdef_comps


! ------------------------- def_points ----------------------------------

subroutine yac_fdef_points_reg2d_r4 ( grid_id,       &
                                      nbr_points,    &
                                      location,      &
                                      x_points_real, &
                                      y_points_real, &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r4

  integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points in x and y
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
  real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points
  integer, intent(out) :: point_id                     !< [OUT] point identifier

  double precision    ::  x_points(nbr_points(1))
  double precision    ::  y_points(nbr_points(2))

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_cdef_points_reg2d ( grid_id,    &
                               nbr_points, &
                               location,   &
                               x_points,   &
                               y_points,   &
                               point_id )

end subroutine yac_fdef_points_reg2d_r4

subroutine yac_fdef_points_reg2d_r8 ( grid_id,    &
                                      nbr_points, &
                                      location,   &
                                      x_points,   &
                                      y_points,   &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r8

  integer, intent(in)  :: grid_id                         !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points in x and y
  integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                        !< [OUT] point identifier

  call yac_cdef_points_reg2d ( grid_id, &
                               nbr_points,  &
                               location,    &
                               x_points,    &
                               y_points,    &
                               point_id )

end subroutine yac_fdef_points_reg2d_r8


subroutine yac_fdef_points_unstruct_r4 ( grid_id,       &
                                         nbr_points,    &
                                         location,      &
                                         x_points_real, &
                                         y_points_real, &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r4

  integer, intent(in)  :: grid_id                   !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points                !< [IN]  number of points
  integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex

  real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                  !< [OUT] point identifier

  double precision    ::  x_points(nbr_points)
  double precision    ::  y_points(nbr_points)

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_cdef_points_unstruct ( grid_id,    &
                                  nbr_points, &
                                  location,   &
                                  x_points,   &
                                  y_points,   &
                                  point_id )

end subroutine yac_fdef_points_unstruct_r4

subroutine yac_fdef_points_unstruct_r8 ( grid_id,    &
                                         nbr_points, &
                                         location,   &
                                         x_points,   &
                                         y_points,   &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r8

  integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
  integer, intent(in)  :: nbr_points                   !< [IN]  number of points
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                     !< [OUT] grid identifier

  call yac_cdef_points_unstruct  ( grid_id,    &
                                   nbr_points, &
                                   location,   &
                                   x_points,   &
                                   y_points,   &
                                   point_id )

end subroutine yac_fdef_points_unstruct_r8

! ------------------------- def_grid -------------------------------

subroutine yac_fdef_grid_nonuniform_r4 ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_connections,       &
                                         nbr_vertices_per_cell, &
                                         x_vertices_real,       &
                                         y_vertices_real,       &
                                         cell_to_vertex_in,     &
                                         grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_nonuniform_r4

  character(len=*), intent(in) :: grid_name                  !< [IN]  grid name
  integer, intent(in)  :: nbr_vertices                       !< [IN]  number of vertices
  integer, intent(in)  :: nbr_cells                          !< [IN]  number of cells
  integer, intent(in)  :: nbr_connections                    !< [IN]  total size of cell_to_vertex
  integer, intent(in)  :: nbr_vertices_per_cell(nbr_cells)   !< [IN]  number of vertices for each cell

  real, intent(in)     :: x_vertices_real(nbr_vertices)      !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices)      !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_connections) !< [IN] connectivity between vertices and cells\n
                                                             !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer, intent(out) :: grid_id                            !< [OUT] grid identifier

  double precision     :: x_vertices(nbr_vertices)
  double precision     :: y_vertices(nbr_vertices)

  integer              :: cell_to_vertex(nbr_connections)

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_nonuniform_r4): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:) = cell_to_vertex_in(:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_nonuniform_r4

subroutine yac_fdef_grid_nonuniform_r8 ( grid_name,             &
                                         nbr_vertices,          &
                                         nbr_cells,             &
                                         nbr_connections,       &
                                         nbr_vertices_per_cell, &
                                         x_vertices,            &
                                         y_vertices,            &
                                         cell_to_vertex_in,     &
                                         grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_nonuniform_r8

  character(len=*), intent(in) :: grid_name                          !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
  integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
  integer, intent(in)          :: nbr_connections                    !< [IN]  total size of cell_to_vertex
  integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells)   !< [IN]  number of vertices for each cell

  double precision, intent(in) :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices

  integer, intent(in)          :: cell_to_vertex_in(nbr_connections) !< [IN] connectivity between vertices and cells\n
                                                                     !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer, intent(out)         :: grid_id                            !< [OUT] grid identifier

  integer                      :: cell_to_vertex(nbr_connections)

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_nonuniform_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:) = cell_to_vertex_in(:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_nonuniform_r8

subroutine yac_fdef_grid_unstruct_r4 ( grid_name,                &
                                       nbr_vertices,             &
                                       nbr_cells,                &
                                       nbr_vertices_per_cell_in, &
                                       x_vertices_real,          &
                                       y_vertices_real,          &
                                       cell_to_vertex_in,        &
                                       grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_unstruct_r4

  character(len=*), intent(in) :: grid_name                 !< [IN]  grid name
  integer, intent(in)  :: nbr_vertices                      !< [IN]  number of vertices
  integer, intent(in)  :: nbr_cells                         !< [IN]  number of cells
  integer, intent(in)  :: nbr_vertices_per_cell_in          !< [IN]  number of vertices for each cell

  real, intent(in)     :: x_vertices_real(nbr_vertices)     !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices)     !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_vertices_per_cell_in,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                                !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer, intent(out)   :: grid_id                         !< [OUT] grid identifier

  double precision       :: x_vertices(nbr_vertices)
  double precision       :: y_vertices(nbr_vertices)

  integer                :: nbr_vertices_per_cell(nbr_cells)
  integer                :: cell_to_vertex(nbr_vertices_per_cell_in,nbr_cells)

  nbr_vertices_per_cell(:) = nbr_vertices_per_cell_in

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_unstruct_r4): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:,:) = cell_to_vertex_in(:,:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_unstruct_r4

subroutine yac_fdef_grid_unstruct_r8 ( grid_name,                &
                                       nbr_vertices,             &
                                       nbr_cells,                &
                                       nbr_vertices_per_cell_in, &
                                       x_vertices,               &
                                       y_vertices,               &
                                       cell_to_vertex_in,        &
                                       grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_unstruct_r8

  character(len=*), intent(in) :: grid_name                !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices             !< [IN]  number of vertices
  integer, intent(in)          :: nbr_cells                !< [IN]  number of cells
  integer, intent(in)          :: nbr_vertices_per_cell_in !< [IN]  number of vertices for each cell

  double precision, intent(in) :: x_vertices(nbr_vertices) !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices) !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_vertices_per_cell_in,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                                !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer, intent(out)         :: grid_id                  !< [OUT] grid identifier

  integer                      :: nbr_vertices_per_cell(nbr_cells)
  integer                      :: cell_to_vertex(nbr_vertices_per_cell_in,nbr_cells)

  nbr_vertices_per_cell(:) = nbr_vertices_per_cell_in

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_grid_unstruct_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:,:) = cell_to_vertex_in(:,:) - 1

  call yac_cdef_grid_unstruct  ( TRIM(grid_name) // c_null_char, &
                                 nbr_vertices,                   &
                                 nbr_cells,                      &
                                 nbr_vertices_per_cell,          &
                                 x_vertices,                     &
                                 y_vertices,                     &
                                 cell_to_vertex,                 &
                                 grid_id )

end subroutine yac_fdef_grid_unstruct_r8

subroutine yac_fdef_grid_reg2d_r4 ( grid_name,       &
                                    nbr_vertices,    &
                                    cyclic,          &
                                    x_vertices_real, &
                                    y_vertices_real, &
                                    grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_reg2d_r4

  character(len=*), intent(in) :: grid_name                !< [IN]  grid name
  integer, intent(in)  :: nbr_vertices(2)                  !< [IN]  number of cells in each dimension
  integer, intent(in)  :: cyclic(2)                        !< [IN]  cyclic behavior of cells in each dimension

  real, intent(in)     :: x_vertices_real(nbr_vertices(1)) !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices(2)) !< [IN]  latitudes of vertices

  integer, intent(out) :: grid_id                          !< [OUT] grid identifier

  double precision     ::  x_vertices(nbr_vertices(1))
  double precision     ::  y_vertices(nbr_vertices(2))

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  call yac_cdef_grid_reg2d ( TRIM(grid_name) // c_null_char, &
                             nbr_vertices,                   &
                             cyclic,                         &
                             x_vertices,                     &
                             y_vertices,                     &
                             grid_id )

end subroutine yac_fdef_grid_reg2d_r4

subroutine yac_fdef_grid_reg2d_r8 ( grid_name,    &
                                    nbr_vertices, &
                                    cyclic,       &
                                    x_vertices,   &
                                    y_vertices,   &
                                    grid_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_grid_reg2d_r8

  character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
  integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
  integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension

  double precision, intent(in) :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices

  integer, intent(out)         :: grid_id                     !< [OUT] grid identifier

  call yac_cdef_grid_reg2d ( TRIM(grid_name) // c_null_char, &
                             nbr_vertices,                   &
                             cyclic,                         &
                             x_vertices,                     &
                             y_vertices,                     &
                             grid_id )

end subroutine yac_fdef_grid_reg2d_r8

! ------------------------- set global_index ------------------------------

subroutine yac_fset_global_index ( global_index, &
                                   location,     &
                                   grid_id )

  use mo_yac_finterface, dummy => yac_fset_global_index

  integer, intent(in)  :: global_index(*) !< [IN] global index
  integer, intent(in)  :: location        !< [IN] location
  integer, intent(in)  :: grid_id         !< [IN] point set indentifier

  call yac_cset_global_index ( global_index, &
                               location, &
                               grid_id )

end subroutine yac_fset_global_index

! ------------------------- set core_mask ------------------------------

subroutine yac_fset_core_lmask ( is_core,    &
                                 location, &
                                 grid_id )

  use mo_yac_finterface, dummy => yac_fset_core_lmask
  use, intrinsic :: iso_c_binding, only : c_size_t

  logical, intent(in)  :: is_core(*) !< [IN] core flag
                                     !< false, cell/vertex/edge is halo
                                     !< true, cell/vertex/edge is core
  integer, intent(in)  :: location   !< [IN] location
  integer, intent(in)  :: grid_id    !< [IN] point set indentifier

  integer (kind=c_size_t) :: i, count
  integer, allocatable :: int_is_core(:)

  interface

     function yac_get_grid_size ( location, &
                                  grid_id ) &
    &      result ( grid_size )             &
    &      bind ( c, name='yac_get_grid_size' )

       use, intrinsic :: iso_c_binding, only : c_int, c_size_t

       integer ( kind=c_int ), value :: location
       integer ( kind=c_int ), value :: grid_id
       integer ( kind=c_size_t) :: grid_size

     end function yac_get_grid_size

  end interface

  count = yac_get_grid_size(location, grid_id)
  allocate(int_is_core(count))

  do i = 1, count
     if ( is_core(i) ) then
        int_is_core(i) = 1
     else
        int_is_core(i) = 0
     endif
  enddo

  call yac_cset_core_mask ( int_is_core,    &
                            location, &
                            grid_id )

end subroutine yac_fset_core_lmask

subroutine yac_fset_core_imask ( is_core,    &
                                 location, &
                                 grid_id )

  use mo_yac_finterface, dummy => yac_fset_core_imask

  integer, intent(in)  :: is_core(*) !< [IN] core flag
                                     !< 0, cell/vertex/edge is halo
                                     !< 1, cell/vertex/edge is core
  integer, intent(in)  :: location   !< [IN] location
  integer, intent(in)  :: grid_id    !< [IN] point set indentifier

  call yac_cset_core_mask ( is_core, &
                            location, &
                            grid_id )

end subroutine yac_fset_core_imask

! ------------------------- set_mask ------------------------------

subroutine yac_fset_lmask ( is_valid,     &
                            points_id )

  use mo_yac_finterface, dummy => yac_fset_lmask
  use, intrinsic :: iso_c_binding, only : c_size_t

  logical, intent(in)  :: is_valid(*)  !< [IN] logical mask
                                       !< false, point is masked out
                                       !< true, point is valid
  integer, intent(in)  :: points_id    !< [IN] point set indentifier

  integer ( kind=c_size_t) :: i, count
  integer, allocatable :: int_is_valid(:)

  interface

     function yac_get_points_size ( points_id ) &
    &      result ( points_size )             &
    &      bind ( c, name='yac_get_points_size' )

       use, intrinsic :: iso_c_binding, only : c_int, c_size_t

       integer ( kind=c_int ), value :: points_id
       integer ( kind=c_size_t) :: points_size

     end function yac_get_points_size

  end interface

  count = yac_get_points_size(points_id)
  allocate(int_is_valid(count))

  do i = 1, count
     if ( is_valid(i) ) then
        int_is_valid(i) = 1
     else
        int_is_valid(i) = 0
     endif
  enddo

  call yac_cset_mask ( int_is_valid, &
                       points_id )

end subroutine yac_fset_lmask

subroutine yac_fset_imask ( is_valid, &
                            points_id )

  use mo_yac_finterface, dummy => yac_fset_imask

  integer, intent(in)  :: is_valid(*)  !< [IN] integer mask
                                       !< 0, point is masked out
                                       !< 1, point is valid
  integer, intent(in)  :: points_id    !< [IN] point set indentifier

  call yac_cset_mask ( is_valid, &
                       points_id )

end subroutine yac_fset_imask

! ----------------------------- def_field -------------------------------

subroutine yac_fdef_field ( field_name,    &
                            component_id,  &
                            grid_id,       &
                            point_ids,     &
                            num_pointsets, &
                            field_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_field

  !
  ! Definition of coupling fields
  !
  character(len=*), intent (in) :: field_name    !< [IN] short name of the field
  integer, intent (in)          :: component_id  !< [IN] component identifier
  integer, intent (in)          :: grid_id       !< [IN] grid identifier
  integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
  integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
  integer, intent (out)         :: field_id      !< [OUT] returned field handle

  call yac_check_strlength ( field_name )

  call yac_cdef_field ( TRIM(field_name) // c_null_char, &
                        component_id,                    &
                        grid_id,                         &
                        point_ids,                       &
                        num_pointsets,                   &
                        field_id )

end subroutine yac_fdef_field

! ---------------------------------- put --------------------------------

subroutine yac_fput_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   nbr_pointsets,    &
     &                   collection_size,  &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r4

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  real, intent (in)     :: send_field(nbr_hor_points, nbr_pointsets, collection_size)
                                           !< [IN] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error

  double precision :: send_buffer(nbr_hor_points, nbr_pointsets, collection_size)

  send_buffer = DBLE(send_field)

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_buffer,      &
     &             info,             &
     &             ierror )

end subroutine yac_fput_r4

subroutine yac_fput_r4_ptr ( field_id,         &
     &                       nbr_pointsets,    &
     &                       collection_size,  &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r4_ptr

  integer, intent (in)     :: field_id        !< [IN] field identifier
  integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                              !< [IN] returned field
  integer, intent (out)    :: info            !< [OUT] returned info
  integer, intent (out)    :: ierror          !< [OUT] returned error

  integer :: i, j
  type(yac_r8_ptr) :: send_field_r8(nbr_pointsets, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      allocate(send_field_r8(j, i)%p(SIZE(send_field(j, i)%p)))
      send_field_r8(j, i)%p = dble(send_field(j, i)%p)
    end do
  end do

  call yac_fput_r8_ptr ( field_id,         &
     &                   nbr_pointsets,    &
     &                   collection_size,  &
     &                   send_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      deallocate(send_field_r8(j, i)%p)
    end do
  end do

end subroutine yac_fput_r4_ptr

subroutine yac_fput_single_pointset_r4 ( field_id,         &
     &                                   nbr_hor_points,   &
     &                                   collection_size,  &
     &                                   send_field,       &
     &                                   info,             &
     &                                   ierror )

  use mo_yac_finterface, dummy => yac_fput_single_pointset_r4

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  real, intent (in)             :: send_field(nbr_hor_points, collection_size)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: send_buffer(nbr_hor_points, collection_size)

  send_buffer = DBLE(send_field)

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_buffer,      &
     &             info,             &
     &             ierror )

end subroutine yac_fput_single_pointset_r4

subroutine yac_fput_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  double precision, intent (in) :: send_field(nbr_hor_points, nbr_pointsets, collection_size)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fput_r8

subroutine yac_fput_r8_ptr ( field_id,         &
     &                       nbr_pointsets,    &
     &                       collection_size,  &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  integer, intent (in)     :: field_id        !< [IN] field identifier
  integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                              !< [IN] returned field
  integer, intent (out)    :: info            !< [OUT] returned info
  integer, intent (out)    :: ierror          !< [OUT] returned error

  integer :: i, j
  type(c_ptr) :: send_field_(nbr_pointsets, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_pointsets
      send_field_(j, i) = c_loc(send_field(j, i)%p(1))
    end do
  end do

  call yac_cput_ptr_ ( field_id,         &
     &                 collection_size,  &
     &                 send_field_,      &
     &                 info,             &
     &                 ierror )

end subroutine yac_fput_r8_ptr

subroutine yac_fput_single_pointset_r8 ( field_id,         &
     &                                   nbr_hor_points,   &
     &                                   collection_size,  &
     &                                   send_field,       &
     &                                   info,             &
     &                                   ierror )

  use mo_yac_finterface, dummy => yac_fput_single_pointset_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  double precision, intent (in) :: send_field(nbr_hor_points, collection_size)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fput_single_pointset_r8

! ---------------------------------- get -------------------------------

subroutine yac_fget_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r4

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  real, intent (inout)          :: recv_field(nbr_hor_points, collection_size)
                                                   !< [INOUT] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: recv_buffer(nbr_hor_points, collection_size)

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_buffer,      &
     &             info,             &
     &             ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fget_r4



subroutine yac_fget_r4_ptr ( field_id,         &
  &                          collection_size,  &
  &                          recv_field,       &
  &                          info,             &
  &                          ierror )

  use mo_yac_finterface, dummy => yac_fget_r4_ptr

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r4_ptr)      :: recv_field(collection_size) !< [INOUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i
  type(yac_r8_ptr) :: recv_field_r8(collection_size)

  do i = 1, collection_size
    allocate(recv_field_r8(i)%p(SIZE(recv_field(i)%p)))
    recv_field_r8(i)%p = dble(recv_field(i)%p)
  end do

  call yac_fget_r8_ptr ( field_id,         &
     &                   collection_size,  &
     &                   recv_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    deallocate(recv_field_r8(i)%p)
  end do

end subroutine yac_fget_r4_ptr

subroutine yac_fget_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  double precision, intent (inout):: recv_field(nbr_hor_points, collection_size)
                                                  !< [INOUT] returned field
  integer, intent (out)         :: info           !< [OUT] returned info
  integer, intent (out)         :: ierror         !< [OUT] returned error

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fget_r8

subroutine yac_fget_r8_ptr ( field_id,         &
                             collection_size,  &
                             recv_field,       &
                             info,             &
                             ierror )

  use mo_yac_finterface, dummy => yac_fget_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  type(yac_r8_ptr)      :: recv_field(collection_size) !< [OUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i
  type(c_ptr) :: recv_field_(collection_size)

  do i = 1, collection_size
    recv_field_(i) = c_loc(recv_field(i)%p(1))
  end do

  call yac_cget ( field_id,         &
     &            collection_size,  &
     &            recv_field_,      &
     &            info,             &
     &            ierror )

end subroutine yac_fget_r8_ptr

! ----------------------------------------------------------------------

subroutine yac_fget_localcomm ( local_communicator, comp_id )

   use mo_yac_finterface, dummy => yac_fget_localcomm

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   integer, intent (in)  :: comp_id            !< [IN] component identifier

   call yac_get_localcomm ( local_communicator, comp_id )

end subroutine yac_fget_localcomm

! ----------------------------------------------------------------------

subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fget_comp_io_proc_step

   integer, intent (in)  :: comp_id            !< [IN] component identifier
   integer, intent (out) :: io_proc_step       !< [OUT] stride between ranks of io processes

   call yac_cget_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fget_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fset_comp_io_proc_step

   integer, intent (in) :: comp_id            !< [IN] component identifier
   integer, intent (in) :: io_proc_step       !< [IN] stride between ranks of io processes

   call yac_cset_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fset_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fget_pair_rootcomm ( local_communicator, comp_name1, comp_name2 )

   use, intrinsic :: iso_c_binding, only : c_null_char
   use mo_yac_finterface, dummy => yac_fget_pair_rootcomm

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
   character(len=*), intent(in) :: comp_name2  !< [IN]  second component name

   call yac_get_pair_rootcomm ( local_communicator, TRIM(comp_name1) // c_null_char, TRIM(comp_name2)  // c_null_char )

end subroutine yac_fget_pair_rootcomm

! ------------------- search/end of definition -------------------------

subroutine yac_fsearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

   use mo_yac_finterface, dummy => yac_fsearch

   integer, intent (in)  :: nbr_comps             !< [IN] number of components
   integer, intent (in)  :: comp_ids(nbr_comps)   !< [IN] list of components IDs
   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (in)  :: field_ids(nbr_fields) !< [IN] list of field IDs
   integer, intent (out) :: ierror                !< [OUT] returned error

   call yac_csearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

end subroutine yac_fsearch

! ------------------------ query routines -----------------------------

subroutine yac_fget_nbr_fields ( nbr_fields )

   use mo_yac_finterface, dummy => yac_fget_nbr_fields

   integer, intent (out) :: nbr_fields !< [OUT] number of coupling fields

   call yac_cget_nbr_fields ( nbr_fields )

end subroutine yac_fget_nbr_fields

! ---------------------------------------------------------------------

subroutine yac_fget_field_ids ( nbr_fields, field_ids )

   use mo_yac_finterface, dummy => yac_fget_field_ids

   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

   call yac_cget_field_ids ( nbr_fields, field_ids )

end subroutine yac_fget_field_ids
! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_id ( field_id, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_id

   integer, intent (in)  :: field_id  !< [IN]  field identifier
   integer, intent (out) :: role      !< [OUT] role for this field
                                      !<       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_cget_role_of_field_by_id ( field_id, role )

end subroutine yac_fget_role_of_field_by_id

! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_name ( field_name, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)    :: field_name !< [IN] field name
   integer, intent (out)             :: role       !< [OUT] role for this field
                                                   !<       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )

   call yac_cget_role_of_field_by_name ( TRIM(field_name) // c_null_char, role )

end subroutine yac_fget_role_of_field_by_name

! ---------------------------------------------------------------------

subroutine yac_fget_field_name_from_id ( field_id, field_name )

   use mo_yac_finterface, dummy => yac_fget_field_name_from_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id   !< [IN]  field identifier
   character (len=*), intent (out) :: field_name !< [OUT] field name

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_field_name_from_id ( field_id, field_name )

   do i = 1, len(field_name)
      if (field_name(i:i) == c_null_char) exit
   end do

   field_name(i:len(field_name)) = ' '

end subroutine yac_fget_field_name_from_id

! ---------------------------------------------------------------------

subroutine yac_fget_field_id_from_name ( field_name, field_id )

   use mo_yac_finterface, dummy => yac_fget_field_id_from_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in) :: field_name !< [IN] field name
   integer, intent (out)          :: field_id   !< [OUT]  field identifier

   call yac_check_strlength ( field_name )

   call yac_cget_field_id_from_name &
                     ( TRIM(field_name) // c_null_char, field_id )

end subroutine yac_fget_field_id_from_name

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_model_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_model_timestep_name &
                     ( TRIM(field_name) // c_null_char, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_name

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_coupling_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_coupling_timestep_name &
                     ( TRIM(field_name) // c_null_char, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_name

! ---------------------------------------------------------------------

subroutine yac_check_strlength ( in_string )

  use mo_yac_finterface

  character (len=*), intent (in)  :: in_string
  character (len=YAC_MAX_CHARLEN) :: message

  if ( LEN_TRIM(in_string) > YAC_MAX_CHARLEN ) then
     message = "string too long"
     print *, "string ", TRIM(in_string), "exceeds length of ", YAC_MAX_CHARLEN
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

end subroutine
