/**
 * @file yac_interface.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "couple_config.h"
#include "xml/config_xml.h"
#include "component.h"
#include "ensure_array_size.h"
#include "event.h"
#include "fields.h"
#include "interpolation.h"
#include "grid.h"
#include "yac_interface.h"
#include "search.h"
#include "utils.h"
#include "version.h"
#include "interp_grid.h"

int const YAC_LOCATION_CELL = CELL;
int const YAC_LOCATION_CORNER = CORNER;
int const YAC_LOCATION_EDGE = EDGE;

static int nbr_defined_components = 0;
static int current_components_size = 0;
static struct component ** list_of_components = NULL;

static size_t nbr_defined_grids = 0;
static size_t list_of_grids_array_size = 0;
static struct basic_grid ** list_of_grids = NULL;

static size_t nbr_interp_fields = 0;
static size_t list_of_interp_fields_array_size = 0;
static struct {
  struct interp_field * field;
  struct basic_grid * grid;
} * list_of_interp_fields = NULL;

static int nbr_defined_fields = 0;
static int current_field_id_size = 0;
static int *list_of_field_ids = NULL;

static char job_start_datetime[MAX_CALENDAR_STR_LEN];
static char job_end_datetime[MAX_CALENDAR_STR_LEN];

/* ---------------------------------------------------------------------- */

void yac_cinit ( const char * xml_filename, const char * schema_filename ) {

  yac_init ( xml_filename, schema_filename );

  /* initialise start- and end dates with setting from xml.
   * This can be overwritten by calls to yac_cdef_datetime.
   */

#ifdef YAC_XML_ENABLED
  strcpy (job_start_datetime, yac_xml_get_start_date());
  strcpy (job_end_datetime, yac_xml_get_end_date());
#endif

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cfinalize () {

  /* free memory */

  for (int i = 0; i < nbr_defined_fields; ++ i ) {
    struct coupling_field * cpl_field = yac_unique_id_to_pointer(list_of_field_ids[i]);
    yac_coupling_field_delete ( cpl_field );
  }
  free(list_of_field_ids);
  nbr_defined_fields = 0;
  current_field_id_size = 0;
  list_of_field_ids = NULL;

  for (int i = 0; i < nbr_defined_grids; ++ i) {
    yac_basic_grid_free(*(list_of_grids[i]));
    free(list_of_grids[i]);
  }
  free(list_of_grids);
  nbr_defined_grids = 0;
  list_of_grids_array_size = 0;
  list_of_grids = NULL;

  for (int i = 0; i < nbr_defined_components; ++i)
    yac_component_delete(list_of_components[i]);
  free(list_of_components);

  /* close MPI */

  yac_finalize();

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_datetime ( const char * start_datetime, const char * end_datetime ) {

#ifdef YAC_XML_ENABLED

  if ( start_datetime == NULL || strlen(start_datetime) == 0 )
      strcpy (job_start_datetime, yac_xml_get_start_date());
  else strcpy (job_start_datetime, start_datetime);

  if ( end_datetime == NULL || strlen(end_datetime) == 0 )
       strcpy (job_end_datetime, yac_xml_get_end_date());
  else strcpy (job_end_datetime, end_datetime );

#endif

  return;
}

/* ---------------------------------------------------------------------- */

char * yac_get_start_datetime () {
  return job_start_datetime;
}

/* ---------------------------------------------------------------------- */

char * yac_get_end_datetime () {
  return job_end_datetime;
}

/* ---------------------------------------------------------------------- */

/* The YAC version number is provided in version.h
 */

char * yac_cget_version () {
  return yac_version;
}

/* ---------------------------------------------------------------------- */

/* internal routine to serve the Fortran interface */

void yac_get_localcomm ( MPI_Fint *comm, int comp_id ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);
  MPI_Comm tmp_comm;

  /* replace with communicator generated within init_comp */

  MPI_Comm_dup(yac_component_get_communicator(comp), &tmp_comm);
  *comm = MPI_Comm_c2f(tmp_comm);
}

/* c interface routine */

void yac_cget_localcomm ( MPI_Comm *comm, int comp_id ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  MPI_Comm_dup(yac_component_get_communicator(comp), comm);
}

/* ---------------------------------------------------------------------- */

void yac_cset_comp_io_proc_step ( int comp_id, int io_proc_step ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  yac_component_set_io_proc_step(comp, io_proc_step);
}

/* ---------------------------------------------------------------------- */

void yac_cget_comp_io_proc_step ( int comp_id, int * io_proc_step ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  *io_proc_step = yac_component_get_io_proc_step(comp);
}

/* ---------------------------------------------------------------------- */

/* internal routine to serve the Fortran interface */

void yac_get_pair_rootcomm ( MPI_Fint *comm, const char * comp_name1, const char * comp_name2 ) {
#ifdef YAC_XML_ENABLED
  *comm = MPI_Comm_c2f(yac_get_pair_rootcomm_ ( comp_name1, comp_name2 ));
#else
  *comm = MPI_Comm_c2f(MPI_COMM_NULL);
#endif
}

/* c interface routine */

void yac_cget_pair_rootcomm ( MPI_Comm *comm, const char * comp_name1, const char * comp_name2 ) {

#ifdef YAC_XML_ENABLED
  *comm = yac_get_pair_rootcomm_ ( comp_name1, comp_name2 );
#else
  *comm = MPI_Comm_c2f(MPI_COMM_NULL);
#endif
}

/* ---------------------------------------------------------------------- */

void yac_cdef_comp ( char const * comp_name, int * comp_id ) {

  struct component * component = yac_component_new(comp_name);

  *comp_id = yac_pointer_to_unique_id(component);

  int redirect = yac_xml_redirect_stdout();

  if ( redirect ) {
      int ierr;
      int lenstr = (int) strlen(comp_name);
      MPI_Comm comm =
        yac_component_get_communicator(yac_unique_id_to_pointer(*comp_id));
      int rank, size;
      yac_mpi_call(MPI_Comm_rank(comm, &rank), comm);
      yac_mpi_call(MPI_Comm_size(comm, &size), comm);
      if ( redirect == 1 )
        yac_redirstdout ( comp_name, lenstr, 0, rank, size, &ierr ); // all redirect
      else
        yac_redirstdout ( comp_name, lenstr, 1, rank, size, &ierr ); // only root redirect
  }

  nbr_defined_components++;
  ENSURE_ARRAY_SIZE(list_of_components, current_components_size, nbr_defined_components);
  list_of_components[nbr_defined_components-1] = component;
}

/* ---------------------------------------------------------------------- */

void
yac_cdef_comps ( char const ** comp_names, int num_comps, int * comp_ids ) {

  struct component * components[num_comps];

  yac_components_new(comp_names, num_comps, components);

  for (int i = 0; i < num_comps; ++i)
    comp_ids[i] = yac_pointer_to_unique_id(components[i]);

  int redirect = yac_xml_redirect_stdout();

  if ( redirect ) {
      int ierr;
      int lenstr = (int) strlen(comp_names[0]);
      MPI_Comm comm =
        yac_component_get_communicator(yac_unique_id_to_pointer(comp_ids[0]));
      int rank, size;
      yac_mpi_call(MPI_Comm_rank(comm, &rank), comm);
      yac_mpi_call(MPI_Comm_size(comm, &size), comm);
      if ( redirect == 1 )
        yac_redirstdout ( comp_names[0], lenstr, 0, rank, size, &ierr ); // all redirect
      else
        yac_redirstdout ( comp_names[0], lenstr, 1, rank, size, &ierr ); // only root redirect
  }

  ENSURE_ARRAY_SIZE(list_of_components, current_components_size,
                    nbr_defined_components + num_comps);
  for (int i = 0; i < num_comps; ++i, ++nbr_defined_components)
    list_of_components[nbr_defined_components] = components[i];
}

/* ---------------------------------------------------------------------- */

static size_t get_data_size(
  struct basic_grid * grid, enum yac_location location) {

  switch (location) {
    case (CELL): return grid->data.num_cells;
    case (CORNER): return grid->data.num_vertices;
    case (EDGE): return grid->data.num_edges;
    default:
      yac_internal_abort_message(
        "ERROR(get_data_size): invalid location", __FILE__, __LINE__);
      return SIZE_MAX;
  };
}

void yac_cdef_points_reg2d ( int const grid_id,
                             int const *nbr_points,
                             int const located,
                             double const *x_points,
                             double const *y_points,
                             int *point_id ) {

  if (nbr_interp_fields > INT_MAX)
    yac_internal_abort_message(
      "ERROR(yac_cdef_points_reg2d): too many fields", __FILE__, __LINE__);

  ENSURE_ARRAY_SIZE(
    list_of_interp_fields, list_of_interp_fields_array_size,
    nbr_interp_fields + 1);

  *point_id = nbr_interp_fields;

  struct interp_field * field =
    (list_of_interp_fields[nbr_interp_fields].field =
       xmalloc(1 * sizeof(*(list_of_interp_fields[nbr_interp_fields].field))));
  struct basic_grid * grid =
    (list_of_interp_fields[nbr_interp_fields].grid =
       yac_unique_id_to_pointer(grid_id));
  ++nbr_interp_fields;

  enum yac_location location = yac_get_location(located);
  size_t num_points = get_data_size(grid, location);

  if (((size_t)nbr_points[0] * (size_t)nbr_points[1]) != num_points)
    yac_internal_abort_message(
      "ERROR(yac_cdef_points_reg2d): nbr_points does not match with grid",
      __FILE__, __LINE__);

  field->location = location;
  field->coordinates = xmalloc(num_points * sizeof(*(field->coordinates)));
  field->mask = NULL;

  num_points = 0;
  for (int i = 0; i < nbr_points[1]; ++i)
    for (int j = 0; j < nbr_points[0]; ++j)
      LLtoXYZ(x_points[j], y_points[i], field->coordinates[num_points++]);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_points_unstruct ( int const grid_id,
                                int const nbr_points,
                                int const located,
                                double const *x_points,
                                double const *y_points,
                                int *point_id ) {

  if (nbr_interp_fields > INT_MAX)
    yac_internal_abort_message(
      "ERROR(yac_cdef_points_unstruct): too many fields", __FILE__, __LINE__);

  ENSURE_ARRAY_SIZE(
    list_of_interp_fields, list_of_interp_fields_array_size,
    nbr_interp_fields + 1);

  *point_id = nbr_interp_fields;

  struct interp_field * field =
    (list_of_interp_fields[nbr_interp_fields].field =
       xmalloc(1 * sizeof(*(list_of_interp_fields[nbr_interp_fields].field))));
  struct basic_grid * grid =
    (list_of_interp_fields[nbr_interp_fields].grid =
       yac_unique_id_to_pointer(grid_id));
  ++nbr_interp_fields;

  enum yac_location location = yac_get_location(located);
  size_t num_points = get_data_size(grid, location);

  if ((size_t)nbr_points != num_points)
    yac_internal_abort_message(
      "ERROR(yac_cdef_points_unstruct): nbr_points does not match with grid",
      __FILE__, __LINE__);

  field->location = location;
  field->coordinates = xmalloc(num_points * sizeof(*(field->coordinates)));
  field->mask = 0;

  num_points = 0;
  for (int i = 0; i < nbr_points; ++i)
    LLtoXYZ(x_points[i], y_points[i], field->coordinates[num_points++]);
}

/* ---------------------------------------------------------------------- */

void yac_cset_mask ( int const * is_valid,
                     int points_id ) {

  if ((points_id < 0) || (points_id >= nbr_interp_fields))
    yac_internal_abort_message(
      "ERROR(yac_cdef_mask): invalid points_id", __FILE__, __LINE__);

  struct basic_grid * grid = list_of_interp_fields[points_id].grid;
  int ** mask = &(list_of_interp_fields[points_id].field->mask);

  if (*mask != NULL)
    yac_internal_abort_message(
      "ERROR(yac_cdef_mask): mask already defined", __FILE__, __LINE__);

  size_t num_points =
    get_data_size(grid, list_of_interp_fields[points_id].field->location);

  *mask = xmalloc(num_points * sizeof(**mask));
  memcpy(*mask, is_valid, num_points * sizeof(**mask));
}

/* ---------------------------------------------------------------------- */

void yac_cdef_field ( char const * name,
                      int const component_id,
                      int const grid_id,
                      int const * point_ids,
                      int const num_pointsets,
                      int * field_id ) {

  struct basic_grid * grid = yac_unique_id_to_pointer(grid_id);

  struct interp_field * interp_fields = NULL;

  if (point_ids != NULL) {
    interp_fields = xmalloc((size_t)num_pointsets * sizeof(*interp_fields));
    for (int i = 0; i < num_pointsets; ++i) {
      if (grid != list_of_interp_fields[point_ids[i]].grid)
        yac_internal_abort_message(
          "ERROR(yac_cdef_field): grid of point_ids do not match",
          __FILE__, __LINE__);
      interp_fields[i] = *(list_of_interp_fields[point_ids[i]].field);
    }
  }

  struct coupling_field * cpl_field =
    yac_coupling_field_new(
      name, component_id, grid, interp_fields, num_pointsets);
  free(interp_fields);

  *field_id = yac_pointer_to_unique_id(cpl_field);

  nbr_defined_fields++;
  ENSURE_ARRAY_SIZE(list_of_field_ids, current_field_id_size, nbr_defined_fields);
  list_of_field_ids[nbr_defined_fields-1] = *field_id;
}

/* ---------------------------------------------------------------------- */

void yac_cget_ ( int const field_id,
                 int const collection_size,
                 double *recv_field,   // recv_field[collection_size][Points]
                 int    *info,
                 int    *ierr ) {


  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field * cpl_field = yac_unique_id_to_pointer(field_id);

  if (yac_coupling_field_get_num_interp_fields(cpl_field) != 1)
    yac_internal_abort_message(
      "ERROR(yac_cget_): invalid number of interpolation fields "
      "(should be one)", __FILE__, __LINE__);

  enum yac_location location =
    yac_coupling_field_get_interp_fields(cpl_field)->location;
  size_t num_points =
    yac_coupling_field_get_data_size(cpl_field, location);

  double ** recv_field_ = xmalloc(collection_size * sizeof(*recv_field_));

  for (int i = 0; i < collection_size; ++i) {
    recv_field_[i] = recv_field;
    recv_field += num_points;
  }

  yac_cget ( field_id, collection_size, recv_field_, info, ierr );

  free(recv_field_);
}

/* ---------------------------------------------------------------------- */

void yac_cget ( int const field_id,
                int collection_size,
                double ** recv_field, // recv_field[collection_size][Points]
                int    *info,
                int    *ierr ) {

  *ierr = -1;
  *info = NONE;

#ifndef YAC_XML_ENABLED
  yac_internal_abort_message("ERROR: Compiled without xml support! "
                             "Coupling is not possible.", __FILE__, __LINE__);
#endif

  struct coupling_field * cpl_field = yac_unique_id_to_pointer(field_id);

  if (yac_get_coupling_field_exchange_type(cpl_field) != TARGET) return;

  /* --------------------------------------------------------------------
     Check for restart and coupling events
     -------------------------------------------------------------------- */

  int event_id = yac_get_coupling_field_get_op_event_id(cpl_field);

  int action = yac_event_check ( event_id );

  /* --------------------------------------------------------------------
     add one model time step (as provided in coupling.xml) to the current event date
     -------------------------------------------------------------------- */

  yac_event_update ( event_id );

  /* ------------------------------------------------------------------
     return in case we are already beyond the end of the run
     ------------------------------------------------------------------ */

  if ( action == OUT_OF_BOUND ) {

      fprintf (stderr, "WARNING: yac get action is beyond end of run date!\n");
      fprintf (stdout, "WARNING: yac get action is beyond end of run date!\n");

      *info = action;
      *ierr = 0;

      return;
  }

  /* --------------------------------------------------------------------
     start actions
     -------------------------------------------------------------------- */

  if ( action == NONE ) {
    *info = action;
    *ierr = 0;

    return;
  }

  if ( collection_size != yac_get_coupling_field_collection_size(cpl_field) )
    yac_internal_abort_message(
      "ERROR: collection size does not match with xml.", __FILE__, __LINE__);

  if ( action == RESTART ) action = GET_FOR_RESTART;

  struct interpolation * interpolation =
    yac_get_coupling_field_get_op_interpolation(cpl_field);

  yac_interpolation_execute_get(interpolation, recv_field);

  *info = action;
  *ierr = 0;

  return;
}

/* ---------------------------------------------------------------------- */


void yac_cput_ ( int const field_id,
                 int const collection_size,
                 double *send_field,   // send_field[collection_size]
                                       //           [nPointSets]
                                       //           [Points]
                 int    *info,
                 int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field * cpl_field = yac_unique_id_to_pointer(field_id);

  size_t num_interp_fields =
    yac_coupling_field_get_num_interp_fields(cpl_field);

  size_t * num_points = xmalloc(num_interp_fields * sizeof(*num_points));

  for (size_t i = 0; i < num_interp_fields; i++) {

    enum yac_location location =
      yac_coupling_field_get_interp_fields(cpl_field)[i].location;
    num_points[i] = yac_coupling_field_get_data_size(cpl_field, location);
  }

  double *** send_field_ = xmalloc(collection_size * sizeof(*send_field_));

  for (int i = 0; i < collection_size; ++i) {

    send_field_[i] = xmalloc(num_interp_fields * sizeof(**send_field_));

    for (size_t j = 0; j < num_interp_fields; ++j) {

      send_field_[i][j] = send_field;
      send_field += num_points[j];
    }
  }

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);
  free(num_points);
}

/* ---------------------------------------------------------------------- */


void yac_cput_ptr_ ( int const field_id,
                     int const collection_size,
                     double ** send_field,   // send_field[collection_size *
                                             //            nPointSets][Points]
                     int    *info,
                     int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field * cpl_field = yac_unique_id_to_pointer(field_id);

  size_t num_interp_fields = yac_coupling_field_get_num_interp_fields(cpl_field);

  size_t * num_points = xmalloc(num_interp_fields * sizeof(*num_points));

  for (size_t i = 0; i < num_interp_fields; i++) {

    enum yac_location location =
      yac_coupling_field_get_interp_fields(cpl_field)[i].location;
    num_points[i] = yac_coupling_field_get_data_size(cpl_field, location);
  }

  double *** send_field_ = xmalloc(collection_size * sizeof(*send_field_));

  for (int i = 0, k = 0; i < collection_size; ++i) {

    send_field_[i] = xmalloc(num_interp_fields * sizeof(**send_field_));
    for (size_t j = 0; j < num_interp_fields; ++j, ++k)
      send_field_[i][j] = send_field[k];
  }

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);
  free(num_points);

}

/* ---------------------------------------------------------------------- */

void yac_cput ( int const field_id,
                int const collection_size,
                double *** const send_field,   // send_field[collection_size]
                                        //           [nPointSets][Points]
                int    *info,
                int    *ierr ) {

  *info = NONE;
  *ierr = -1;

#ifndef YAC_XML_ENABLED
  yac_internal_abort_message("ERROR: Compiled without xml support! "
                             "Coupling is not possible.", __FILE__, __LINE__);
#endif

  static int action;

  struct coupling_field * cpl_field = yac_unique_id_to_pointer(field_id);

  size_t num_interp_fields = yac_coupling_field_get_num_interp_fields(cpl_field);
  unsigned * num_points = NULL;

  if (yac_get_coupling_field_exchange_type(cpl_field) != SOURCE)
    return;

  for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(cpl_field);
       ++put_idx) {

    /* ------------------------------------------------------------------
       Check for restart and coupling events
       ------------------------------------------------------------------ */

    int event_id = yac_get_coupling_field_put_op_event_id(cpl_field, put_idx);

    action = yac_event_check ( event_id );

    /* ------------------------------------------------------------------
       add one model time step (as provided in coupling.xml) to the current event date
       ------------------------------------------------------------------ */

    yac_event_update ( event_id );

    /* ------------------------------------------------------------------
       return in case we are already beyond the end of the run
       ------------------------------------------------------------------ */

    if ( action == OUT_OF_BOUND ) {

        fprintf (stderr, "WARNING: put action is beyond end of run date!\n");
        fprintf (stdout, "WARNING: put action is beyond end of run date!\n");

        *info = action;
        *ierr = 0;

        return;
    }

    /* ------------------------------------------------------------------
       start actions
       ------------------------------------------------------------------ */

    int time_operation = yac_get_event_time_operation(event_id);

    if ( action == NONE && time_operation == INSTANT ) continue;

    /* ------------------------------------------------------------------
       If it is time for restart, set appropriate flags and continue
       ------------------------------------------------------------------ */

    if ( action == RESTART ) {
      *info = MAX (*info, PUT_FOR_RESTART);
      *ierr = 0;
    }

    struct interpolation * interpolation =
      yac_get_coupling_field_put_op_interpolation(cpl_field, put_idx);

    if ( collection_size != yac_get_coupling_field_collection_size(cpl_field) )
      yac_internal_abort_message(
        "ERROR: collection size does not match with xml.", __FILE__, __LINE__ );

    /* ------------------------------------------------------------------
       First deal with instant sends
       ------------------------------------------------------------------ */

    if ( time_operation == INSTANT ) {

      yac_interpolation_execute_put ( interpolation, send_field );

      *info = MAX (*info, COUPLING);
      *ierr = 0;

      continue;
    }

    /* ------------------------------------------------------------------
       Accumulation & Averaging
       ------------------------------------------------------------------ */

    /* prepare for accumulation */

    if (num_points == NULL) {

      num_points = xmalloc(num_interp_fields * sizeof(*num_points));

      for (size_t i = 0; i < num_interp_fields; i++)
        num_points[i] =
          yac_coupling_field_get_data_size(
            cpl_field,
            yac_coupling_field_get_interp_fields(cpl_field)[i].location);
    }

    /* accumulate data */

    double *** send_field_acc =
      yac_get_coupling_field_put_op_send_field_acc(cpl_field, put_idx);

    int time_accumulation_count =
      yac_get_coupling_field_put_op_time_accumulation_count(cpl_field, put_idx);

    time_accumulation_count++;

    for (int h = 0; h < collection_size; h++)
      for (size_t i = 0; i < num_interp_fields; i++)
        for (size_t j = 0; j < num_points[i]; j++ )
          send_field_acc[h][i][j] += send_field[h][i][j];

  /* --------------------------------------------------------------------
     Check whether we have to perform the coupling in this call
     -------------------------------------------------------------------- */

    if ( action == NONE ) {
      yac_set_coupling_field_put_op_time_accumulation_count(
        cpl_field, put_idx, time_accumulation_count);
      continue;
    }

  /* --------------------------------------------------------------------
     Average data if required
     -------------------------------------------------------------------- */

    if ( time_operation == TIME_AVERAGE ) {
      double weight = 1.0 / time_accumulation_count;

      for (int h = 0; h < collection_size; ++h)
        for (size_t i = 0; i < num_interp_fields; i++)
          for (size_t j = 0; j < num_points[i]; j++)
            send_field_acc[h][i][j] *= weight;
    }

  /* --------------------------------------------------------------------
     Apply interpolation and send away
     -------------------------------------------------------------------- */

      yac_interpolation_execute_put ( interpolation, send_field_acc );

      *info = MAX (*info, COUPLING);

  /* --------------------------------------------------------------------
     Initialise memory for the next accumulation
     -------------------------------------------------------------------- */

    yac_init_coupling_field_put_op_send_field_acc(cpl_field, put_idx, 0.0);
    yac_set_coupling_field_put_op_time_accumulation_count(cpl_field, put_idx, 0);

  }

  free (num_points);

  *ierr = 0;

  return;
}

/* ---------------------------------------------------------------------- */

void yac_csearch ( int nbr_comps,
                   int * comp_ids,
                   int nbr_fields,
                   int * field_ids,
                   int * ierror ) {

  struct component ** components = xmalloc(nbr_comps * sizeof(*components));
  for (int i = 0; i < nbr_comps; ++i)
    components[i] = yac_unique_id_to_pointer(comp_ids[i]);
  struct coupling_field ** fields = xmalloc(nbr_fields * sizeof(*fields));
  for (int i = 0; i < nbr_fields; ++i)
    fields[i] = yac_unique_id_to_pointer(field_ids[i]);
  yac_setup_couple_config(nbr_comps, components, nbr_fields, fields);

  free(components);
  free(fields);

  *ierror = 0;
}

/* ----------------------------------------------------------------------
                   query functions
   ----------------------------------------------------------------------*/

void yac_cget_nbr_fields ( int *nbr_fields ) {
  *nbr_fields = nbr_defined_fields;
  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_ids ( int nbr_fields, int *field_ids ) {
  if ( nbr_fields != nbr_defined_fields )
    yac_internal_abort_message ( "ERROR: nbr of fields does not match.", __FILE__, __LINE__ );
  for ( int i = 0; i < nbr_defined_fields; ++i )
    field_ids[i] = list_of_field_ids[i];
  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_id_from_name ( const char *field_name,
                                   int *field_id ) {

  int i;

  for ( i = 0; i < nbr_defined_fields; ++i ) {
    struct coupling_field * field = yac_unique_id_to_pointer(list_of_field_ids[i]);
    if ( !strcmp(yac_get_coupling_field_name(field),field_name) ) {
      *field_id = list_of_field_ids[i];
      break;
    }
  }

  if ( i == nbr_defined_fields )
    yac_internal_abort_message ( "ERROR: no matching field name!", __FILE__, __LINE__ );

}

/* ---------------------------------------------------------------------- */

void yac_cget_field_name_from_id ( int field_id,
                                   char *field_name ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  strcpy(field_name,yac_get_coupling_field_name(field));

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_id ( int field_id, int *role ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  *role = yac_get_coupling_field_exchange_type(field);

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_name ( char *field_name, int *role ) {

  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  *role = yac_get_coupling_field_exchange_type(field);

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string ) {
  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );
  yac_cget_coupling_timestep_id ( field_id, timestep_string );

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_id ( int field_id,
                                     char *timestep_string ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  int event_id;

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  int role = yac_get_coupling_field_exchange_type(field);

  if ( role == SOURCE )
    for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(field);
         ++put_idx)
      event_id = yac_get_coupling_field_put_op_event_id(field, put_idx);

  else if ( role == TARGET )
    event_id = yac_get_coupling_field_get_op_event_id(field);

  char my_string[MAX_TIMEDELTA_STR_LEN];

  if ( role == SOURCE || role == TARGET )
    strcpy(timestep_string,yac_get_event_coupling_timestep(event_id,my_string));

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string ) {
  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );
  yac_cget_model_timestep_id ( field_id, timestep_string );

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  int event_id;

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  int role = yac_get_coupling_field_exchange_type(field);

  if ( role == SOURCE )
    for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(field);
         ++put_idx)
      event_id = yac_get_coupling_field_put_op_event_id(field, put_idx);

  else if ( role == TARGET )
    event_id = yac_get_coupling_field_get_op_event_id(field);

  char my_string[MAX_TIMEDELTA_STR_LEN];

  if ( role == SOURCE || role == TARGET )
    strcpy(timestep_string,yac_get_event_model_timestep(event_id,my_string));

  return;
}

void yac_cdef_grid_reg2d(
  const char * grid_name, int nbr_vertices[2], int cyclic[2],
  double *x_vertices, double *y_vertices, int *grid_id) {

  for (size_t i = 0; i < nbr_defined_grids; ++i)
    if (!strcmp(list_of_grids[i]->name, grid_name))
      yac_internal_abort_message(
        "ERROR(yac_cdef_grid_reg2d): multiple definitions of grid with "
        "identical name", __FILE__, __LINE__);

  size_t nbr_vertices_size_t[2] =
    {(size_t)nbr_vertices[0], (size_t)nbr_vertices[1]};

  struct basic_grid * grid = xmalloc(1 * sizeof(*grid));

  size_t grid_name_len = strlen(grid_name) + 1;
  grid->name = xmalloc(grid_name_len);
  strncpy(grid->name, grid_name, grid_name_len);
  grid->data =
    yac_generate_basic_grid_data_reg_2d(
      nbr_vertices_size_t, cyclic, x_vertices, y_vertices);
  *grid_id = yac_pointer_to_unique_id(grid);

  nbr_defined_grids++;
  ENSURE_ARRAY_SIZE(list_of_grids, list_of_grids_array_size, nbr_defined_grids);
  list_of_grids[nbr_defined_grids-1] = grid;
}

void yac_cdef_grid_unstruct(
  const char * grid_name, int nbr_vertices,
  int nbr_cells, int *num_vertices_per_cell, double *x_vertices,
  double *y_vertices, int *cell_to_vertex, int *grid_id) {

  for (size_t i = 0; i < nbr_defined_grids; ++i)
    if (!strcmp(list_of_grids[i]->name, grid_name))
      yac_internal_abort_message(
        "ERROR(yac_cdef_grid_unstruct): multiple definitions of grid with "
        "identical name", __FILE__, __LINE__);

  struct basic_grid * grid = xmalloc(1 * sizeof(*grid));

  size_t grid_name_len = strlen(grid_name) + 1;
  grid->name = xmalloc(grid_name_len);
  strncpy(grid->name, grid_name, grid_name_len);
  grid->data =
    yac_generate_basic_grid_data_unstruct(
      (size_t)nbr_vertices, (size_t)nbr_cells, num_vertices_per_cell,
      x_vertices, y_vertices, cell_to_vertex);
  *grid_id = yac_pointer_to_unique_id(grid);

  nbr_defined_grids++;
  ENSURE_ARRAY_SIZE(list_of_grids, list_of_grids_array_size, nbr_defined_grids);
  list_of_grids[nbr_defined_grids-1] = grid;
}

void yac_cset_global_index(
  yac_int const * global_index, int location, int grid_id) {

  struct basic_grid * grid = yac_unique_id_to_pointer(grid_id);

  yac_int ** grid_global_ids;
  size_t count;

  switch (location) {
    case (CELL): {
      grid_global_ids = &(grid->data.cell_ids);
      count = grid->data.num_cells;
      break;
    }
    case (CORNER): {
      grid_global_ids = &(grid->data.vertex_ids);
      count = grid->data.num_vertices;
      break;
    }
    case (EDGE): {
      grid_global_ids = &(grid->data.edge_ids);
      count = grid->data.num_edges;
      break;
    }
    default: {
      grid_global_ids = NULL;
      yac_internal_abort_message(
        "ERROR(yac_cset_global_index): invalid location", __FILE__, __LINE__);
    }
  }

  size_t global_ids_size = count * sizeof(**grid_global_ids);
  if (*grid_global_ids == NULL) *grid_global_ids = xmalloc(global_ids_size);
  memcpy(*grid_global_ids, global_index, global_ids_size);
}

void yac_cset_core_mask(
  int const * is_core, int location, int grid_id) {

  struct basic_grid * grid = yac_unique_id_to_pointer(grid_id);

  int ** grid_core_mask;
  size_t count;

  switch (location) {
    case (CELL): {
      grid_core_mask = &(grid->data.core_cell_mask);
      count = grid->data.num_cells;
      break;
    }
    case (CORNER): {
      grid_core_mask = &(grid->data.core_vertex_mask);
      count = grid->data.num_vertices;
      break;
    }
    case (EDGE): {
      grid_core_mask = &(grid->data.core_edge_mask);
      count = grid->data.num_edges;
      break;
    }
    default: {
      grid_core_mask = NULL;
      yac_internal_abort_message(
        "ERROR(yac_cset_core_mask): invalid location", __FILE__, __LINE__);
    }
  }

  size_t core_mask_size = count * sizeof(**grid_core_mask);
  if (*grid_core_mask == NULL) *grid_core_mask = xmalloc(core_mask_size);
  memcpy(*grid_core_mask, is_core, core_mask_size);
}

size_t yac_get_grid_size ( int located,
                           int grid_id ) {

  struct basic_grid * grid = yac_unique_id_to_pointer(grid_id);
  enum yac_location location = yac_get_location(located);

  return get_data_size(grid, location);
}

size_t yac_get_points_size ( int points_id  ) {

  struct interp_field * field = list_of_interp_fields[points_id].field;
  struct basic_grid * grid = list_of_interp_fields[points_id].grid;

  return get_data_size(grid, field->location);
}
