/**
 * @file test_interp_weights_parallel.c
 *
 * @copyright Copyright  (C)  2019 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <unistd.h>

#include "tests.h"
#include "test_common.h"
#include "geometry.h"
#include "read_icon_grid.h"
#include "interp_method.h"
#include "interp_method_file.h"
#include "interp_weights.h"
#include "yac_mpi.h"
#include "dist_grid_utils.h"

#if defined(YAC_NETCDF_ENABLED)

#include <mpi.h>
#include <yaxt.h>
#include <netcdf.h>

static void get_basic_grid_data(
  char * filename, size_t * num_cells, size_t * num_vertices,
  size_t * num_edges);
static int check_direct_results(
  struct basic_grid_data * grid, int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, size_t collection_size,
  double *** src_data, double ** tgt_data);

int main(void) {

  MPI_Init(NULL, NULL);

  xt_initialize(MPI_COMM_WORLD);

  int comm_rank, comm_size;
  yac_mpi_call(MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank), MPI_COMM_WORLD);
  yac_mpi_call(MPI_Comm_size(MPI_COMM_WORLD, &comm_size), MPI_COMM_WORLD);

  if (comm_size < 4) {
    PUT_ERR("too few processes\n");
    xt_finalize();
    MPI_Finalize();
    return TEST_EXIT_CODE;
  }

  enum interp_weights_reorder_type reorder_types[] = {
    MAPPING_ON_SRC,
    MAPPING_ON_TGT
  };
  size_t reorder_type_count =
    sizeof(reorder_types) / sizeof(reorder_types[0]);

  {
    int is_src = comm_rank < (comm_size / 2);

    MPI_Comm local_grid_comm;
    int local_grid_comm_size;
    yac_mpi_call(
      MPI_Comm_split(
        MPI_COMM_WORLD, is_src, 0, &local_grid_comm), MPI_COMM_WORLD);
    yac_mpi_call(
      MPI_Comm_size(
        local_grid_comm, &local_grid_comm_size), local_grid_comm);

    char * filenames[] ={"icon_grid_R02B02.nc", "icon_grid_R02B03_G.nc"};
    struct basic_grid_data * grids[2];
    int grid_ids[2] = {2, 3};

    grids[0] =
      read_icon_grid_information_parallel2(
        filenames[is_src], local_grid_comm, MIN(local_grid_comm_size, 2));
    grids[1] = &basic_grid_data_empty;

    struct dist_grid_pair * grid_pair =
      yac_dist_grid_pair_new(
        grids[0], grid_ids[is_src], grids[1], grid_ids[is_src^1],
        MPI_COMM_WORLD);

    struct interp_field src_fields[] =
      {{.location = CELL, .coordinates = NULL, .mask = NULL}};
    size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
    struct interp_field tgt_field =
      {.location = CELL, .coordinates = NULL, .mask = NULL};

    struct interp_grid * interp_grid =
      yac_interp_grid_new(grid_pair, grid_ids[1], grid_ids[0],
                          num_src_fields, src_fields, tgt_field);

    struct interp_weights * weights = yac_interp_weights_new(interp_grid);

    size_t num_tgt_cells, num_tgt_vertices, num_tgt_edges;
    get_basic_grid_data(
      filenames[0], &num_tgt_cells, &num_tgt_vertices, &num_tgt_edges);

    {
      size_t count;
      size_t * tgt_points;
      yac_interp_grid_get_tgt_points(interp_grid, &tgt_points, &count);
      yac_int * tgt_global_ids = xmalloc(count * sizeof(*tgt_global_ids));
      size_t * size_t_buffer = xmalloc(2 * count * sizeof(*size_t_buffer));
      size_t * odd_tgt_points = size_t_buffer;
      size_t * even_tgt_points = size_t_buffer + count;
      yac_interp_grid_get_tgt_global_ids(
        interp_grid, tgt_points, count, tgt_global_ids);

      size_t odd_count = 0;
      size_t even_count = 0;
      for (size_t i = 0; i < count; ++i) {
        if (tgt_global_ids[i] & 1) odd_tgt_points[odd_count++] = tgt_points[i];
        else even_tgt_points[even_count++] = tgt_points[i];
      }

      struct remote_points odd_tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, odd_tgt_points, odd_count),
        .count = odd_count};
      struct remote_points even_tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, even_tgt_points, even_count),
        .count = even_count};

      yac_interp_weights_add_fixed(weights, &odd_tgts, -1.0);
      yac_interp_weights_add_fixed(weights, &even_tgts, -2.0);

      free(odd_tgts.data);
      free(even_tgts.data);
      free(size_t_buffer);
      free(tgt_global_ids);
      free(tgt_points);
    }

    for (size_t i = 0; i < reorder_type_count; ++i) {

      struct interpolation * interpolation =
        yac_interp_weights_get_interpolation(weights, reorder_types[i], 1);

      double * field_data =
        xmalloc(grids[0]->num_cells * sizeof(*field_data));
      if (is_src)
        for (size_t j = 0; j < grids[0]->num_cells; ++j)
          field_data[j] = (double)1;
      else
        for (size_t j = 0; j < grids[0]->num_cells; ++j)
          field_data[j] = (double)-3;

      double ** src_fields, * tgt_field;
      if (is_src) {
        src_fields = &field_data;
        tgt_field = NULL;
      } else {
        src_fields = NULL;
        tgt_field = field_data;
      }

      yac_interpolation_execute(interpolation, &src_fields, &tgt_field);

      if (!is_src) {
        for (size_t j = 0; j < grids[0]->num_cells; ++j) {
          if (grids[0]->core_cell_mask[j]) {
            if (grids[0]->cell_ids[j] & 1) {
              if (tgt_field[j] != -1.0)
                PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                        "invalid interpolation result");
            } else {
              if (tgt_field[j] != -2.0)
                PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                        "invalid interpolation result");
            }
          } else {
            if (tgt_field[j] != -3.0)
              PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                      "invalid interpolation result");
          }
        }
      }

      free(field_data);
      yac_interpolation_delete(interpolation);
    }

    yac_interp_weights_delete(weights);
    yac_interp_grid_delete(interp_grid);
    yac_dist_grid_pair_delete(grid_pair);
    yac_mpi_call(MPI_Comm_free(&local_grid_comm), MPI_COMM_WORLD);

    for (int i = 0; i < 2; ++i) yac_basic_grid_data_delete(grids[i]);
  }

  {
    int is_src = comm_rank < (comm_size / 2);

    MPI_Comm local_grid_comm;
    int local_grid_comm_size;
    yac_mpi_call(
      MPI_Comm_split(
        MPI_COMM_WORLD, is_src, 0, &local_grid_comm), MPI_COMM_WORLD);
    yac_mpi_call(
      MPI_Comm_size(
        local_grid_comm, &local_grid_comm_size), local_grid_comm);

    char * filenames[] ={"icon_grid_R02B02.nc", "icon_grid_R02B02.nc"};
    struct basic_grid_data * grids[2];
    int grid_ids[2] = {2, 3};

    grids[0] =
      read_icon_grid_information_parallel2(
        filenames[is_src], local_grid_comm, MIN(local_grid_comm_size, 2));
    grids[1] = &basic_grid_data_empty;

    struct dist_grid_pair * grid_pair =
      yac_dist_grid_pair_new(
        grids[0], grid_ids[is_src], grids[1], grid_ids[is_src^1],
        MPI_COMM_WORLD);

    struct interp_field src_fields[] =
      {{.location = CELL, .coordinates = NULL, .mask = NULL}};
    size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
    struct interp_field tgt_field =
      {.location = CELL, .coordinates = NULL, .mask = NULL};

    struct interp_grid * interp_grid =
      yac_interp_grid_new(grid_pair, grid_ids[1], grid_ids[0],
                          num_src_fields, src_fields, tgt_field);

    struct interp_weights * weights = yac_interp_weights_new(interp_grid);

    size_t num_tgt_cells, num_tgt_vertices, num_tgt_edges;
    get_basic_grid_data(
      filenames[0], &num_tgt_cells, &num_tgt_vertices, &num_tgt_edges);

    {
      // get all local target points
      size_t count;
      size_t * tgt_points;
      yac_interp_grid_get_tgt_points(interp_grid, &tgt_points, &count);

      // get global ids of local target points
      yac_int * tgt_global_ids = xmalloc(count * sizeof(*tgt_global_ids));
      yac_interp_grid_get_tgt_global_ids(
        interp_grid, tgt_points, count, tgt_global_ids);

      // get local ids of matching source points
      size_t * src_points = xmalloc(count * sizeof(*src_points));
      yac_intep_grid_src_global_to_local(
        interp_grid, 0, tgt_global_ids, count, src_points);
      free(tgt_global_ids);

      struct remote_points tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, tgt_points, count),
        .count = count};

      struct remote_point * srcs =
        yac_interp_grid_get_src_remote_points(
          interp_grid, 0, src_points, count);

      yac_interp_weights_add_direct(weights, &tgts, srcs);

      free(tgts.data);
      free(tgt_points);
      free(srcs);
      free(src_points);
    }

    double *** src_data, ** tgt_data;

    if (is_src) {
      src_data = xmalloc(4 * sizeof(*src_data));
      for (size_t i = 0; i < 4; ++i)
        src_data[i] = xmalloc(1 * sizeof(**src_data));

      for (size_t collection_idx = 0; collection_idx < 4;
           ++collection_idx) {
        double * src_field =
          xmalloc(grids[0]->num_cells * sizeof(*src_field));
        for (size_t i = 0; i < grids[0]->num_cells; ++i)
          src_field[i] =
            (grids[0]->core_cell_mask[i])?
              (double)(grids[0]->cell_ids[i]):(-1.0);
        src_data[collection_idx][0] = src_field;
      }
      tgt_data = NULL;
    } else {
      tgt_data = xmalloc(4 * sizeof(*tgt_data));
      for (size_t collection_idx = 0; collection_idx < 4;
           ++collection_idx)
        tgt_data[collection_idx] =
          xmalloc(grids[0]->num_cells * sizeof(**tgt_data));
      src_data = NULL;
    }

    // checking with different reorder types and collection sizes
    for (size_t reorder_type = 0; reorder_type < reorder_type_count;
         ++reorder_type)
      for (size_t collection_size = 1; collection_size <= 4;
           collection_size *= 2)
          if (check_direct_results(
                grids[0], is_src, weights, reorder_types[reorder_type],
                collection_size, src_data, tgt_data))
            PUT_ERR("ERROR(yac_interp_weights_add_direct): "
                    "invalid interpolation result");

    // check writing and reading
    {
      char const * weight_file_name =
        "test_interp_weights_parallel_direct.nc";
      // write weights to file
      yac_interp_weights_write_to_file(
        weights, weight_file_name, "src_grid", "dst_grid", -1);

      // read weights
      struct interp_method * method_stack[2] = {
        yac_interp_method_file_new(
          weight_file_name, -1, "src_grid", "dst_grid"), NULL};
      struct interp_weights * weights_from_file =
        yac_interp_method_do_search(method_stack, interp_grid);
      yac_interp_method_delete(method_stack);

      // check results
      if (check_direct_results(
            grids[0], is_src, weights, MAPPING_ON_SRC, 1, src_data, tgt_data))
        PUT_ERR("ERROR(yac_interp_weights_add_direct): "
                "invalid interpolation result (read from file)");
      if (comm_rank == 0) unlink(weight_file_name);
      yac_interp_weights_delete(weights_from_file);
    }

    if (is_src) {
      for (size_t collection_idx = 0; collection_idx < 4; ++collection_idx) {
        free(src_data[collection_idx][0]);
        free(src_data[collection_idx]);
      }
      free(src_data);
    } else {
      for (size_t collection_idx = 0; collection_idx < 4; ++collection_idx)
        free(tgt_data[collection_idx]);
      free(tgt_data);
    }

    yac_interp_weights_delete(weights);
    yac_interp_grid_delete(interp_grid);
    yac_dist_grid_pair_delete(grid_pair);
    yac_mpi_call(MPI_Comm_free(&local_grid_comm), MPI_COMM_WORLD);

    for (int i = 0; i < 2; ++i) yac_basic_grid_data_delete(grids[i]);
  }

  xt_finalize();

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static int check_direct_results(
  struct basic_grid_data * grid, int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, size_t collection_size,
  double *** src_data, double ** tgt_data) {

  int err_count = 0;

  struct interpolation * interpolation =
    yac_interp_weights_get_interpolation(
      weights, reorder_type, collection_size);

  if (!is_src)
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx)
      for (size_t k = 0; k < grid->num_cells; ++k)
        tgt_data[collection_idx][k] = -1;

  yac_interpolation_execute(interpolation, src_data, tgt_data);

  if (!is_src) {
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx) {
      for (size_t j = 0; j < grid->num_cells; ++j) {
        if (grid->core_cell_mask[j]) {
          if (tgt_data[collection_idx][j] !=
              (double)(grid->cell_ids[j])) err_count++;
        } else {
          if (tgt_data[collection_idx][j] != -1.0) err_count++;
        }
      }
    }
  }

  if (is_src) {
    yac_interpolation_execute_put(interpolation, src_data);
  } else {
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx)
      for (size_t k = 0; k < grid->num_cells; ++k)
        tgt_data[collection_idx][k] = -1;
    yac_interpolation_execute_get(interpolation, tgt_data);
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx) {
      for (size_t j = 0; j < grid->num_cells; ++j) {
        if (grid->core_cell_mask[j]) {
          if (tgt_data[collection_idx][j] !=
              (double)(grid->cell_ids[j])) err_count++;
        } else {
          if (tgt_data[collection_idx][j] != -1.0) err_count++;
        }
      }
    }
  }
  yac_interpolation_delete(interpolation);

  return err_count;
}

static void handle_error(int status) {
  if (status == NC_NOERR) return;
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  exit(EXIT_FAILURE);
}

static void get_basic_grid_data(
  char * filename, size_t * num_cells, size_t * num_vertices,
  size_t * num_edges) {

  int ncid;

  handle_error(nc_open(filename, NC_NOWRITE, &ncid));

  int dimid;
  handle_error(nc_inq_dimid(ncid, "cell", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_cells));
  handle_error(nc_inq_dimid(ncid, "vertex", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_vertices));
  handle_error(nc_inq_dimid(ncid, "edge", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_edges));
}

#else
int main(void) {return EXIT_SKIP_TEST;}
#endif // YAC_NETCDF_ENABLED
